
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     Interpretation of 'ASTPostgreSQL' statements into the 'DBase' type.
--     This conversion works only for simple CREATE and INSERT statements, with no queries.
--     Data types conversion is not safe, and few are supported, as long as 2LT only supports very basic types
-- 
-- by
--
--     Digo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Data.SQL.ASTToDBase where

import Data.SQL.ASTPostgreSQL
import Control.Monad (MonadPlus,mzero)
import Data.SQL.DBase
import Data.TwoLevelExchangeMisc
import Data.List as List
import Data.Map as Map
import Text.XML.HaXml.OneOfN (OneOf3(..))
import Pointless.Combinators

-- | extends the given 'DBase' with the SQL represented by 'ASTPostgreSQL'
addAstToDBase :: MonadPlus m => ASTPostgreSQL -- ^ the SQL input
		-> DBase -- ^ the input database
		-> m DBase -- ^the return database
addAstToDBase (AST stmts) (DB db) = do
	let (creates,inserts) = filterStatements stmts
	    dbatts = Map.fromDistinctAscList $ List.map (fst3 /\ snd3) db
	    dbvals = Map.fromDistinctAscList $ List.map (fst3 /\ thr3) db
	tablesatts <- createStmtsToTables dbatts creates
	let atts = dbatts `Map.union` tablesatts
	tablesvals <- insertStmtsToValues inserts atts
	let vals = dbvals `assocUnion` tablesvals	
	return $ DB $ List.map (\(table,att) -> (table,att,maybe2List $ Map.lookup table vals)) $ Map.toList atts

-- | converts the SQL represented by 'ASTPostgreSQL' to a 'DBase'
astToDBase :: MonadPlus m => ASTPostgreSQL -> m DBase
astToDBase (AST stmts) = do
	let (creates,inserts) = filterStatements stmts
	-- ast variables
	tablesatts <- createStmtsToTables Map.empty creates
	tablesvals <- insertStmtsToValues inserts tablesatts
	return $ DB $ List.map (\(table,att) -> (table,att,maybe2List $ Map.lookup table tablesvals)) $ Map.toList tablesatts

-- | returns the string value of a 'ColId'
colIdtoName :: ColId -> String
colIdtoName = Data.SQL.ASTPostgreSQL.sshow

-- | converts a list of CREATE 'Statements' to tables with 'DBattributes' only
createStmtsToTables :: MonadPlus m => Map TableName DBattributes -- ^ a map of existing tables
		      -> Statements -- ^ the list of statements to convert
		      -> m (Map TableName DBattributes) -- ^ the return list of tables
createStmtsToTables m [] = return m
createStmtsToTables m (x:xs) = do
	(table,atts) <- createStmtToTable m x 
	createStmtsToTables (Map.singleton table atts `Map.union` m) xs

-- | converts a single table to a table with 'DBattributes' only
createStmtToTable :: MonadPlus m => Map TableName DBattributes -- ^ a map of existing table
		    -> Statement -- ^ the list of statements to convert
		    -> m (TableName,DBattributes) -- ^ the return table
createStmtToTable m (CreateStmt _ name _ atts _ _ _ _) = do
	tn <- getTableName name
	if tn `elem` (keys m)
		then fail $ "Error in AST to DBase conversion: table '" ++ show name ++ "' already exists"
		else case atts of
			Nothing -> return (tn,([],[]))
			(Just a) -> do
				atts <- getTableAttrs m a
				return (tn,atts)
createStmtToTable m x = fail $ "Error in AST to DBase conversion:\n" ++ Data.SQL.ASTPostgreSQL.sshow x ++ "\n is not a create statement"

-- | filters the string 'TableName' from a 'QualifiedName'
getTableName :: MonadPlus m => QualifiedName -> m TableName
getTableName (QualNameRel (Right str)) = return $ colIdtoName str
getTableName (QualNameRelInd (Right str) _) = return $ colIdtoName str
getTableName _ = fail $ "Error in AST to DBase conversion: indirection names not supported"

-- | gets the attributes from table attributes definition ('TableElementList')
getTableAttrs :: MonadPlus m => Map TableName DBattributes -> TableElementList -> m DBattributes
getTableAttrs m (MyLst lst) = do
	(pkeys,fkeys) <- getTableKeys m lst
	atts <- mapM (uncurry (getTableAtt fkeys)) (zip [1..length lst] lst) >>= return . concat
	return $ (List.map fst >< List.map fst) $ List.partition (isPKey pkeys) atts
	where
	isPKey :: [VariableName] -> (DBattribute,Bool) -> Bool
	isPKey pks (_,True) = True
	isPKey pks ((_,n,_,_,_,_),_) = elem n pks

-- The boolean is for primary key or not distinction

-- | gets the definition of a 'DBattribute' from its 'ASTPostgreSQL' representation ('TableElement')
getTableAtt :: MonadPlus m => Map VariableName ForeignKey -> Int -- ^ the index of the attribute in the original table
	      -> TableElement -- ^ the attribute 'ASTPostgreSQL' definition
	      -> m [(DBattribute,Bool)]-- ^ (the return attribute,the boolean for primary key or not distinction)
getTableAtt m i (OneOf3 (CDef n t cstrs)) = do
	tp <- getAttType t
	let att = (i,colIdtoName n,tp,False,Nothing,Nothing)
	(isnull,ispkey,def,fk) <- getAttConstraints m cstrs att
	let att' = (i,colIdtoName n,tp,isnull,def,fk)
	return [(att',ispkey)]
	
getTableAtt _ _ (ThreeOf3 x) = return []
getTableAtt _ _ (TwoOf3 _) = fail "Error in AST to DBase conversion: table-like clause definition not supported"

-- | gets the 'DBtype' from a 'Typename' definition
getAttType :: MonadPlus m => Typename -> m DBtype
getAttType (TypenameConst _ x _) = getSimpleType x

-- | gets the 'DBtype' from a 'SimpleTypename' definition
getSimpleType :: MonadPlus m => SimpleTypename -> m DBtype
getSimpleType (SimpTypeGeneric x) = fail "Error in AST to DBase conversion: generic types not supported"
getSimpleType (SimpTypeNum x) = return $ getNumeric x
getSimpleType (SimpTypeBit x) = return $ getBit x
getSimpleType (SimpTypeChar x) = return $ getCharacter x
getSimpleType (SimpTypeDate x) = return $ getDate x
getSimpleType (SimpTypeInterval _ _ _) = fail "Error in AST to DBase conversion: interval type not supported"
getSimpleType (SimpType x y) = fail "Error in AST to DBase conversion: string types not supported"

getDate :: ConstDatetime -> DBtype
getDate (ConstTimestamp _ _) = VARCHAR 1024

-- | gets the 'DBtype' from a 'Numeric' definition
getNumeric :: Numeric -> DBtype
getNumeric NumInt = INTEGER
getNumeric NumInteger = INTEGER
getNumeric NumBigInt = INTEGER
getNumeric NumReal = FLOAT
getNumeric (NumFloat _) = FLOAT
getNumeric NumDouble = FLOAT
getNumeric (NumDecimal x) = getNumeric $ NumNumeric x
getNumeric (NumNumeric (OptNumericConst (Just (x,Just y)))) = NUMERIC x y
getNumeric (NumNumeric (OptNumericConst (Just (x,Nothing)))) = NUMERIC x 12
getNumeric (NumNumeric (OptNumericConst Nothing)) = NUMERIC 64 12
getNumeric NumBool = BOOLEAN

-- | gets the 'DBtype' from a 'Bit' definition
getBit :: Bit -> DBtype
getBit _ = VARCHAR 2048
-- ^ the 'Bit' type is represented by an SQL string 'VARCHAR'

-- | gets the 'DBtype' from a 'Character' definition
getCharacter :: Character -> DBtype
getCharacter (CharDef char2 (Just i) _) = getChar2 char2 i
getCharacter (CharDef char2 Nothing _) = getChar2 char2 2048

-- gets the 'DBtype' from a 'Charater2' definition
getChar2 :: Character2 -- ^ the 'Character2' definition
	   -> Iconst -- ^ the maximum length of the string
	   -> DBtype -- ^ the returned type
getChar2 (Char2Character _) _ = VARCHAR 1
getChar2 Char2VarChar i = VARCHAR i
getChar2 (Char2National _) _ = VARCHAR 1
-- ^ the returned 'DBtype' is always a SQL string 'VARCHAR'

-- | gets the constraints for a list of constraints for an attribute
getAttConstraints :: MonadPlus m => Map VariableName ForeignKey -> ColQualList -- ^ the list of constraints in 'ASTPostgreSQL'
		    -> DBattribute -- ^ the attribute to which the constraints refer
		    -> m (Bool,Bool,Maybe DBvalue,Maybe ForeignKey) -- ^ (the null or not null constraint,the primary key constraint,the default value)
getAttConstraints m l att = do
	cstrs <- mapM (\x -> getAttConstraint x att) l
	let isnull = or $ List.map fst4 cstrs
	    ispkey = or $ List.map snd4 cstrs
	    def = singleDefault $ foldr (curry (uncurry appendMaybe . (thr4 >< id))) [] cstrs
	    fk = singleFK $ foldr (curry (uncurry appendMaybe . (fth4 >< id))) [] (cstrs ++ mapfk)
	    mapfk = case (Map.lookup (snd6 att) m) of
			Nothing  -> []
			(Just x) -> [(False,False,Nothing,Just x)]
	return (isnull,ispkey,def,fk)
	where
	singleDefault :: MonadPlus m => [a] -> m a
	singleDefault [] = mzero
	singleDefault [el] = return el
	singleDefault _ = fail $ "Error in AST to DBase conversion: more than one default values assigned to variable " ++ show (snd6 att)
	
	singleFK :: MonadPlus m => [a] -> m a
	singleFK [] = mzero
	singleFK [el] = return el
	singleFK _ = fail $ "Error in AST to DBase conversion: more than one foreign keys assigned to variable " ++ show (snd6 att)

-- SQL default is not null, when not specified

-- | gets the constraints defined in a 'ColConstraint'
getAttConstraint :: MonadPlus m => ColConstraint -- ^ the constraint 'ASTPostgreSQL' representation
		   -> DBattribute -- ^ the attribute used for type checking in the default constraint later
		   -> m (Bool,Bool,Maybe DBvalue,Maybe ForeignKey) -- ^ the constraints
getAttConstraint (OneOf3 (_,_,x)) att = getAttConstraint' x att
getAttConstraint (TwoOf3 x) att = getAttConstraint' x att
getAttConstraint _ _ = return (False,False,Nothing,Nothing)
-- ^ only one of the constraints can be returned as true, as long as its only one constraint definition

-- | 'getAttConstraint' auxiliary function
getAttConstraint' :: MonadPlus m => ColConstraintElem -> DBattribute -> m (Bool,Bool,Maybe DBvalue,Maybe ForeignKey)
getAttConstraint' CstrNotNull _ = return (False,False,Nothing,Nothing)
getAttConstraint' CstrNull _ = return  (True,False,Nothing,Nothing)
getAttConstraint' (CstrPrimaryKey _) _ = return (False,True,Nothing,Nothing)
getAttConstraint' (CstrDefault bexp) att = do
	tc <- getBExpr bexp >>= typeCheck att
	return (False,False,Just tc ,Nothing)
getAttConstraint' (CstrReferences ft (OptColList (Just (MyLst [fc]))) Nothing Nothing) _ = do
	tn <- getTableName ft
	return (False,False,Nothing,Just (Nothing,tn,colIdtoName fc))
getAttConstraint' (CstrUnique _) _ = return (False,True,Nothing,Nothing)
getAttConstraint' _ _ = fail "Error in AST to DBase conversion: unsupported constraint"

-- | gets the 'DBvalue' from an 'AExpr'
getAExpr :: MonadPlus m => AExpr -> m DBvalue
getAExpr (AExpCExp cexp) = getCExpr cexp
getAExpr _ = fail "Error in AST to DBase conversion: aexp not supported"

-- | gets the 'DBvalue' from a 'BExpr'
getBExpr :: MonadPlus m => BExpr -> m DBvalue
getBExpr (BExpCExp cexp) = getCExpr cexp
getBExpr _ = fail "Error in AST to DBase conversion: bexp not supported"

-- | gets the 'DBvalue' from a 'CExpr'
getCExpr :: MonadPlus m => CExpr -> m DBvalue
getCExpr (CExpAExpC x) = getAExprConst x
getCExpr (CExpAExp x []) = getAExpr x
getCExpr (CExpAExp _ _) = fail "Error in AST to DBase conversion: indirection not supported in cexp"
getCExpr _ = fail "Error in AST to DBase conversion: cexp not supported"

-- | gets the 'DBvalue' from a constant 'AExpr' ('AExprConst')
getAExprConst :: MonadPlus m => AExprConst -> m DBvalue
getAExprConst (AExpCI x) = return $ NUMERIC_T $ toEnum x
getAExprConst (AExpCF x) = return $ VARCHAR_T x
getAExprConst (AExpCS x) = return $ VARCHAR_T x
getAExprConst (AExpCB x) = return $ VARCHAR_T x
getAExprConst (AExpCX x) = return $ VARCHAR_T x
getAExprConst AExpCTrue = return $ BOOLEAN_T True
getAExprConst AExpCFalse = return $ BOOLEAN_T False
getAExprConst AExpCNull = return $ Data.SQL.DBase.NULL
getAExprConst (AExpCType _ s) = return $ VARCHAR_T s
getAExprConst (AExpCInt _ _ _ _) = fail "Error in AST to DBase conversion: interval type not supported"

-- gets the names of the attributes that are primary keys or foreign keys from a list of table elements
getTableKeys :: MonadPlus m => Map TableName DBattributes -> [TableElement] -> m ([VariableName],Map VariableName ForeignKey) -- ^ Variable Name, foreign table name, foreign variable name
getTableKeys m l = do
	let pks = concat . List.map (either (:[]) (const []))
	    fks = Map.fromDistinctAscList . List.map (\(a,b,c,d) -> (b,(a,c,d))) . concat . List.map (either (const []) (:[]))
	mapM (getTableKeys' m) l >>= return . (pks /\ fks) . concat
	where
	getTableKeys' :: MonadPlus m => Map TableName DBattributes -> TableElement -> m [Either VariableName (Maybe String,VariableName,TableName,VariableName)]
	getTableKeys' _ (ThreeOf3 (TConstraint _ (CstrElemPrimaryKey (MyLst keys) _))) = return $ List.map (Left . colIdtoName) keys
	getTableKeys' ts (ThreeOf3 (TConstraint c (CstrElemForeignKey (MyLst fkeys) ft (OptColList fvs) _ _ _))) = do
		let fkname = case c of
			Nothing  -> Nothing
			(Just x) -> Just $ colIdtoName x
		let len = length variables
		    variables = List.map colIdtoName fkeys
		ftable <- getTableName ft
		fvariables <- do
			case fvs of
				Nothing            -> getTableVariables ftable ts
				Just (MyLst fvars) -> return $ List.map colIdtoName fvars		
		return $ List.map Right $ zip4 (replicate len fkname) variables (replicate len ftable) fvariables
	getTableKeys' _ _ = return []

-- INSERT statements conversion: reponsible for the values in the tables

-- | converts INSERT 'Statements' to a list of tables with attributes only
insertStmtsToValues :: MonadPlus m => Statements -- ^ the CREATE 'Statements'
		      -> Map TableName DBattributes -- ^ the declaration of existing tables
		      -> m (Map TableName DBvalues) -- ^ the converted table values
insertStmtsToValues atts createtables = do
	let tables = nub $ keys createtables
	inserts <- mapM (\x -> insertStmtToValues x createtables) atts
	return $ Map.fromDistinctAscList $ List.map (\table -> (table,List.map snd $ List.filter ((==table) . fst) inserts)) tables

-- | converts an INSERT 'Statement' to a table with values only
insertStmtToValues :: MonadPlus m => Statement -> Map TableName DBattributes -> m (TableName,LineDBvalues)
insertStmtToValues (InsertStmt qn ins) tables = do
	name <- getTableName qn
	case (Map.lookup name tables) of
		Nothing -> fail $ "Error in AST to DBase conversion: no table '" ++ name ++ "' found"
		(Just atts) -> do
			stmts <- insertStmtToValues' name ins atts
			return (name,stmts)
insertStmtToValues x _ = fail $ "Error in AST to DBase conversion:\n" ++ Data.SQL.ASTPostgreSQL.sshow x ++ "\nis not a insert statement"

-- | 'insertStmtToValues' auxiliary function		
insertStmtToValues' :: MonadPlus m => TableName -> InsertRest -> DBattributes -> m LineDBvalues
insertStmtToValues' n (InsValues (MyLst targets)) (l,r) = do
	let atts = sort $ l ++ r
	vals <- mapM aux $ zip atts targets
	let (lvals,rvals) = List.partition (\(i,_) -> lookup6 i l) vals	
	if (length lvals /= length l) || (length rvals /= length r)
		then fail $ "Error in AST to DBase conversion: values for table '" ++ n ++ "' don't match with attributes"
		else return $ (List.map snd lvals,List.map snd rvals)
	where
	aux x = do
		let l = fst6 $ fst x
		r <- fillValue x
		return (l,r)
insertStmtToValues' _ InsDefaultValues (l,r) = do
	l <- mapM fillDefaultValue l
	r <- mapM fillDefaultValue r
	return (l,r)
insertStmtToValues' n (InsSelect _) _ = fail $ "Error in AST to DBase conversion: no insert queries supported for table '" ++ n ++ "'"
insertStmtToValues' n (InsColValues (MyLst vals) (MyLst targets)) (l,r) = do
	values <- mapM (getInsertColumnItem n) vals >>= return . \x -> zip x targets
	let (lvals,rvals) = List.partition findPkeyItem values
	a <- orderAtts lvals l
	b <- orderAtts rvals r
	return (a,b)
	where
	findPkeyItem :: (String,InsertTargetEl) -> Bool
	findPkeyItem (x,_) = maybe2Bool $ find (\(_,n,_,_,_,_) -> x == n) l
	
	-- puts the target elements in the AST declaration in the same order as the table's attributes
	orderAtts :: MonadPlus m => [(String,InsertTargetEl)] -> [DBattribute] -> m[DBvalue]
	orderAtts x = mapM (fillValue . matchAtt x)

	-- matches an attribute with a target element
	matchAtt :: [(String,InsertTargetEl)] -> DBattribute -> (DBattribute,InsertTargetEl)
	matchAtt l att@(_,n,a,b,def,_) =
		case (find ((==n) . fst) l) of
			Nothing -> (att,Right DEFAULT)
			(Just (_,y)) -> (att,y)
	
	-- the name of an insert target element
	getInsertColumnItem :: MonadPlus m => TableName -> InsertColumnItem -> m String
	getInsertColumnItem _ (InsColItem x []) = return $ showColId x
	getInsertColumnItem n (InsColItem x _) = fail $ "Error in AST to DBase conversion: no indirection insert column item names supported for table '" ++ n ++ "'"
	
	--the string name inside a ColId
	showColId :: ColId -> String
	showColId (OneOf3 x) = x
	showColId (TwoOf3 x) = show x
	showColId (ThreeOf3 x) = show x
	
insertStmtToValues' n (InsCol _ _) _ = fail $ "Error in AST to DBase conversion: no insert queries supported for table '" ++ n ++ "'" 

-- | fills a 'DBattribute' with its default 'DBvalue'
fillDefaultValue :: MonadPlus m => DBattribute -> m DBvalue
fillDefaultValue (_,_,_,_,Just def,_) = return def
fillDefaultValue (_,n,_,_,Nothing,_) = fail $ "Error in AST to DBase conversion: no default value specified for attribute '" ++ n ++ "'"


-- | fills a 'DBattribute' with the value received from 'InsertTargetEl'
fillValue :: MonadPlus m => (DBattribute,InsertTargetEl) -> m DBvalue
fillValue (att,Left aexp) = getAExpr aexp >>= typeCheck att
fillValue (x,_) = fillDefaultValue x

-- | CREATE and INSERT statements filtering
filterStatements :: Statements -> (Statements,Statements)
filterStatements = (fst . List.partition isCreateStmt) /\ (List.filter isInsertStmt . snd . List.partition isCreateStmt)

-- | checks is a 'Statement' is a CREATE 'Statement'
isCreateStmt (CreateStmt _ _ _ _ _ _ _ _) = True
isCreateStmt _ = False

-- | checks is a 'Statement' is a INSERT 'Statement'
isInsertStmt (InsertStmt _ _) = True
isInsertStmt _ = False
