
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     DBase: a non-typesafe SQL database representation.
--     Supports tables with values and primary key, null or not null and default constraints
-- 
-- by
--
--     Digo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------


module Data.SQL.DBase where

import GHC.Float
import Pointless.Combinators
import Data.Map as Map
import Control.Monad (MonadPlus)
import Data.TwoLevelExchangeMisc
import Data.List as List

-- * Data Types

-- | Data Base definition
data DBase = DB [DBtable]
    deriving Show

-- | Table = Table name, attributes (Create Table), Values (Insert)

type DBattributes = ([DBattribute], [DBattribute])

type DBvalues = [LineDBvalues]
-- ^ pre-condition: each sub list of the list of values must be the same lenght of the atributes list (for each attribute you must have a value)
-- this pre-condition is not checked

type LineDBvalues = ([DBvalue], [DBvalue])

type TableName = String

type VariableName = String

type DBtable = (TableName,DBattributes,DBvalues)

-- | Represents a foreign key attributes.
-- (foreign key constraint name, referred table name, referred column name)
type ForeignKey = (Maybe String,TableName,VariableName)

-- | Attribute = (index,name,type,null or not null,default value,froreign key)
type DBattribute = (Int,VariableName, DBtype, Bool, Maybe DBvalue,Maybe ForeignKey)

instance (Show a,Show b,Show c,Show d,Show e,Show f) => Show (a,b,c,d,e,f) where
	show (a,b,c,d,e,f) = parenthesis $ [show a,show b,show c,show d,show e,show f]

type DBattributeNoIndex = (VariableName, DBtype, Bool, Maybe DBvalue)

-- | Attributes supported types
data DBtype = VARCHAR Int | NUMERIC Int Int | INTEGER | BOOLEAN | FLOAT
    deriving (Show,Ord,Eq)

-- | Attributes values
data DBvalue = VARCHAR_T String | NUMERIC_T Double | INTEGER_T Int | BOOLEAN_T Bool | FLOAT_T Float | NULL
    deriving (Eq,Show,Ord)

-- * Functions

-- | Returns a list of variables belonging to a 'DBtable'.
-- The variables are in their original order.
getTableVariables :: MonadPlus m => TableName -> Map TableName DBattributes -> m [VariableName]
getTableVariables t ts =
		case (Map.lookup t ts) of
			Nothing -> fail $ "DBase function error: table " ++ show t ++ " not found\n"
			(Just atts) -> return $ getTableVariables' atts
	where
	getTableVariables' :: DBattributes -> [VariableName]
	getTableVariables' = List.map snd . sort . List.map (fst6 /\ snd6) . uncurry (++)

-- | matches the 'DBattribute' with the 'DBvalue'
-- If they do not match the 'DBvalue' is converted to a valid value or an error is returned
typeCheck :: MonadPlus m => DBattribute -> DBvalue -> m DBvalue
typeCheck (_,n,VARCHAR i,_,_,_) (VARCHAR_T s) = if (length s > i)
	then fail $ "DBase function error:  exceeded maximum string value for attribute '" ++ n ++ "'"
	else return $ VARCHAR_T s
typeCheck att (VARCHAR_T s) = numericstring att s
typeCheck att (NUMERIC_T d) = numeric att d
typeCheck (_,_,BOOLEAN,_,_,_) (BOOLEAN_T b) = return $ BOOLEAN_T b
typeCheck (_,_,_,True,_,_) NULL = return $ NULL
typeCheck (_,n,_,False,_,_) NULL = fail $ "DBase function error:  attribute '" ++ n ++ "' not nullable"
typeCheck (_,n,_,_,_,_) _ = fail $ "DBase function error:  error in value assignment for attribute '" ++ n ++ "'"

-- | returns the correct 'DBvalue' of a numeric constant
numeric :: MonadPlus m => DBattribute -> Double -> m DBvalue
numeric (_,_,INTEGER,_,_,_) i = return $ INTEGER_T $ fromEnum i
numeric (_,_,FLOAT,_,_,_) f = return $ FLOAT_T $ double2Float f
numeric (_,_,NUMERIC _ _,_,_,_) d = return $ NUMERIC_T d
numeric (_,n,_,_,_,_) _ = fail $ "DBase function error:  error in numeric value assignment for attribute '" ++ n ++ "'"
-- ^ all numeric values are converted to double and need to be matched with the correct 'DBattribute'

-- | tries to match a string with a numeric 'DBattribute'
-- this function is only used when the lexer returns an invalid value
-- casting may fail, but at least it tries to avoid an error
numericstring :: MonadPlus m => DBattribute -> String -> m DBvalue
numericstring (_,_,INTEGER,_,_,_) s = return $ INTEGER_T (read s :: Int)
numericstring (_,_,FLOAT,_,_,_) s = return $ FLOAT_T (read s :: Float)
numericstring (_,_,NUMERIC _ _,_,_,_) s = return $ NUMERIC_T (read s :: Double)
numericstring (_,n,_,_,_,_) _ = fail $ "DBase function error:  error in string value assignment for attribute '" ++ n ++ "'"
