
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     DBase: a type-safe SQL database representation.
--     Supports tables with values and primary key, null or not null and default constraints
-- 
-- by
--
--     Digo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Data.SQL.DBaseTS where

import GHC.Float
import GHC.Base as Base
import Data.Map as Map
import Data.TwoLevelExchangeMisc as Misc
import Data.List as List
import Pointless.Combinators
-- * Data Types

-- | Data Base definition
data DBase = DB [DBTable]
	deriving Show

type DDL = DBase -- type
type DML = DBase -- values

data DBTable where
	DBTable	:: DBtable a b -> DBTable

type TableName = String

type VariableName = String

data DBtable a b = DBtable TableName (DBtype a) (DBtype b) (Map a b)

-- | DBattribute = (VariableName,Default Value) 
type DBattribute a = (VariableName,Maybe a)

data DBAttribute where
	DBAttribute :: DBattribute a -> DBAttribute

data DBType where
	DBType :: DBtype a -> DBType

data DBtype a where
	VarChar :: DBattribute String -> Int -> DBtype String
	Numeric :: DBattribute Double -> Int -> Int -> DBtype Double
	Integer :: DBattribute Int -> DBtype Int
	Boolean :: DBattribute Bool -> DBtype Bool
	Float	:: DBattribute Float -> DBtype Float
	Null    :: DBtype a -> DBtype (Either a ())
	Prod	:: DBtype a -> DBtype b -> DBtype (a,b)
	FK	:: Maybe String -> TableName -> VariableName -> DBtype a -> DBtype a
	One	:: DBtype ()

data DBValue where
	DBValue :: DBtype a -> a -> DBValue

-- | An encapsulation for a 'DBtype' and a corresponding list of values.
data DBValueList where
	DBValueList :: DBtype a -> [a] -> DBValueList

-- * Instances

instance Eq DBTable where
	(DBTable (DBtable t1 _ _ _)) == (DBTable (DBtable t2 _ _ _)) = t1 == t2

instance Ord DBTable where
	(DBTable (DBtable t1 _ _ _)) > (DBTable (DBtable t2 _ _ _)) = t1 > t2
	(DBTable (DBtable t1 _ _ _)) < (DBTable (DBtable t2 _ _ _)) = t1 < t2


instance Show DBTable where
	show (DBTable t) = show t

instance Show DBType where
	show (DBType x) = show x

instance Show (DBtable a b) where
	show (DBtable n ta tb m) = "(DBtable " ++ n ++ " " ++ gshow ta tb m ++ ")"
		where	
		gshow :: DBtype a -> DBtype b -> Map a b -> String
		gshow a b m = show a ++ " " ++ show b ++ " " ++ showMap m
			where
			delim = ('(':) . (++")") . init . foldr (\h r -> h ++ ',' : r) []
			showMap = concat . Base.map showEntry . uncurry zip . (Map.keys /\ Map.elems)
			showEntry (k,v) = delim $ (drawValues $ DBValue a k) ++ (drawValues $ DBValue b v)

instance Show (DBtype a) where
	show (VarChar att i)	 = "(VarChar " ++ show att ++ " " ++ show i ++ ")"
	show (Numeric att i1 i2) = "(Numeric " ++ show att ++ " " ++ show i1 ++ " " ++ show i2 ++ ")"
	show (Integer i)	 = "(Integer " ++ show i ++ ")"
	show (Boolean b)	 = "(Boolean " ++ show b ++ ")"
	show (Float f)		 = "(Float " ++ show f ++ ")"
	show (Null dbt)		 = "(Null " ++ show dbt ++ ")"
	show (Data.SQL.DBaseTS.Prod a b) = "(Prod " ++ show a ++ " " ++ show b ++ ")"
	show Data.SQL.DBaseTS.One = "One"
	show (FK n tn cn t) = "(FK " ++ show n ++ " " ++ show tn ++ " " ++ show cn ++ " " ++ show t

instance Show DBValue where
	show (DBValue t v) = gshow t v
		where
		gshow :: DBtype a -> a -> String
		gshow (VarChar att l) s = '\'' : escapeQuotes s ++ "'"
		gshow (Numeric att a b) n = show n
		gshow (Integer att) i = show i
		gshow (Boolean att) b = show b
		gshow (Float att) f = show f
		gshow (Null t) (Left v) = gshow t v
		gshow (Null t) (Right ()) = "NULL"
		gshow (FK _ _ _ t) v = gshow t v
		gshow One () = "TRUE"
		gshow t _ = error $ "\n\ngshow error :" ++ show t

-- * Functions

-- | Applies a generic function to an encapsulated 'DBtable'.
applyDBTable :: (forall a b . DBtable a b -> c) -> DBTable -> c
applyDBTable f (DBTable t) = f t	

-- | Applies a generic function to an encapsulated 'DBtype'.
applyDBType :: (forall a . DBtype a -> b) -> DBType -> b	 
applyDBType f (DBType t) = f t	 

-- | Applies a generic function to an encapsulated 'DBtype' and it's corresponding list of values.
applyDBValueList :: (forall a . DBtype a -> [a] -> b) -> DBValueList -> b
applyDBValueList f (DBValueList t vs) = f t vs

-- | Fetches the name of a 'DBtype' variable.	 
getName :: DBtype a -> String	 
getName (VarChar (n,_) _) = n	 
getName (Null x) = getName x	 
getName (Numeric (n,_) _ _) = n	 
getName (Integer (n,_)) = n	 
getName (Boolean (n,_)) = n	 
getName (Float (n,_)) = n
getName (FK _ _ _ t) = getName t
getName One = "TLT_One"	 

-- | Fetches the foreign key definition from a 'DBtype' variable.
getFKeys :: DBtype a -> [(Maybe String,VariableName,TableName,VariableName)]
getFKeys (FK fk tn cn t) = (fk,getName t,tn,cn) : getFKeys t
getFKeys _ = []

-- | Pretty prints a list of 'DBvalue's
drawValues :: DBValue -> [String]
drawValues (DBValue (Data.SQL.DBaseTS.Prod a b) (x,y)) = drawValues (DBValue a x) ++ drawValues (DBValue b y)
drawValues x = [show x]

-- | Maps a function into a 'DBType'
mapDBType :: (DBType -> b) -> DBType -> [b]
mapDBType f (DBType t@(Data.SQL.DBaseTS.Prod a b)) = mapDBType f (DBType a) ++ mapDBType f (DBType b)
mapDBType f x = [f x]

-- | Removes all One occurences from a 'DBtype'.
removeOnes :: DBtype a -> [a] -> DBValueList
removeOnes (Prod a b) v = aux (removeOnes a (List.map fst v)) (removeOnes b (List.map snd v))
	where
	aux :: DBValueList -> DBValueList -> DBValueList
	aux (DBValueList One []) dbvl = dbvl
	aux dbvl (DBValueList One []) = dbvl
	aux (DBValueList a va) (DBValueList b vb) = DBValueList (Prod a b) $ zip va vb
removeOnes One _ = DBValueList One []
removeOnes (Null One) _ = DBValueList One []
removeOnes (FK x y z t) lv = aux $ removeOnes t lv
	where
	aux :: DBValueList -> DBValueList
	aux x@(DBValueList One []) = x
	aux (DBValueList a lb) = DBValueList (FK x y z a) lb
removeOnes t lv = DBValueList t lv

-- | Distributes a 'DBValueList' into a list of 'DBValue's.
distDBValueList :: DBValueList -> [DBValue]
distDBValueList (DBValueList t lv) = List.map (DBValue t) lv


