
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     DBase type-safe to 2LT Type conversion
-- 
-- by
--
--     Diogo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Data.SQL.DBaseTStoType where

import Data.Transform.Type as Type
import Data.Transform.TwoLevelCombinators
import Data.Transform.RulesWithRefs as TLT
import Pointless.Combinators
import Data.SQL.DBaseTS as DBaseTS
import Data.TwoLevelExchangeMisc
import Data.Map as Map
import Data.List as List
import Control.Monad.State
import Control.Monad (MonadPlus)

-- | The data type to represent an index.
-- (index key , referred table name , referred column name)
type Index = (KeyId,TableName,VariableName)

-- | The data type of the 'State' monad used for index discovery.
-- (actual table name , index counter)
type IndexState = (TableName,Int)

-- | The return type for the 'State' monad used for index discovery.
-- (dynamic type , list of discovered indexes)
type IndexRet = (DynType,[Index])

-- | Creates a product of 'IndexRet's.
indexProd :: IndexRet -> IndexRet -> IndexRet
indexProd (DynT t1,l1) (DynT t2,l2) = (DynT $ Type.Prod t1 t2,l1 ++ l2)

-- | Create a dynamic type in the form of a 'DynType' from a 'DBase'
-- The created 'Type' structure has tags for table and variable names.
dBase2dyn :: MonadPlus m => DBase -> m Dynamic
dBase2dyn db = dBase2dyntype db >>= aux
	where
	aux :: MonadPlus m => DynType -> m Dynamic
	aux (DynT t) = dBase2type db t >>= return . Dyn t
		
-- | Creates 'Index'es in the proper variables of an encapsulated relational 'Type' structure.
createIndexesDyn :: MonadPlus m => DynType -> [Index] -> m DynType
createIndexesDyn (DynT t) lidx = createIndexes t lidx >>= return . DynT

-- | Creates 'Index'es in the proper variable of a relational 'Type' structure.
createIndexes :: MonadPlus m => Type a -> [Index] -> m (Type a)
createIndexes t [] = return t
createIndexes t [x] = createIndex t x
createIndexes t (x:xs) = do
	idx <- createIndex t x
	createIndexes idx xs

-- | Creates an 'Index' in the proper variables of a relational 'Type' structure.
createIndex :: MonadPlus m => Type a -> Index -> m (Type a)
createIndex (Type.Prod a b) i = do
	l <- createIndex a i
	r <- createIndex b i
	return $ Type.Prod l r
createIndex t@(Ann l (Map a b)) i = (liftM (==snd3 i) (mb2mplus $ getFieldName l) >>= bool2mplus >> (do
			left <- createIndex a i
			right <- createIndex b i
			return $ Ann l $ Map left right)) `mplus` return t
createIndex t@(Ann l basict) (idk,_,cn) = (liftM (==cn) (mb2mplus $ getFieldName l) >>= bool2mplus >> (return $ Ann (l {key = Just idk}) basict)) `mplus` return t
createIndex Type.One _ = return $ Type.One
createIndex t _ = fail $ "DBase type-safe to Type conversion error: creating index in type '" ++ show t ++ "'"

-- Converts a list of 'DBTable' to a dynamic type.
dBase2dyntype :: MonadPlus m => DBase -> m DynType
dBase2dyntype (DB ts) = uncurry createIndexesDyn $ evalState (aux tables) ("",0)
	where
	tables = ts
	aux :: [DBTable] -> State IndexState IndexRet
	aux [] = return (DynT Type.One,[])
	aux [x] = dBTable2dyn x
	aux (x:xs) = do
			l <- dBTable2dyn x
			r <- aux xs
			return $ indexProd l r

-- | Create a dynamic type in the form of a 'DynType' from a 'DBTable'
dBTable2dyn :: DBTable -> State IndexState IndexRet
dBTable2dyn (DBTable (DBtable n k a _)) = 
		do
		(tn,i) <- get
		put (n,i)
		l <- dBtype2dyn k
		r <- dBtype2dyn a
		return $ aux l r
	where
	aux :: IndexRet -> IndexRet -> IndexRet
	aux (DynT l,l1) (DynT r,l2) = (DynT $ Ann (emptyTag {fieldName = Just n}) $ Map l r,l1 ++ l2)

-- | Creates a dynamic type in the form of a 'DynType' type from a 'DBtype'.
dBtype2dyn :: DBtype a -> State IndexState IndexRet
dBtype2dyn (DBaseTS.Prod a b) = do
				l <- dBtype2dyn a
				r <- dBtype2dyn b
				return $ indexProd l r

dBtype2dyn (VarChar (n,_) _) = return (DynT $ Ann (emptyTag {fieldName = Just n}) String,[])
dBtype2dyn (Numeric (n,_) _ _) = return (DynT $ Ann (emptyTag {fieldName = Just n}) Int,[])
dBtype2dyn (Integer (n,_)) = return (DynT $ Ann (emptyTag {fieldName = Just n}) Int,[])
dBtype2dyn (Boolean (n,_)) = return (DynT $ Ann (emptyTag {fieldName = Just n}) Bool,[])
dBtype2dyn (Float (n,_)) = return (DynT $ Ann (emptyTag {fieldName = Just n}) Int,[])
dBtype2dyn (Null t) = do
			x <- dBtype2dyn t
			return $ aux x
	where
	aux :: IndexRet -> IndexRet
	aux (DynT (Ann l t),li) = (DynT $ Ann l $ Either t Type.One,li)
	
dBtype2dyn (FK fk ftn fcn t) = do
				x <- dBtype2dyn t
				ret <- aux x
				return ret
	where
	aux :: IndexRet -> State IndexState IndexRet
	aux ret@(DynT (Ann l tp),idxs) = maybe (return ret) (\cn -> do
		(tn,i) <- get
		let idx = (i,[0])
		put (tn,i+1)
		return (DynT $ Ann (addReference fk idx l) tp , (idx,ftn,fcn):idxs)) $ getFieldName l
dBtype2dyn DBaseTS.One = return (DynT Type.One,[])

-- | Converts a 'DBase' and it's corresponding type representation and fills the values for that type.
-- The given'Type' structure is supposed to have tags for table and variable names.
dBase2type :: MonadPlus m => DBase -> Type a -> m a
dBase2type (DB ts) t = mapProdType t
	where
	mapProdType :: MonadPlus m => Type a -> m a
	mapProdType (Type.Prod a b) = do
		l <- mapProdType a
		r <- mapProdType b
		return (l,r)
	mapProdType x = lookupName ts x

-- | Finds the value for the 'DBTable' with name equal to the tag of a 'DBtype' and fills the values for that type.
lookupName :: MonadPlus m => [DBTable] -> Type a -> m a
lookupName ts (Ann l t) = mb2mplus $ getFieldName l >>= \x -> dBTable2type (head $ List.filter (isTableName x) ts) t
	where
	isTableName :: String -> DBTable -> Bool
	isTableName n (DBTable (DBtable tn _ _ _)) = n == tn

-- | Converts a 'DBTable' and it's corresponding type representation and fills the values for that type.
dBTable2type :: MonadPlus m => DBTable -> Type a -> m a
dBTable2type (DBTable (DBtable n k a m)) (Map x y) = return $ Map.fromDistinctAscList $ zip (dBtype2type k x $ Map.keys m) (dBtype2type a y $ Map.elems m)
dBTable2type _ _ = fail "DBase type-safe to Type conversion error: when converting a table to a map"

-- | Converts a 'DBtype' and it's values and creates the equivalent values for the given 'Type' structure.
dBtype2type :: DBtype a -> Type b -> [a] -> [b]
dBtype2type dbt t dbvs = apply (normalizeDBtype $ flattenDBtype dbt dbvs) t
	where
	apply :: DBValueList -> Type a -> [a]
	apply (DBValueList a b) c = List.map (dBtype2type' a c) b
	
	prodDBValueList :: DBValueList -> DBValueList -> DBValueList
	prodDBValueList (DBValueList t vs) (DBValueList t1 vs1) = DBValueList (DBaseTS.Prod t t1) (zip vs vs1)
	
	flattenDBtype :: DBtype a -> [a] -> [DBValueList]
	flattenDBtype (DBaseTS.Prod a b) = uncurry (++) . (flattenDBtype a >< flattenDBtype b) . unzip
	flattenDBtype t = (:[]) . DBValueList t
	
	normalizeDBtype :: [DBValueList] -> DBValueList
	normalizeDBtype [x] = x
	normalizeDBtype l = prodDBValueList (normalizeDBtype $ init l) (last l)	
	
	dBtype2type' :: DBtype a -> Type b -> a -> b
	dBtype2type' dbt (Ann _ t) = dBtype2type' dbt t
	dBtype2type' (DBaseTS.Prod a b) (Type.Prod c d) = (dBtype2type' a c >< dBtype2type' b d)
	dBtype2type' (VarChar _ _) String = id
	dBtype2type' (Numeric _ _ _) Int = fromEnum
	dBtype2type' (Integer _) Int = id
	dBtype2type' (Boolean _) Bool = id
	dBtype2type' (Float _) Int = fromEnum
	dBtype2type' (Null dbt) (Either Type.One t) = either (Right . dBtype2type' dbt t) Left
	dBtype2type' (Null dbt) (Either t Type.One) = either (Left . dBtype2type' dbt t) Right
	dBtype2type' DBaseTS.One Type.One = id
	dBtype2type' (FK _ _ _ t) tp = dBtype2type' t tp





