
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     DBase non-type-safe to DBase type-safe data types conversion.
-- 
-- by
--
--     Diogo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Data.SQL.DBasetoDBaseTS where

import Data.SQL.DBase as DB
import Data.SQL.DBaseTS as TS
import Data.Map as Map
import Data.List as List
import Control.Monad (MonadPlus)

-- | Converts a non-type-safe relational database representation ('DB.DBase') into a type-safe relational database representation ('TS.DBase').
dBase2dBaseTS :: MonadPlus m => DB.DBase -> m TS.DBase
dBase2dBaseTS (DB.DB ts) = mapM dBtable2dBTableTS ts >>= return . TS.DB

-- | Converts a non-type-safe table ('DB.DBtable') into a type-safe table representation ('TS.DBTable')
dBtable2dBTableTS :: MonadPlus m => DB.DBtable -> m TS.DBTable
dBtable2dBTableTS (name,(x,y),vs) = do
	let (a,b) = unzip vs
	ax <- aux x a
	by <- aux y b
	return $ mergeAtts ax by
	where
	aux :: MonadPlus m => [DB.DBattribute] -> [[DB.DBvalue]] -> m DBValueList
	aux x y = aux1 y $ dBattributes2dBType x
		
	aux1 :: MonadPlus m => [[DBvalue]] -> DBType -> m DBValueList
	aux1 l (DBType t) = mapM (dBvalues2valueTS t) l >>= return . DBValueList t
	
	mergeAtts :: DBValueList -> DBValueList -> DBTable
	mergeAtts (DBValueList t1 vs1) (DBValueList t2 vs2) = DBTable $ DBtable name t1 t2 $ Map.fromDistinctAscList $ zip vs1 vs2

-- | Converts a list of non-type-safe values, given a product of type-safe variables, to a pair of type-safe values.
dBvalues2valueTS :: MonadPlus m => TS.DBtype a -> [DBvalue] -> m a
dBvalues2valueTS (TS.Prod a b) (x:xs) = do
	l <- dBvalue2valueTS x a
	r <- dBvalues2valueTS b xs
	return (l,r)
dBvalues2valueTS t [x] = dBvalue2valueTS x t
dBvalues2valueTS One [] = return ()
dBvalues2valueTS x  y = fail $ "DBase to DBase type-safe conversion error: " ++ show x ++ "\n\n" ++ show y 

-- | Converts a non-type-safe value, given the type-safe variable for that value, to the variable's matching value.
dBvalue2valueTS :: MonadPlus m => DBvalue -> TS.DBtype a -> m a
dBvalue2valueTS (VARCHAR_T s) (VarChar _ l)
	| length s <= l = return s
	| otherwise = fail $ "DBase to DBase type-safe conversion error: string value '" ++ show s ++ "(" ++ show (length s) ++ ")' larger than maximum size " ++ show l ++ "\n"
dBvalue2valueTS (INTEGER_T i) (Integer _) = return i
dBvalue2valueTS (NUMERIC_T d) (Numeric _ _ _) = return d
dBvalue2valueTS (BOOLEAN_T b) (Boolean _) = return b
dBvalue2valueTS (FLOAT_T f) (Float _) = return f
dBvalue2valueTS NULL (Null t) = return $ Right ()
dBvalue2valueTS v (Null t) = do
	l <- dBvalue2valueTS v t
	return $ Left l
dBvalue2valueTS v (FK _ _ _ t) = dBvalue2valueTS v t
dBvalue2valueTS x y = fail $ "DBase to DBase type-safe conversion error: can't match value '" ++ show x ++ "' with type-safe representation '" ++ show y ++ "'\n"

-- Converts a list of non-type-safe table variables to a product of type-safe table variables.
dBattributes2dBType :: [DB.DBattribute] -> DBType
dBattributes2dBType [] = DBType TS.One
dBattributes2dBType [x] = dBattribute2dBType x
dBattributes2dBType (x:xs) = aux (dBattribute2dBType x) (dBattributes2dBType xs)
	where
	aux :: DBType -> DBType -> DBType
	aux (DBType x) (DBType y) = DBType $ TS.Prod x y

-- | Converts a non-type-safe variable representation ('DB.DBattribute') to a variable type-safe representation ('DBType' encapsulated, without products)
dBattribute2dBType :: DB.DBattribute -> DBType
dBattribute2dBType (_,name,t,isnull,def,fk) = mkFk fk $ nillable isnull $ hasDefault def
	where
	mkFk :: Maybe ForeignKey -> TS.DBType -> TS.DBType
	mkFk Nothing = id
	mkFk (Just (n,tn,cn)) = applyDBType (DBType . FK n tn cn)
	
	nillable :: Bool -> TS.DBType -> TS.DBType
	nillable True = applyDBType (DBType . Null)
	nillable False = id
	
	hasDefault :: Maybe DBvalue -> TS.DBType
	hasDefault Nothing = dBtype2dBTypeTS name t
	hasDefault (Just def) = attMatch name t def

-- | Converts a non-type-safe 'DB.DBtype' to an encapsulated type-safe 'TS.DBtype' without a default value.
dBtype2dBTypeTS :: String -> DB.DBtype -> TS.DBType
dBtype2dBTypeTS n (VARCHAR i) = DBType $ VarChar (n,Nothing) i
dBtype2dBTypeTS n (NUMERIC x y) = DBType $ Numeric (n,Nothing) x y
dBtype2dBTypeTS n INTEGER = DBType $ Integer (n,Nothing)
dBtype2dBTypeTS n BOOLEAN = DBType $ Boolean (n,Nothing)
dBtype2dBTypeTS n FLOAT = DBType $ Float (n,Nothing)

-- | Converts a non-type-safe 'DBtype' to an encapsulated type-safe 'DBtype' with a default value.
attMatch :: String -> DB.DBtype -> DBvalue -> DBType
attMatch n (VARCHAR i) (VARCHAR_T s) = DBType $ VarChar (n,Just s) i
attMatch n (NUMERIC x y) (NUMERIC_T d) = DBType $ Numeric (n,Just d) x y
attMatch n INTEGER (INTEGER_T i) = DBType $ Integer (n,Just i)
attMatch n BOOLEAN (BOOLEAN_T b) = DBType $ Boolean (n,Just b)
attMatch n FLOAT (FLOAT_T f) = DBType $ Float (n,Just f)


	