
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     Misc functions to deal with SQL DBase and Type representations.
-- 
-- by
--
--     Diogo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Data.SQL.SQLexts where

import Control.Monad.State
import Data.List as List
import Data.Map as Map
import Data.Maybe
import Pointless.Combinators
import Data.TwoLevelExchangeMisc
import Data.Transform.Type
import Data.Transform.RulesWithRefs

-- * RelationalDB class

-- | Verify if an Haskell data type represents a relational database
class (RelationalDB a) where
    isRelationalDB :: a -> Bool

instance RelationalDB (Type a) where
    isRelationalDB = isRelationalDB_Type

-- | Checks if a 2LT Type structure given is a relation database.
-- a relational database type must contain tags in maps (for tables) and in basic types (for variables)
isRelationalDB_Type :: Type a -> Bool
-- isRelationalDB_Type (Prod a b) = isRelationalDB_Type a && isRelationalDB_Type b
isRelationalDB_Type (Ann l (Prod a b)) = (isRelationalDB_Type a && isRelationalDB_Type b) && l == emptyTag
isRelationalDB_Type (Ann l (Map One b)) = isRelationalDB_type' b && maybe False (const True) (getFieldName l)
isRelationalDB_Type (Ann l (Map a One)) = isRelationalDB_type' a  && maybe False (const True) (getFieldName l)
isRelationalDB_Type (Ann l (Map a b)) = isRelationalDB_type' a && isRelationalDB_type' b  && maybe False (const True) (getFieldName l)
isRelationalDB_Type _ = False

-- | 'isRelationalDB_Type' auxiliary function.
-- It is for internal use only
isRelationalDB_type' :: Type a -> Bool
isRelationalDB_type' (Ann l Int) = maybe False (const True) (getFieldName l)
isRelationalDB_type' (Ann l Bool) = maybe False (const True) (getFieldName l)
isRelationalDB_type' (Ann l Char) = maybe False (const True) (getFieldName l)
isRelationalDB_type' (Ann l String) = maybe False (const True) (getFieldName l)
-- isRelationalDB_type' (Prod x y) = isRelationalDB_type' x && isRelationalDB_type' y
isRelationalDB_type' (Ann l (Prod x y)) = (isRelationalDB_type' x && isRelationalDB_type' y) && getFieldName l == Nothing
isRelationalDB_type' (Ann _ (Maybe a)) = isBasicType a
isRelationalDB_type' (Ann l (Either One y)) = isBasicType y  && maybe False (const True) (getFieldName l)	-- This should have been replaced with maybe
isRelationalDB_type' (Ann l (Either x One)) = isBasicType x && maybe False (const True) (getFieldName l)	-- This should have been replaced with maybe
isRelationalDB_type' One = True
isRelationalDB_type' (Ann _ One) = True
isRelationalDB_type' _ = False

-- * Normalizing Functions

-- | puts necessary relational database tags in a Type structure (doesn't change current tags)
-- a relational database type must contain tags in maps (for tables) and in basic types (for variables)
putTags :: Type a -> Type a
putTags a = evalState (putTagsST a) (1,1)
	where
	putTagsST :: Type a -> State (Int, Int) (Type a)
	putTagsST (Ann l (Map a b)) = do
					next <- get
				 	put (fst next + 1, snd next)
			                na <- putTagsST a
        	        		nb <- putTagsST b
			                return (Ann l (Map na nb))
	putTagsST (Map a b) = do
				next <- get
			 	put (fst next + 1, snd next)
		                na <- putTagsST a
                		nb <- putTagsST b
		                return (Ann (emptyTag {fieldName = return $ "Table" ++ ((show . fst) next)}) (Map na nb))
	putTagsST (Prod a b) = do
				na <- putTagsST a
				nb <- putTagsST b
				return (Prod na nb)
	putTagsST (Ann l a)
		| isBasicType a || isBasicEither a = do
							next <- get
			                		put (fst next, snd next + 1)
						        return (Ann l a)
		| otherwise     = do return a
	putTagsST a
		| isBasicType a || isBasicEither a = do
							next <- get
			                		put (fst next, snd next + 1)
						        return (Ann (emptyTag {fieldName = return $ "Column" ++ ((show . snd) next)}) a)
		| otherwise     = do return a
		            
	isBasicEither (Either One y) = isBasicType y
	isBasicEither (Either x One) = isBasicType x
	isBasicEither _ = False

-- | removes all tags from a Type structure
removeTags :: Type a -> Type a
removeTags (Map a b) = Map (removeTags a) (removeTags b)
removeTags (Prod a b) = Prod (removeTags a) (removeTags b)
removeTags (Either a b) = Either (removeTags a) (removeTags b)
removeTags (Ann _ t) = t
removeTags t = t

-- | Union of two maps, with deletion of all occurences of multiple keys.
myunion :: (Ord k) => Map k a -> Map k a -> Map k a
myunion = curry (uncurry Map.difference . (uncurry Map.union /\ uncurry Map.intersection))

-- | The data type to hold a map of indexes to their corresponding tables and columns.
type IndexMap = Map KeyId (String,String)

-- | Change the 'KeyId' of a 'Type' to a given one.
changeIndex :: Maybe KeyId -> Type a -> Type a
changeIndex idx (Ann l t) = Ann (changeIndex' idx l) t
	where
	changeIndex' :: Maybe KeyId -> TypeAnn -> TypeAnn
	changeIndex' idx tag = maybe tag (\x -> tag {key = Just x}) idx
changeIndex idx t = Ann (emptyTag {key = idx}) t

-- | Add a new depth value to a 'KeyId'.
incIndex :: Int -> Maybe KeyId -> Maybe KeyId
incIndex i = maybe Nothing (Just . (id >< (++[i])))

-- | Adds a list of references to a 'Type'.
addRefs :: RefType -> Type a -> Type a
addRefs ref (Ann l t) = Ann (addRefs' ref l) t
	where
	addRefs' :: RefType -> TypeAnn -> TypeAnn
	addRefs' ref tag = tag {reference = getReference tag ++ ref}
addRefs ref One = One

-- Add a new depth value to all references in a 'RefType'
incRefs :: Int -> RefType -> RefType
incRefs i = List.map (id >< fromJust . incIndex i . Just)

-- | Collects all 'Index' definitions in a 'Type'.
-- Map (Index name,order) (tablename,variablename)
collectIndexes :: Type a -> IndexMap
collectIndexes a = evalState (collectIndexesST a) ""
	where
	collectIndexesST :: Type a -> State String IndexMap
	collectIndexesST (Ann l (Map a b)) = maybe (return Map.empty) (\y -> do
		put y
		l <- collectIndexesST a
		r <- collectIndexesST b
		return $ myunion l r) $ getFieldName l
	collectIndexesST (Ann l (Prod a b)) = do
		let idx = getKey l
		    lidx = incIndex 0 idx
		    ridx = incIndex 1 idx
		-- removes current indexes in Prod childs, where no indexes should exist if an index exists in the Prod.
		l <- collectIndexesST $ changeIndex lidx a
		r <- collectIndexesST $ changeIndex ridx b
		return $ myunion l r
	collectIndexesST (Ann _ (Either a b)) = do
		l <- collectIndexesST a
		r <- collectIndexesST b
		return $ myunion l r
	collectIndexesST (Ann l _) = maybe (return Map.empty) (\(cn,key) -> do
		tn <- get
		return $ Map.singleton key (tn,cn)) $ mergeMb (getFieldName l) (key l)
	collectIndexesST One = return Map.empty
	collectIndexesST x = fail $ "Type to DBase type-safe conversion error: type '" ++ show x ++ "' has no tag."


