
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     2LT Type to DBase type-safe conversion
-- 
-- by
--
--     Diogo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Data.SQL.TypetoDBaseTS where

import Data.SQL.DBaseTS as DBaseTS
import Data.TwoLevelExchangeMisc as Misc hiding (trace)
import Data.Transform.Type as Type
import Data.Transform.RulesWithRefs as TLT
import Pointless.Combinators
import Data.Map as Map
import Data.SQL.SQLexts
import Data.List as List
import Control.Monad (MonadPlus,mzero,liftM)

-- | Converts a 'Type' into a 'DBase' type-safe representation.
-- Corrects existant tags if needed, by adding new ones and removes unproppriate ones.
-- Assumes a correct type declaration, the function doesn't accept incorrect patterns.
type2dBase :: MonadPlus m => Type a -> a -> m DBase
type2dBase t v = liftM DB $ type2dBase' (collectIndexes t) t v

-- | Converts a product of maps into a set of 'DBTable'.
type2dBase' :: MonadPlus m => IndexMap -> Type a -> a -> m [DBTable]
type2dBase' m (Ann emptyTag (Type.Prod a b)) (x,y) = do
	l <- type2dBase' m a x
	r <- type2dBase' m b y
	return $ l ++ r
type2dBase' m (Ann l t@(Map a b)) mvs = liftM return $ mb2mplus $ getFieldName l >>= \x -> map2dBTable m x t mvs

-- | Converts a 'Type' into a 'DBase' type-safe representation with empty values.
type2dBaseType :: MonadPlus m => Type a -> m DBase
type2dBaseType t = do
	if (isRelationalDB t)
		then liftM DB $ type2dBaseType' (collectIndexes t) t
		else mzero

-- | Converts a product of maps into a set of 'DBTable' with empty values.
type2dBaseType' :: MonadPlus m => IndexMap -> Type a -> m [DBTable]
type2dBaseType' m (Ann emptyTag (Type.Prod a b)) = do
	l <- type2dBaseType' m a
	r <-type2dBaseType' m b
	return $ l ++ r
type2dBaseType' m (Ann l t@(Map a b)) = liftM return $ mb2mplus $ getFieldName l >>= \x -> map2dBTable m x t Map.empty


-- | Converts a map 'Type' into an encapsulated 'DBtable'.
map2dBTable :: MonadPlus m => IndexMap -> String -> Type (Map a b) -> Map a b -> m DBTable
map2dBTable m n (Map a b) mv = do
	l <- type2dBtype m a $ Map.keys mv
	r <- type2dBtype m b $ Map.elems mv
	return $ aux l r
	where
	aux :: DBValueList -> DBValueList -> DBTable
	aux (DBValueList lt lvs) (DBValueList rt rvs) = DBTable $ DBtable n lt rt $ Map.fromDistinctAscList $ zip lvs rvs

-- | Converts a 'Type' and a list of values for that type to the similar encapsulated 'DBtype'.
type2dBtype :: MonadPlus m => IndexMap -> Type a -> [a] -> m DBValueList
type2dBtype m t@(Ann tag (Type.Prod a b)) = \x -> do
	let (z1,z2) = unzip x
	let ref = getReference tag
	    lref = incRefs 0 ref
	    rref = incRefs 1 ref
	l <- type2dBtype m (addRefs lref a) z1
	r <- type2dBtype m (addRefs rref b) z2
	return $ aux l r
	where
	aux :: DBValueList -> DBValueList -> DBValueList
	aux (DBValueList lt lvs) (DBValueList rt rvs) = DBValueList (DBaseTS.Prod lt rt) (zip lvs rvs)
	
type2dBtype m (Ann l String) = liftMbf (\x -> uncurry DBValueList . (processFks m (reference l) . stringType2dBtype x . len /\ id)) $ getFieldName l
	where
	len = (\x -> if x > 1000 then x else 1000) . maxInt . List.map length
	
type2dBtype m (Ann l Int) = liftMbf (\x -> DBValueList (processFks m (reference l) $ basicType2dBtype x Int)) $ getFieldName l

type2dBtype m (Ann l Bool) = liftMbf (\x -> DBValueList (processFks m (reference l) $ basicType2dBtype x Bool)) $ getFieldName l

type2dBtype m (Ann l Char) = liftMbf (\x -> DBValueList (processFks m (reference l) $ stringType2dBtype x 1) . List.map (:[])) $ getFieldName l

type2dBtype m (Ann l (Maybe t)) = type2dBtype m (Ann l (Either t Type.One)) . List.map (maybe (Right ()) Left)

type2dBtype m (Ann l (Either String Type.One)) = liftMbf (\x -> uncurry DBValueList . (processFks m (reference l) . Null . stringType2dBtype x . lenLeft /\ id)) $ getFieldName l
	where
	lenLeft = (\x -> if x > 1000 then x else 1000) . maxInt . List.map sizeLeft
	sizeLeft (Left x) = length x
	sizeLeft _ = 0

type2dBtype m (Ann l (Either t Type.One)) = liftMbf (\x -> DBValueList (processFks m (reference l) $ Null (basicType2dBtype x t))) $ getFieldName l

type2dBtype m (Ann l (Either Type.One String)) = liftMbf (\x -> uncurry DBValueList . (processFks m (reference l) . Null . stringType2dBtype x . lenRight /\ List.map (either Right Left))) $ getFieldName l
	where
	lenRight = (\x -> if x > 1000 then x else 1000) . maxInt . List.map sizeRight
	sizeRight (Right x) = length x
	sizeRight _ = 0
	
type2dBtype m (Ann l (Either Type.One t)) = liftMbf (\x -> DBValueList (processFks m (reference l) $ Null (basicType2dBtype x t)) . List.map (either Right Left)) $ getFieldName l

type2dBtype _ Type.One = return . DBValueList DBaseTS.One

type2dBtype _ t = error $ "tp2dberr: " ++ show t

-- | Converts a basic 'Type' into a 'DBtype' representation.
basicType2dBtype :: String -> Type a -> DBtype a
basicType2dBtype n Int = Integer (n,Nothing)
basicType2dBtype n Bool = Boolean (n,Nothing)

-- | Converts a String basic 'Type' into a 'DBtype' representation.
stringType2dBtype :: String -> Int -> DBtype String
stringType2dBtype n = VarChar (n,Nothing)

-- | Adds a list of foreign keys to a 'DBtype'.
processFks :: IndexMap -> RefType -> DBtype a -> DBtype a
processFks _ [] t = t
processFks m [(n,idx)] t = case (Map.lookup idx m) of
	Nothing  -> error $ "\nindex not found: " ++ show m ++ show n ++ " " ++ show idx
	-- index not found: probably the index was repeated due to duplicate foreign keys
	(Just (tn,cn)) -> FK n tn cn t
processFks m (fk:fks) t = processFks m fks $ processFks m [fk] t

