-----------------------------------------------------------------------------
-- |
-- Maintainer  :  pabloberdaguer@alfa.di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     Type-safe Two-level Data Transformation
--
-- by
--
--     Alcino Cunha, Jose Nuno Oliveira, Joost Visser
--
-----------------------------------------------------------------------------


module Data.Transform.CommandLine where

import Data.SQL.ASTPostgreSQL	hiding (Type,Schema,Database)	-- The AST for the PostgreSQL grammar
import Data.SQL.ASTToDBase		-- Conversion from AST to DBase
import Data.SQL.DBaseTS as TS hiding (One)		  -- Our database type-safe representation
import Data.SQL.DBase as DB		  -- Our database non-type-safe representation
import Data.SQL.DBaseTStoType		-- DBase to 2LT Type conversion
import Data.SQL.SQLexts			-- The link between DBase and 2LT
import Data.SQL.TypetoDBaseTS		-- 2LT Type to DBase conversion Non-exhaustive patterns in function basicType2dBtype


import Data.SQL.DBasetoDBaseTS		  -- DBase non-type-safe to DBase type-safe conversion

import Text.SQL.PrettyPrintDBaseTS	-- Pretty printing for DBase

--import Text.SQL.Lexer			-- A Parsec lexer for SQL
import Text.SQL.PostgreSQL		-- the PostgreSQL parser


import Text.XML.HaXml.Pretty		  -- Pretty printer for HaXml documents
import Data.XML.TypetoXML		  -- Type to XML conversion
import Text.XML.HaXmlAliases		  -- HaXml support
import Text.XML.HaXml.Types		  -- HaXml types definition
import Data.XML.XSDtoType		  -- Xsd to Type conversion
import Data.XML.XMLtoType		  -- XML to Type conversion
import Text.XML.XsdAliases		  -- Joost Visser's XSD parser support
import Language.XMLSchema.NameSpaces

import Data.Transform.Type
import Data.Transform.TwoLevelCombinators
import Data.Transform.RulesWithRefs hiding (when)
import System.Environment (getArgs, getProgName)		-- For options (PLB)
import System.Exit (exitWith, exitFailure, ExitCode(..))	-- For halts (PLB)
import System.Console.GetOpt
import System.IO
import Data.List as List
import Control.Monad.State
import System.IO.Unsafe
import Data.Transform.TwoLevelFunctions

-----------------------------------------------------------------------------
-- * Command line options

-- | Record type to hold all program options.
data Opt = Opt
	{
		  optInputFiles	:: [FilePath]
		, optOutputFile	:: Maybe FilePath
		, optVerbose	:: Bool
		, optQuiet	:: Bool
		, optMap	:: Bool
		, optEvolve	:: Bool
		, optBack	:: Bool
		, optXML	:: Maybe FilePath
		, optSQL	:: Maybe FilePath
	}


-- | Default options.
startOpt :: Opt
startOpt = Opt
	{ optInputFiles		= []
	, optOutputFile		= Nothing
	, optVerbose		= False
	, optQuiet		= False
	, optMap		= False
	, optEvolve		= False
	, optBack		= False
	, optXML		= Nothing
	, optSQL		= Nothing
	}


-- | Description of all available program options.
options :: [OptDescr (Opt -> IO Opt)]
options =
   [ Option "h" ["help"]
        (NoArg (\opt -> exitHelp))
        "Show usage info"
   , Option "i" ["input"]
        (ReqArg
           (\arg opt -> do
              return opt { optInputFiles = arg:(optInputFiles opt) })
           "FILES")
        "Input file names"
   , Option "o" ["output"]
        (ReqArg
           (\arg opt -> do
              return opt { optOutputFile = Just arg })
           "FILE")
        "Output file (default: stdout)"
   , Option "x" ["xsd"]
        (ReqArg (\arg opt -> do return opt { optXML = Just arg }) "XSD")
        "Input xml schema."
   , Option "s" ["sql"]
        (ReqArg (\arg opt -> do return opt { optSQL = Just arg }) "SQL")
        "Input sql schema."
   , Option "v" ["verbose"]
        (NoArg (\opt -> do return opt { optVerbose = True }))
        "Shows the intermediate Type structure."
   , Option "q" ["quiet"]
        (NoArg (\opt -> do return opt { optQuiet = True }))
        "No messages."
   , Option "m" ["map"]
        (NoArg (\opt -> do return opt { optMap = True }))
        "Perform X/R mapping."
   , Option "e" ["evolve"]
        (NoArg (\opt -> do return opt { optEvolve = True }))
        "Perform built in evolution."
   , Option "b" ["back"]
        (NoArg (\opt -> do return opt { optBack = True }))
        "Convert data backwards."
   ]

-- * Functions to parse and deal with program options

-- | Function that reads arguments from the environment and parses it according
-- with the rules specified in the 'options' type.
parseOptions :: IO Opt
parseOptions = do
   -- Reads the options to a triple
   (optsActions, rest, errs) <- getArgs >>= return . getOpt RequireOrder options

   -- If there are errors show them, show the usage and exit with failure code
   when (not (Prelude.null errs)) $ do
      mapM_ (hPutStrLn stderr) errs
      showHelp
      exitFailure

   -- If no options were specified then show the help and exit with failure
   -- code. The "rest" is used to save multiple files passed as options, thus
   -- it must be tested.
   when ((Prelude.null optsActions) && (Prelude.null rest)) $ do
      showHelp
      exitFailure

   -- create a record with all program options
   opts <- foldl (>>=) (return startOpt) optsActions

   -- Verify at exactly 1 schema
   when (optXML opts == Nothing && optSQL opts == Nothing) $ do
      exitErrorHelp "Exactly one schema is required."

   when (optXML opts /= Nothing && optSQL opts /= Nothing) $ do
      exitErrorHelp "Exactly one schema is required."


   -- verify if at least one input argument exists. This can be specified both
   -- with "i" option or just passing the names of the files

   when (optSQL opts == Nothing && (Prelude.null rest) && (Prelude.null (optInputFiles opts))) $ do
      exitErrorHelp "An input file is required."


   -- return all the parsed options
   return opts { optInputFiles = (optInputFiles opts) ++ rest }


-----------------------------------------------------
-- | Transformations

-- Applies the given transformation with whatever is specified in the command line options
transformWithExternalOptions :: Rule -> Rule -> IO ()
transformWithExternalOptions transformation notrensformation = do
	opts <- parseOptions
	when (not (optQuiet opts)) $ do
		putStrLn "*** Parameters ***"
		putStrLn $ "Input files:   "++(show $ optInputFiles opts)
		putStrLn $ "Output files:   "++(show $ optOutputFile opts)
		putStrLn $ "XML Schema file:   "++(show $ optXML opts)
		putStrLn $ "SQL Schema file:   "++(show $ optSQL opts)
		putStrLn $ "Verbose:       "++(show $ optVerbose opts)
		putStrLn $ "Map:           "++(show $ optMap opts)
		putStrLn $ "Evolve:        "++(show $ optEvolve opts)
		putStrLn $ "Back:          "++(show $ optBack opts)
		putStrLn $ "Quiet:         "++(show $ optQuiet opts)
		putStrLn "\n"
	
	let l = optInputFiles opts
	let t = if (optEvolve opts) then transformation else notrensformation
	case (optXML opts,optSQL opts, optMap opts) of
 		(Just xsd, Nothing, True) -> xsd2sql t xsd l (optBack opts) (optOutputFile opts) (optVerbose opts)
		(Just xsd, Nothing, False) -> xsd2xsd t xsd l (optBack opts) (optOutputFile opts) (optVerbose opts)
		(Nothing, Just sql, _) -> sql2sql t sql l (optBack opts) (optOutputFile opts) (optVerbose opts)
                _ -> putStrLn "OOPS: Unknown Combination of options!!"
	
	when (not (optQuiet opts)) $ do
		putStrLn "*** Success! ***"
	

xsd2xsd :: Rule -> FilePath -> [FilePath] -> Bool -> Maybe FilePath -> Bool -> IO ()
xsd2xsd transformation xsd xmls back output verbose = do
	if (back)
		then do
			putStrLn "Starting to parse!!"
			(DynT t) <- xsd2type xsd
			putStrLn "Parsing done!!"
			transformBckwdXML t transformation xmls output verbose
		else do
			putStrLn "Starting to parse!!"
			(Dyn t v) <- xml2type xsd xmls
			putStrLn "Parsing done!!"
			transfomFwdXML t v transformation output verbose

-- Applies the given transformation to the DB in the list of files
sql2sql :: Rule -> FilePath -> [FilePath] -> Bool -> Maybe FilePath -> Bool -> IO ()
sql2sql transformation sql l back output verbose = do
	(Dyn t v) <- sql2type (sql:l) (DB.DB [])
	if (back) 
		then do
			(DynT t) <- sql2dbase [sql] (DB.DB []) >>= dBase2dyntype
			transformBckwdSQL t transformation l output verbose
		else do
			(Dyn t v) <- sql2type (sql:l) (DB.DB [])
			transformFwd t v transformation output verbose

xsd2sql :: Rule -> FilePath -> [FilePath] -> Bool -> Maybe FilePath -> Bool -> IO ()
xsd2sql transformation xsd l back output verbose = do
	if (back)
		then do
			putStrLn "Starting to parse!!"
			(DynT t) <- xsd2type xsd
-- 			let t = (Tag TagType {fieldName = Just "imdb", key = Nothing, reference = []} (Data.Transform.RulesWithRefs.Prod (Tag TagType {fieldName = Just "Map$movie", key = Nothing, reference = []} (Data.Transform.RulesWithRefs.Map (Data.Transform.RulesWithRefs.Prod (Tag TagType {fieldName = Just "Key$title", key = Just (3,[]), reference = []} String) (Tag TagType {fieldName = Just "Key$year", key = Just (4,[]), reference = []} Int)) (Data.Transform.RulesWithRefs.Prod (Data.Transform.RulesWithRefs.List (Tag TagType {fieldName = Just "review", key = Nothing, reference = []} String)) (Data.Transform.RulesWithRefs.Prod (Tag TagType {fieldName = Just "director", key = Nothing, reference = []} String) (Data.Transform.RulesWithRefs.List (Tag TagType {fieldName = Just "box_office", key = Nothing, reference = []} (Data.Transform.RulesWithRefs.Prod (Tag TagType {fieldName = Just "country", key = Nothing, reference = []} String) (Tag TagType {fieldName = Just "value", key = Nothing, reference = []} Int)))))))) (Data.Transform.RulesWithRefs.List (Tag TagType {fieldName = Just "actor", key = Nothing, reference = []} (Data.Transform.RulesWithRefs.Prod (Tag TagType {fieldName = Just "name", key = Nothing, reference = []} String) (Data.Transform.RulesWithRefs.List (Tag TagType {fieldName = Just "played", key = Nothing, reference = []} (Data.Transform.RulesWithRefs.Prod (Tag TagType {fieldName = Just "title", key = Nothing, reference = [(Just "FKtitle",(3,[]))]} String) (Data.Transform.RulesWithRefs.Prod (Tag TagType {fieldName = Just "year", key = Nothing, reference = [(Just "FKyear",(4,[]))]} Int) (Data.Transform.RulesWithRefs.Prod (Tag TagType {fieldName = Just "role", key = Nothing, reference = []} String) (Data.Transform.RulesWithRefs.List (Tag TagType {fieldName = Just "award", key = Nothing, reference = []} (Tag TagType {fieldName = Just "award_name", key = Nothing, reference = []} String)))))))))))))
			putStrLn "Parsing done!!"
			transformBckwd t transformation l output verbose
		else do
			putStrLn "Starting to parse!!"
 			(Dyn t v) <- xml2type xsd l
			putStrLn "Parsing done!!"
			transformFwd t v transformation output verbose


transfomFwdXML :: Type a -> a		-- original XML Schema converted to Type
	    -> Rule			-- XML evolution rule
	    -> Maybe FilePath		-- XSD and XML output file
	    -> Bool			-- verbosity
	    -> IO ()
	    
transfomFwdXML t v transformation outFileName verbose = do
	let (Just evolutionView) = evalStateT (transformation t) (initNameSupply t)
	outputHandle <- maybe (return stdout) (\f -> openFile f WriteMode) outFileName

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Source):\n# " ++ (show t)

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Target):\n# " ++ showType evolutionView

	let (Just xmlEvolve) = forthDyn evolutionView v
	xmlout <- applyDyn guaranteeXSD xmlEvolve
	
	xsdStr <- applyDyn (curry (type2Xsd . fst)) xmlout
	hPutStrLn outputHandle $ "\n\n#Target XSD output:\n" ++ (show $ document $ xsdStr)
	
	xmlStr <- applyDyn type2Xml xmlout
	hPutStrLn outputHandle $ "\n\n#Target XML output:\n" ++ (show $ document $ xmlStr)
	maybe (return ()) (const $ hClose outputHandle) outFileName
	

transformBckwdXML :: Type a		-- original XML Schema converted to Type
	     	  -> Rule		-- XML evolution rule
	     	  -> [FilePath]		-- XML evoluted input files
	          -> Maybe FilePath	-- SQL output file
	          -> Bool		-- verbosity
	          -> IO ()
	          
transformBckwdXML t transformation xmls outFileName verbose = do
	let (Just evolutionView) = evalStateT (transformation t) (initNameSupply t)
	outputHandle <- maybe (return stdout) (\f -> openFile f WriteMode) outFileName

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Source):\n# " ++ (show t)

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Target):\n# " ++ showType evolutionView

	let (Just xmlEvolve) = forthDynT evolutionView
	xmlBack <- applyDynT ((flip fromXML) xmls) xmlEvolve
	let (Just xmlBackEvolve) = backDyn2 t evolutionView xmlBack
	
	when verbose $ hPutStrLn  outputHandle $ "\n\n#Intermediate Value representation (back - Source):\n" ++ applyDyn gshow xmlBackEvolve
	
	xmlStr <- applyDyn type2Xml xmlBackEvolve
	hPutStrLn outputHandle $ "\n\n#Target XML output:\n" ++ (show $ document $ xmlStr)
	maybe (return ()) (const $ hClose outputHandle) outFileName


transformBckwdSQL :: Type a		-- original SQL Schema converted to Type
	     -> Rule			-- XML evolution rule
	     -> [FilePath]		-- list of SQL evoluted files.
	     -> Maybe FilePath		-- SQL output file
	     -> Bool			-- verbosity
	     -> IO ()

transformBckwdSQL t transformation sqls outFileName verbose = do
	let (Just evolutionView) = evalStateT (transformation t) (initNameSupply t)
	outputHandle <- maybe (return stdout) (\f -> openFile f WriteMode) outFileName

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Source):\n# " ++ (show t)

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Target):\n# " ++ showType evolutionView

	let (Just sqlEvolve) = forthDynT evolutionView
	sqlBack <- applyDynT ((flip fromRelationalDB) sqls) sqlEvolve
	let (Just sqlBackEvolve) = backDyn2 t evolutionView sqlBack
	x <- toDBase sqlBackEvolve
		
	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Database representation (Target):\n" ++ (show x)
	hPutStrLn outputHandle $ "\n\n#Target Database SQL:\n" ++ (Text.SQL.PrettyPrintDBaseTS.sshow x)

	maybe (return ()) (const $ hClose outputHandle) outFileName


transformFwd :: Type a -> a		-- original XML/SQL Schema converted to Type
	     -> Rule			-- XML evolution rule
	     -> Maybe FilePath		-- SQL output file
	     -> Bool			-- verbosity
	     -> IO ()

transformFwd t v transformation outFileName verbose = do
	let (Just evolutionView) = evalStateT (transformation t) (initNameSupply t)
	outputHandle <- maybe (return stdout) (\f -> openFile f WriteMode) outFileName

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Source):\n# " ++ (show t)

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Target):\n# " ++ showType evolutionView

	let (Just xmlEvolve) = forthDyn evolutionView v
	x <- toSQL xmlEvolve
	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Database representation (Target):\n" ++ (show x)
	hPutStrLn outputHandle $ "\n\n#Target Database SQL:\n" ++ (Text.SQL.PrettyPrintDBaseTS.sshow x)

	maybe (return ()) (const $ hClose outputHandle) outFileName


transformBckwd :: Type a   		-- original XML Schema converted to Type
	       -> Rule			-- XML evolution rule
	       -> [FilePath]    	-- SQL input files
	       -> Maybe FilePath	-- XML output file
	       -> Bool          	-- verbosity
	       -> IO ()

transformBckwd t transformation sqls outFileName verbose = do
	let (Just evolutionView) = evalStateT (transformation t) (initNameSupply t)
	outputHandle <- maybe (return stdout) (\f -> openFile f WriteMode) outFileName
	when verbose $ do
		hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Source):\n# " ++ (show t)

	when verbose $ hPutStrLn outputHandle $ "\n\n#Intermediate Type representation (Target):\n# " ++ showType evolutionView
	
	let (Just xmlEvolve) = forthDynT evolutionView
	xmlBack <- applyDynT ((flip fromRelationalDB) sqls) xmlEvolve
	
	when verbose $ hPutStrLn  outputHandle $ "\n\n#Intermediate Value representation (back - Target):\n" ++ applyDyn gshow xmlBack
	
	let (Just xmlBackEvolve) = backDyn2 t evolutionView xmlBack
	
	when verbose $ hPutStrLn  outputHandle $ "\n\n#Intermediate Value representation (back - Source):\n" ++ applyDyn gshow xmlBackEvolve

	xmlStr <- applyDyn type2Xml xmlBackEvolve
	hPutStrLn outputHandle $ "\n\n#Target XML output:\n" ++ (show $ document $ xmlStr)
	maybe (return ()) (const $ hClose outputHandle) outFileName

---------------------------------------------
-- * SQL Interface

-- | Parse SQL.
parseAst :: FilePath -> IO ASTPostgreSQL
parseAst file = do
	x <- readFile file
	let y = astPostgreSQL x
	return y

-- | Fills values from multiple SQL files into a database intermediate type-safe representation.
sqlfromtype :: forall a. [FilePath] -> DB.DBase -> Type a -> IO Dynamic
sqlfromtype files db t = do
	x <- sql2dbase files db
	aux x t
	where
	aux dbts t = do
		y <- dBase2type dbts t
		return $ Dyn t y	

-- | Converts multiple SQL files into a database intermediate type-safe representation.
sql2type :: [FilePath] -> DB.DBase -> IO Dynamic
sql2type files db = sql2dbase files db >>= dBase2dyn

-- | Converts multiple SQL files into an intermediate database type-safe representation.
sql2dbase :: [FilePath] -> DB.DBase -> IO TS.DBase
sql2dbase fs db = sql2dbase' fs db >>= dBase2dBaseTS

-- | Converts multiple SQL files into an intermideate database non-type-safe representation.
sql2dbase' :: [FilePath] -> DB.DBase -> IO DB.DBase
sql2dbase' [] db = return db
sql2dbase' (f:fs) db = do
			x <- parseAst f
			y <- addAstToDBase x db
			sql2dbase' fs y

xml2type :: FilePath -> [FilePath] -> IO Dynamic
xml2type xsd xmls = do	fxmls <- mapM xmlparse xmls
			xsdt <- xsd2type xsd
			applyDynT ((flip xmls2Type) fxmls) xsdt

xmlparse :: FilePath -> IO Document
xmlparse xml = do
		fx <- readFile xml
		case (parseXml "error.log" fx) of
			(Left err) -> error err
			(Right x) -> return x
			
-- | Parses an input XML schema file into Joost Visser's XSD Abstract Syntax Tree.
xsdparse :: FilePath -> IO Language.XMLSchema.NameSpaces.Schema
xsdparse xsd = do
	x <- parseXsd xsd
	x

-- | Parser an input XML schema file into a type-safe representation.
xsd2type :: FilePath -> IO DynType
xsd2type xsd = do	fxsd <- xsdparse xsd
			fns <- mapM xsdparse $ getSchemaLocations fxsd
			xsd2Type (fxsd:fns) fxsd
		
------------------------------------------------------
-- | Handling errors.
panickIfNotRelational :: Type a -> IO ()
panickIfNotRelational t
  | isRelationalDB t = putStr ""
  | otherwise = exitError $ "\n\nERROR! Transformation to relational was NOT successful!\nI don't know how to handle this:\n" ++ (whatIsntRelationalDB_Type t) ++ "\nIn:\n" ++ (whereIsntRelationalDB_Type t)

-------------------------------------------------------
-- | Conversion to relational structure.

makeRelationalDB :: Dynamic -> IO Dynamic
makeRelationalDB d@(Dyn t v)
  | isRelationalDB t
  = return d
  | otherwise
  = do
	let Just tt = evalStateT (normalizeTag t) (initNameSupply t)
-- 	putStrLn ("\nBefore toRDB:\n"++ (showType tt))
	let (Just vw) = toRDB t	-- Tag safe (almost)
-- 	putStrLn ("\nAfter toRDB:\n"++showType vw)
	let (Just d') = forthDyn vw v
	applyDyn (\t _ -> panickIfNotRelational t) d'
	return d'

toSQL :: Dynamic -> IO TS.DBase
toSQL d = do
 	d'<- makeRelationalDB d
  	let (Just db) = applyDyn type2dBase d'
  	return db

toDBase :: Dynamic -> IO TS.DBase
toDBase d = do
	let (Just db) = applyDyn type2dBase d
	return db

-- | Conversion from XML structure to relational structure and back to a XML structure.
-- XML => SQL => XML
fromRelationalDB :: Type a -> [FilePath] -> IO Dynamic
fromRelationalDB t sqls = do
  	let (Just tt) = evalStateT (normalizeTag t) (initNameSupply t)
-- 	putStrLn $ "\nBefore toRDB:\n"++ showType tt
	let (Just vw) = toRDB t	-- Tag safe (almost)
-- 	putStrLn $ "\nAfter toRDB:\n" ++ showType vw
	let (Just d') = forthDynT vw
	applyDynT panickIfNotRelational d'
	-- the SQL RDB type for the matching of the files to the transformed type.
	rdb <- applyDynT (sqlfromtype sqls (DB.DB [])) d'
	putStrLn $ "\nFill with input SQL values:\n" ++ applyDyn typePrint rdb

	-- the XML type with the converted values
	let (Just xmlBack) = backDyn2 t vw rdb
	
	return xmlBack

-- | Fills a XML structure with values from XML files.
fromXML :: Type a -> [FilePath] -> IO Dynamic
fromXML t xmls = do
	fxmls <- mapM xmlparse xmls
	dyn <- xmls2Type t fxmls
	putStrLn $ "\nFill Evoluted type with input XML values:\n" ++ applyDyn typePrint dyn
	return dyn

-- | Conversion to a XML structure.
guaranteeXSD :: Type a -> a -> IO Dynamic
guaranteeXSD t v = do
-- 	putStrLn $ "\nBefore toXSD:\n" ++ show t
	let (Just view) = toXML t
-- 	putStrLn $ "\nAfter toXSD:\n" ++ showType view
	let (Just out) = forthDyn view v
	return out

-- * Help printing functions

-- | Prints a Type separed by type definition and value instances.
typePrint :: Type a -> a -> String
typePrint t v = "\tType Definition:\n" ++ (show t) ++ "\n\n\tValues: \n" ++ (gshow t v)	  

-- | Function that prints the help usage and returns with a success code (used
--   when the help program option is specified.
exitHelp :: IO a
exitHelp = do
    showHelp
    exitWith ExitSuccess


-- | Function that prints the program usage to the sdtderr using the standard
--   'usageInfo' function.
showHelp :: IO ()
showHelp = do
    prg <- getProgName
    let header = unlines [prg++" -- Demo program for two-level XML/SQL transformations","",
                          "Usage: "++prg++" [OPTION...]"]
    hPutStrLn stderr (usageInfo header options)
    hFlush stderr


-- | Helper function that prints a message to stderr and exits with a failure
--   code error
exitErrorHelp :: String -> IO ()
exitErrorHelp msg = do
   hPutStrLn stderr ("Error: " ++ msg)
   hPutStrLn stderr ""
   showHelp
   exitFailure


-- | Function that prints a message to stderr and exits with a failure
--   code error
exitError :: String -> IO ()
exitError msg = do
   hPutStrLn stderr ("Error: " ++ msg)
   hPutStrLn stderr ""
   exitFailure

-- nasty unsafePerformIO !!!!!
schema2type :: Schema -> Maybe DynType
schema2type schema = unsafePerformIO (do {fns <- mapM xsdparse $ getSchemaLocations schema; return $ xsd2Type (schema:fns) schema})

{-class FrontEnd t v | t -> v, v -> t where
	parsetype  :: XSD -> Maybe DynType
	printtype  :: Type a -> Maybe XSD
	parsevalue :: Type a -> XML -> Maybe a
	printvalue :: Type a -> a -> Maybe XML
-}

legrandefinale :: Schema -> Rule -> Maybe (DDL,DDL,DML -> Maybe DML, DML -> Maybe DML)
legrandefinale xsd evolution = do
	(sqlv1,xmlv1tosqlv1,sqlv1toxmlv1) <- transform xsd (toRDB')
	(sqlv2,xmlv1tosqlv2,sqlv2toxmlv1) <- transform xsd (evolution >>> toRDB')
	return (sqlv1,sqlv2, \x -> sqlv1toxmlv1 x >>= xmlv1tosqlv2 , \x -> sqlv2toxmlv1 x >>= xmlv1tosqlv1)

instance FrontEnd Schema Document where
	parsetype = schema2type
	printtype t = do
		doc <- type2Xsd t
		doc2Xsd "" doc
	parsevalue = flip xml2Values
	printvalue = type2Xml
	
instance FrontEnd DDL DML where
	parsetype = dBase2dyntype
	printtype = type2dBaseType
	parsevalue = flip dBase2type
	printvalue = type2dBase

