{-# OPTIONS -fglasgow-exts #-}

-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Point-free rewrite rules
--
-----------------------------------------------------------------------------

module Data.Transform.RulesPF where

import Data.Transform.Type hiding (everything)
import Data.Transform.Rewriting
import Control.Monad

infixr 5 /\
(/\) ::  PF (a -> b) -> PF (a -> c) -> PF (a -> (b,c))
(/\) = SPLIT

-- Predicates for point-free rewriting

projected :: Query Bool
projected (PF t :| x) = (everything (||) aux (PF t :| reduce (many (once (top comp_assocr))) t x))
    where aux :: GenericQ Bool
          aux (PF _) (COMP _ FST HOLE) = True
          aux (PF _) (COMP _ SND HOLE) = True
          aux (PF _) (COMP _ FST (COMP _ HOLE _)) = True
          aux (PF _) (COMP _ SND (COMP _ HOLE _)) = True
          aux _ _ = False

injected :: Query Bool
injected (PF t :| x) = (everything (||) aux (PF t :| reduce (many (once (top comp_assocl))) t x))
    where aux :: GenericQ Bool
          aux (PF _) (COMP _ HOLE INL) = True
          aux (PF _) (COMP _ HOLE INR) = True
          aux (PF _) (COMP _ (COMP _ _ HOLE) INL) = True
          aux (PF _) (COMP _ (COMP _ _ HOLE) INR) = True
          aux _ _ = False

applied :: Query Bool
applied (PF t :| x) = (everything (||) aux (PF t :| reduce (many (once (top comp_assocr))) t x))
    where aux :: GenericQ Bool
          aux (PF _) (COMP _ AP (SPLIT HOLE _)) = True
          aux (PF _) (COMP _ AP (SPLIT (COMP _ HOLE _) _)) = True
          aux _ _ = False

-- Rules for point-free rewriting

comp_assocr :: Rule
comp_assocr _ (COMP a (COMP b f g) h) = return (COMP b f (COMP a g h))
comp_assocr _ _ = fail "Cannot apply Comp-Assoc!"

comp_assocl :: Rule
comp_assocl _ (COMP a f (COMP b g h)) = return (COMP b (COMP a f g) h)
comp_assocl _ _ = fail "Cannot apply Comp-Assoc!"

nat_id :: Rule
nat_id _ (COMP _ ID f) = return f
nat_id _ (COMP _ f ID) = return f
nat_id _ _ = fail "Cannot apply Nat_Id!"

bang_eta :: Rule
bang_eta _ BANG = mzero
bang_eta (Func _ One) _ = return BANG
bang_eta _ _ = mzero

bang_reflex :: Rule
bang_reflex (Func One One) BANG = success "Bang-Reflex" ID
bang_reflex _ _ = mzero

bang_fusion :: Rule
bang_fusion _ (COMP _ BANG _) = success "Bang-Fusion" BANG
bang_fusion _ _ = mzero

eta_prod :: Rule
eta_prod _ (SPLIT _ _) = fail "Cannot apply Eta_Prod to Split!"
eta_prod (Func (Either _ _) (Prod _ _)) _ = fail "Priority to Sums!"
eta_prod (Func _ a@(Prod _ _)) f = 
    do c <- context
       guard (not (projected c))
       success "Eta-Prod" ((COMP a FST f) /\ (COMP a SND f))
eta_prod _ _ = fail "Cannot apply Eta_Prod!"

prod_cancel :: Rule
prod_cancel _ (COMP _ FST (SPLIT f g)) = success "Prod-Cancel" f
prod_cancel _ (COMP _ SND (SPLIT f g)) = success "Prod-Cancel" g
prod_cancel _ (COMP _ FST (COMP a (SPLIT f g) h)) = success "Prod-Cancel" (COMP a f h)
prod_cancel _ (COMP _ SND (COMP a (SPLIT f g) h)) = success "Prod-Cancel" (COMP a g h)
prod_cancel _ _ = fail "Cannot apply Prod_Cancel!"

prod_reflex :: Rule
prod_reflex _ (SPLIT FST SND) = success "Prod-Reflex" ID
prod_reflex _ (PROD ID ID) = success "Prod-Reflex" ID
prod_reflex _ _ = fail "Cannot apply Prod_Reflex!"

prod_fusion :: Rule
prod_fusion _ (COMP a (SPLIT f g) h) = success "Prod-Fusion" ((COMP a f h) /\ (COMP a g h))
prod_fusion _ _ = fail "Cannot apply Prod_Fusion!"

prod_fusion_inv :: Rule
prod_fusion_inv (Func a d) (SPLIT (COMP b f h1) (COMP c g h2)) =
    do Eq <- teq b c
       guard (geq (PF (Func a b)) h1 h2)
       success "Prod-Fusion-Inv" (COMP b (SPLIT f g) h1)
prod_fusion_inv _ _ = fail "Cannot apply Prod-Fusion-Inv!"

prod_eta :: Rule
prod_eta a (SPLIT (COMP b FST f) (COMP c SND g)) =
    do Eq <- teq b c
       guard (geq (PF a) f g)
       success "Prod-Eta" f
prod_eta _ _ = fail "Cannot apply Eta-Prod!"

prod_def :: Rule
prod_def (Func (Prod a b) _) (PROD f g) = success "Prod-Def" ((COMP a f FST) /\ (COMP b g SND))
prod_def (Func (Tag s (Prod a b)) _) (PROD f g) = success "Prod-Def" ((COMP a f FST) /\ (COMP b g SND))
prod_def _ _ = fail "Cannot apply Prod-Def!"

prod_def_inv :: Rule
prod_def_inv _ (SPLIT (COMP _ f FST) (COMP _ g SND)) = success "Prod-Def-Inv" (PROD f g)
prod_def_inv _ (SPLIT (COMP _ f FST) SND) = success "Prod-Def-Inv" (PROD f ID)
prod_def_inv _ (SPLIT FST (COMP _ g SND)) = success "Prod-Def-Inv" (PROD ID g)
prod_def_inv _ _ = mzero

eta_sum :: Rule
eta_sum _ (EITHER _ _) = fail "Cannot apply Eta_Sum to Either!"
eta_sum (Func a@(Either _ _) _) f = 
    do c <- context
       guard (not (injected c))
       success "Eta-Sum" (EITHER (COMP a f INL) (COMP a f INR))
eta_sum _ _ = fail "Cannot apply Eta-Sum!"

sum_cancel :: Rule
sum_cancel _ (COMP _ (EITHER f g) INL) = success "Sum-Cancel" f
sum_cancel _ (COMP _ (EITHER f g) INR) = success "Sum-Cancel" g
sum_cancel _ (COMP _ (COMP a h (EITHER f g)) INL) = success "Sum-Cancel" (COMP a h f)
sum_cancel _ (COMP _ (COMP a h (EITHER f g)) INR) = success "Sum-Cancel" (COMP a h g)
sum_cancel _ _ = fail "Cannot apply Sum-Cancel!"

sum_def :: Rule
sum_def (Func _ (Either a b)) (SUM f g) = success "Sum-Def" (EITHER (COMP a INL f) (COMP b INR g))
sum_def _ _ = fail "Cannot apply Sum-Def!"

eta_exp :: Rule
eta_exp _ (CURRY _) = fail "Cannot apply Eta-Exp to Curry!"
eta_exp (Func a (Func b c)) f =
    do x <- context
       guard (not (applied x))
       success "Eta-Exp" (CURRY (COMP (Prod (Func b c) b) AP (COMP a f FST /\ SND)))
eta_exp _ _ = fail "Cannot apply Eta-Exp!"


exp_cancel :: Rule
exp_cancel (Func a _) (COMP (Prod _ b) AP (SPLIT (CURRY f) h)) = success "Exp-Cancel" (COMP (Prod a b) f (ID /\ h))
exp_cancel _ (COMP (Prod _ b) AP (SPLIT (COMP a (CURRY f) g) h)) = success "Exp-Cancel" (COMP (Prod a b) f (g /\ h))
exp_cancel _ _ = fail "Cannot apply Exp-Cancel!"

exp_def :: Rule
exp_def (Func (Func _ a) _) (EXP f) = success "Exp-Def" (CURRY (COMP a f AP))
exp_def _ _ = fail "Cannot apply Exp-Def!"

listmap_fusion :: Rule
listmap_fusion _ (COMP (List t) (LIST f) (LIST g)) = success "Listmap-Fusion_1" (LIST (COMP t f g))
listmap_fusion _ (COMP (List t) (LIST f) (COMP t' (LIST g) h)) = success "Listmap-Fusion_1" (COMP t' (LIST (COMP t f g)) h)
listmap_fusion _ _ =  fail "Cannot apply Listmap-Fusion"

listmap_id :: Rule
listmap_id _ (LIST ID) = return ID
listmap_id _ _ = fail "Cannot apply Listmap-Id"

mapmap_fusion :: Rule
mapmap_fusion _ (COMP (Map k v) (MAP f) (COMP t (MAP g) h)) = success "Mapmap-Fusion" (COMP t (MAP (COMP v f g)) h)
mapmap_fusion _ (COMP (Map k v) (MAP f) (MAP g)) = success "Mapmap-Fusion" (MAP (COMP v f g))
mapmap_fusion _ _ =  fail "Cannot apply Mapmap-Fusion"

elems_fusion :: Rule
elems_fusion (Func (Map a b) (List c)) (COMP _ (LIST f) ELEMS) = success "Elems-Fusion" (COMP (Map a c) ELEMS (MAP f))
elems_fusion (Func _ (List c)) (COMP _ (LIST f) (COMP (Map a b) ELEMS h)) = success "Elems-Fusion" (COMP (Map a b) (COMP (Map a c) ELEMS (MAP f)) h)
elems_fusion _ _ = fail "Cannot apply Elems-Fusion"

elems_linv :: Rule
elems_linv _ (COMP _ ELEMS LIST2MAP)  = success "Elems-Linv" ID
elems_linv _ (COMP _ ELEMS (COMP _ LIST2MAP f)) = success "Elems-Linv" f
elems_linv _ _ = mzero

list2map_fusion :: Rule
list2map_fusion (Func _ (Map _ a)) (COMP _ (MAP f) LIST2MAP) = success "List2map-Fusion" (COMP (List a) LIST2MAP (LIST f))
list2map_fusion (Func _ (Map _ a)) (COMP _ (MAP f) (COMP b LIST2MAP g)) = success "List2map-Fusion" (COMP b (COMP (List a) LIST2MAP (LIST f)) g)
list2map_fusion _ _ = mzero

mapjoin_fusion :: Rule
mapjoin_fusion (Func _ (Map _ e)) (COMP (Map a (Prod b c)) (MAP f) (COMP _ LJOIN g)) = 
    do (COMP _ h FST) <- return $ reduce defuse (Func (Prod b c) e) (COMP e ID f)
       success "Mapjoin-Fusion" (COMP (Prod (Map a b) (Func a c)) (COMP (Map a b) (MAP h) FST) g)
mapjoin_fusion _ _ = mzero

defuse = many (once (top comp_assocl ||| top prod_fusion_inv))

mapmap_id :: Rule
mapmap_id _ (MAP ID) = return ID
mapmap_id _ _ = fail "Cannot apply Mapmap-Id"

cata_reflex :: Rule
cata_reflex _ (CATA (EITHER NIL CONS)) = success "Cata-Reflex" ID
cata_reflex _ _ = mzero

cata_nil :: Rule
cata_nil _ (CATA (EITHER NIL SND)) = success "Cata-Nil" (COMP One NIL BANG)
cata_nil _ _ = mzero

cata_cancel :: Rule
cata_cancel _ (COMP (List a) (CATA (EITHER f g)) NIL) = success "Cata-Cancel" f
cata_cancel _ (COMP (List a) (CATA (EITHER f g)) (COMP One NIL h)) = success "Cata-Cancel" (COMP One f h)
cata_cancel _ _ = mzero


removeTags :: Rule
removeTags _ (COMP t f g) = return (COMP (rmTags t) f g)
removeTags _ x = return x

-- Rewriting systems for point-free

optimize :: Rule
optimize = everywhere (top removeTags) >>> many (prods ||| maps ||| sums ||| catas)

prods :: Rule
prods = many1 (many (once (top comp_assocr)) >>> once aux)
    where aux :: Rule
	  aux = top nat_id ||| top prod_eta ||| top prod_cancel ||| top prod_reflex ||| top prod_fusion ||| top prod_def ||| top bang_reflex ||| top bang_fusion

sums :: Rule
sums = many1 (many (once (top comp_assocl)) >>> once aux)
    where aux :: Rule
	  aux = top nat_id ||| top sum_cancel ||| top sum_def

maps :: Rule
maps = many1 (many (once (top comp_assocr)) >>> once aux)
   where
     aux = top listmap_id ||| top listmap_fusion ||| top mapmap_fusion ||| top elems_fusion ||| top mapjoin_fusion ||| top list2map_fusion ||| top mapmap_id ||| top elems_linv ||| top bang_eta

catas :: Rule
catas = many1 (many (once (top comp_assocr)) >>> once aux)
    where aux = top cata_reflex ||| top cata_nil ||| top cata_cancel

prods_inv = innermost (top comp_assocl ||| top prod_fusion_inv ||| top prod_def_inv)

simplifypf a = reduce optimize a

-- Example

assocr :: PF (((Int,Char),Bool) -> (Int,(Char,Bool)))
assocr = SPLIT (COMP typeof FST FST) (PROD SND ID)

assocl :: PF ((Int,(Char,Bool)) -> ((Int,Char),Bool))
assocl = SPLIT (PROD ID FST) (COMP typeof SND SND)

associso = simplify True prods (COMP typeof assocr assocl)

