{-# OPTIONS -fglasgow-exts #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
--
-- SYB rewrite rules
--
-----------------------------------------------------------------------------

module Data.Transform.RulesSYB where

import Data.Transform.Type
import Data.Transform.Rewriting
import Data.Transform.RulesPF
import Control.Monad
import Debug.Trace

-----------------------------------------------------------------------------
-- * Type-preserving

-- The identity strategy is a zero for sequential composition.
id_seq :: Rule
id_seq _ (SEQ NOP f)  = success "id_seq" f
id_seq _ (SEQ f NOP)  = success "id_seq"  f
id_seq _ _            = mzero

-- The identity strategy cancels the all combinator.
id_all :: Rule
id_all _ (ALL NOP)    = success "id_all"  NOP
id_all _ _            = mzero


gmapT_fusion :: Rule
gmapT_fusion _ (SEQ (ALL f) (ALL g)) = success "seq_gmapT" $ ALL (SEQ f g)
gmapT_fusion _ _ = mzero

everywhere_def_inv :: Rule
everywhere_def_inv _ (SEQ f (ALL (EVERYWHERE g))) = 
    do guard $ geq (PF TP) f g
       success "everywhere_def_inv" $ EVERYWHERE f
everywhere_def_inv _ _ = mzero

-- Subsequent mkTs for the same type can be merged.
lift_mkT :: Rule
lift_mkT _ (SEQ (MKT a f) (MKT b g)) = do
  Eq <- teq a b
  return (MKT a (COMP a f g))
lift_mkT _ _ = mzero

-- Nested adhoc or mkT for the same type cancel against each other.
adhoc_cancel :: Rule
adhoc_cancel _ (ADHOC a f (ADHOC b g s)) = do
  Eq <- teq a b
  return (ADHOC a f s)
adhoc_cancel _ (ADHOC a f (MKT b g)) = do
  Eq <- teq a b
  return (MKT a f)
adhoc_cancel _ _ = mzero

-- Pushing down mkT

id_pullT :: Rule
id_pullT _ (MKT a ID) = success "id_pullT" NOP
id_pullT _ _ = mzero

map_pullT :: Rule
map_pullT _ (MKT (List a) (LIST f)) = success "map_pullT" $ ALL (MKT a f)
map_pullT _ _ = mzero

prod_pullT :: Rule
prod_pullT _ (MKT (Prod a b) (PROD f g)) =
      do Eq <- teq a b
	 guard (geq (PF (Func a a)) f g)
	 success "prod_pullT" $ ALL (MKT a f)
    `mplus` (success "prod_pullT" $ ALL (SEQ (MKT a f) (MKT b g)))
prod_pullT _ _ = mzero

data_pullT :: Rule
data_pullT _ (MKT (Data _ _ b) (DATAMAP c f)) = 
    do Eq <- teq b c
       success "data_pullT" $ ALL (aux b f)
    where aux :: Type a -> (PF (a -> a)) -> PF T
	  aux (Prod a b) (PROD f g) = SEQ (aux a f) (aux b g)
	  aux (List a) (LIST f)     = aux a f
	  aux a f                   = MKT a f
data_pullT _ _ = mzero


-- Pushing down apply.
nop_applyT :: Rule
nop_applyT _ (APPLY _ NOP) = success "nop_applyT" ID
nop_applyT _ _ = mzero

seq_applyT :: Rule
seq_applyT _ (APPLY t (SEQ f g)) = success "seq_applyT" (COMP t (APPLY t f) (APPLY t g))
seq_applyT _ _ = mzero

gmapT_applyT :: Rule
gmapT_applyT _ (APPLY t (ALL f)) = success "gmapT_applyT" (aux t)
  where
    aux :: Type a -> PF (a->a) 
    aux Int          = ID
    aux String       = ID
    aux (Prod a b)   = PROD (APPLY a f) (APPLY b f)
    aux (Either a b) = SUM (APPLY a f) (APPLY b f)
    aux (List a)     = LIST (APPLY a f)
    aux (Data s _ b) = DATAMAP b (auxN b)
    auxN :: Type a -> PF (a -> a)
    auxN (Prod a b)   = PROD (auxN a) (auxN b)
    auxN (Either a b) = SUM (auxN a) (auxN b)
    auxN (List a)     = LIST (auxN a)
    auxN a            = APPLY a f
gmapT_applyT _ _ = mzero

everywhere_applyT :: Rule
everywhere_applyT _ (APPLY a (EVERYWHERE f)) = 
    success "everywhere_applyT" $ APPLY a (SEQ f (ALL (EVERYWHERE f)))
everywhere_applyT _ _ = mzero

mkT_applyT :: Rule
mkT_applyT _ (APPLY a (MKT b f))  
  = case teq' a b of
      Just Eq   -> success "mkT_applyT" f
      Nothing   -> success "mkT_applyT" ID
mkT_applyT _ _ = mzero

mkT_applyT' :: Rule
mkT_applyT' _ (APPLY a (MKT b f))  
  = case teq' a b of
      Just Eq   -> mzero
      Nothing   -> success "mkT_applyT" ID
mkT_applyT' _ _ = mzero


{-
push_apply' :: Rule 
push_apply' _ (APPLY _ NOP)       = success "push_apply'" ID
push_apply' _ (APPLY t (SEQ f g)) = success "push_apply'" (COMP t (APPLY t f) (APPLY t g))
push_apply' _ (APPLY t (ALL f))   = success "push_apply'" (aux t)
  where
    aux :: Type a -> PF (a->a) 
    aux Int          = ID
    aux String       = ID
    aux (Prod a b)   = PROD (APPLY a f) (APPLY b f)
    aux (Either a b) = SUM (APPLY a f) (APPLY b f)
    aux (List a)     = LIST (APPLY a f)
    aux (Data s _ b) = DATAMAP b (auxN b)
    auxN :: Type a -> PF (a -> a)
    auxN (Prod a b)   = PROD (auxN a) (auxN b)
    auxN (Either a b) = SUM (auxN a) (auxN b)
    auxN (List a)     = LIST (auxN a)
    auxN a            = APPLY a f
push_apply' _ (APPLY a (EVERYWHERE f)) = success "push_apply'" $ APPLY a (SEQ f (ALL (EVERYWHERE f)))
push_apply' _ (APPLY t (MKT t' f))  
  = case teq' t t' of
      Just Eq   -> mzero
      Nothing   -> success "push_apply'" ID
push_apply' _ _ = mzero
-}

-- for non-recursive types
everywhere_mkt :: Rule
everywhere_mkt _ (APPLY a (MKT b f)) = 
    case teq a b of Just Eq -> success "everywhere_intro" $ (APPLY a (EVERYWHERE (MKT b f)))
		    Nothing -> mzero
everywhere_mkt _ _ = mzero


-- Heuristic rules for generalization: should be combined with gtop

{-
everyhole :: Query Bool
everyhole x = Data.Transform.Rewriting.everything (||) aux x
    where aux :: GenericQ Bool
          aux (PF _) (EVERYWHERE HOLE) = True
          aux _ _ = False

everywhere_intro :: Rule
everywhere_intro _ (MKT a f) = 
    do c <- context
       guard (not (everyhole c))
       success "everywhere_intro" (EVERYWHERE (MKT a f))
everywhere_intro _ _ = mzero
-}

everywhere_intro :: Rule
everywhere_intro _ (ALL (MKT a f)) = 
    success "everywhere_intro" (ALL (EVERYWHERE (MKT a f)))
everywhere_intro _ _ = mzero


gmapT_everywhere :: Rule
gmapT_everywhere _ (ALL (EVERYWHERE f)) = success "all_everywhere" (EVERYWHERE f)
gmapT_everywhere _ _ = mzero

-- unpush APPLY

unpush_apply :: Rule
unpush_apply b (LIST (APPLY a t)) =
    do let f = reduce sybt2pf b (LIST (APPLY a t))
	   g = reduce sybt2pf b (APPLY (List a) t)
       guard (geq (PF b) f g)
       success "unpush_apply" (APPLY (List a) t)
unpush_apply d@(Func a@(Data s ep c) _) (DATAMAP _ (APPLY b t)) =
    do Eq <- teq c b
       let f = reduce sybt2pf d (DATAMAP b (APPLY b t))
	   g = reduce sybt2pf d (APPLY (Data s ep b) t)
       guard (geq (PF d) f g)
       success "unpush_apply" (APPLY (Data s ep b) t)
unpush_apply x@(Func (Prod a b) (Prod c _)) (PROD h (APPLY _ t)) =
    do Eq <- teq a c
       let f = reduce sybt2pf x (PROD h (APPLY b t))
	   g = reduce sybt2pf x (APPLY (Prod a b) t)
       guard (geq (PF x) f g)
       success "unpush_apply" (APPLY (Prod a b) t)
unpush_apply x@(Func (Prod a b) (Prod _ c)) (PROD (APPLY _ t) h) =
    do Eq <- teq b c
       let f = reduce sybt2pf x (PROD (APPLY a t) h)
	   g = reduce sybt2pf x (APPLY (Prod a b) t)
       guard (geq (PF x) f g)
       success "unpush_apply" (APPLY (Prod a b) t)
unpush_apply _ _ = mzero

-- Specific maps can be turned into generic maps:
all_intro :: Rule
all_intro = aux1 ||| aux2 
  where
    aux1,aux2 :: Rule
    aux1 (Func (List a) (List b)) (LIST f) = do
      Eq <- teq a b
      return (APPLY (List a) (ALL (MKT a f)))
    aux1 (Func (Prod a c) (Prod b d)) (PROD f g) = do
      Eq <- teq a b; Eq <- teq c d
      guard (maybe False (const True) $ teq a c) 
      return (APPLY (Prod a c) (ALL (ADHOC a f (MKT c g))))
    aux1 _ _ = mzero
    aux2 (Func (Prod a c) (Prod b d)) (PROD f g) = do
      Eq <- teq a b; Eq <- teq c d
      Eq <- teq a c; guard (geq (PF (Func a c)) f g)
      return (APPLY (Prod a c) (ALL (MKT c g)))
    aux2 _ _ = mzero

-- Roll down REC inside apply.
apply_rec :: Rule
apply_rec _ (APPLY t (REC f))   = success "apply_rec" $ APPLY t (f (REC f))
apply_rec _ _                   = mzero

-----------------------------------------------------------------------------
-- * Type-unifying

empty_all :: Rule
empty_all _ (ALLQ Empty)   = return Empty
empty_all _ _              = mzero


{-
-- ....
lift_mkQ :: Rule
lift_mkQ _ (Union (MKQ a b f) (MKQ c d g)) = do
  Eq <- teq a c; Eq <- teq b d
  return (MKQ a b (COMP (Prod (List b) (List b)) APPEND (SPLIT f g)))
lift_mkQ _ _ = mzero
-}

push_applyq :: Rule
push_applyq _ (APPLYQ a (MKQ m b f)) =
    case teq' a b
    of Just Eq   -> success "push_applyq" f
       otherwise -> success "push_applyq" (ZERO m)
push_applyq (Func _ r) (APPLYQ a (UNION m f g)) =
    success "push_applyq" $ COMP (Prod r r) (PLUS m) (SPLIT (APPLYQ a f) (APPLYQ a g))
push_applyq (Func _ r) (APPLYQ Dynamic (GMAPQ m (f :: PF (GQ r)))) = mzero
push_applyq (Func _ r) (APPLYQ a (GMAPQ m (f :: PF (GQ r)))) = success "push_applyq" (aux a)
    where aux :: Type a -> PF (a -> r)
	  aux Int          = ZERO m
	  aux String       = ZERO m
	  aux (Prod a b)   = COMP (Prod r r) (PLUS m) (PROD (APPLYQ a f) (APPLYQ b f))
	  aux (Either a b) = EITHER (APPLYQ a f) (APPLYQ b f)
	  aux (List a)     = COMP (List r) (FOLD m) (LIST (APPLYQ a f))
	  aux (Data _ _ b) = COMP b (auxN b) OUT
	  auxN :: Type a -> PF (a -> r)
	  auxN (Prod a b)   = COMP (Prod r r) (PLUS m) (PROD (auxN a) (auxN b))
	  auxN (Either a b) = EITHER (auxN a) (auxN b)
	  auxN (List a)     = COMP (List r) (FOLD m) (LIST (auxN a))
	  auxN a            = APPLYQ a f
push_applyq (Func _ r) (APPLYQ a (EVERYTHING m f)) = 
    success "push_applyq" $ APPLYQ a (UNION m f (GMAPQ m (EVERYTHING m f)))
push_applyq _ _ = mzero

push_applyq' :: Rule
push_applyq' _ (APPLYQ a (MKQ r b f)) =
    case (teq a b) of Just Eq -> mzero
		      Nothing -> success "push_applyq'" (ZERO r)
push_applyq' (Func _ r) (APPLYQ a (UNION m f g)) =
    success "push_applyq'" $ COMP (Prod r r) (PLUS m) (SPLIT (APPLYQ a f) (APPLYQ a g))
push_applyq' (Func _ r) (APPLYQ Dynamic (GMAPQ m (f :: PF (GQ r)))) = mzero
push_applyq' (Func _ r) (APPLYQ a (GMAPQ m (f :: PF (GQ r)))) = success "push_applyq'" (aux a)
    where aux :: Type a -> PF (a -> r)
	  aux Int          = ZERO m
	  aux String       = ZERO m
	  aux (Prod a b)   = COMP (Prod r r) (PLUS m) (PROD (APPLYQ a f) (APPLYQ b f))
	  aux (Either a b) = EITHER (APPLYQ a f) (APPLYQ b f)
	  aux (List a)     = COMP (List r) (FOLD m) (LIST (APPLYQ a f))
	  aux (Data _ _ b) = COMP b (auxN b) OUT
	  auxN :: Type a -> PF (a -> r)
	  auxN (Prod a b)   = COMP (Prod r r) (PLUS m) (PROD (auxN a) (auxN b))
	  auxN (Either a b) = EITHER (auxN a) (auxN b)
	  auxN (List a)     = COMP (List r) (FOLD m) (LIST (auxN a))
	  auxN a            = APPLYQ a f
push_applyq' (Func _ r) (APPLYQ a (EVERYTHING m f)) = 
    success "push_applyq'" $ APPLYQ a (UNION m f (GMAPQ m (EVERYTHING m f)))
push_applyq' _ _ = mzero


empty_pullQ :: Rule
empty_pullQ _ (MKQ m a (ZERO _)) = success "empty_pullQ" (EMPTYQ m)
empty_pullQ _ _ = mzero

data_pullQ :: Rule
data_pullQ (TU (List d)) (MKQ r (Data _ _ b) (COMP c (f :: PF (b -> [r])) OUT)) =
    do Eq <- teq b c
       success "data_pullQ" $ GMAPQ r (aux b f)
    where aux :: Type a -> (PF (a -> [r])) -> PF (GQ [r])
	  aux (Prod a b) (COMP _ f FST) = aux a f
	  aux (Prod a b) (COMP _ f SND) = aux b f
	  aux (List a) (LIST f) = aux a (COMP d WRAP f)
	  aux a f                   = MKQ r a f
data_pullQ (TU d) (MKQ r (Data _ _ b) (COMP c (f :: PF (b -> r)) OUT)) =
    do Eq <- teq b c
       success "data_pullQ" $ GMAPQ r (aux b f)
    where aux :: Type a -> (PF (a -> r)) -> PF (GQ r)
	  aux (Prod a b) (COMP _ f FST) = aux a f
	  aux (Prod a b) (COMP _ f SND) = aux b f
	  aux (List a) (COMP _ (FOLD _) (LIST f)) = aux a f
	  aux a f                   = MKQ r a f
data_pullQ _ _ = mzero

map_pullQ :: Rule
map_pullQ _ (MKQ r (List a) (COMP _ (FOLD _) (LIST f))) =
    success "map_pullQ" $ GMAPQ r (MKQ r a f)
map_pullQ (TU (List b)) (MKQ r (List a) (LIST f)) =
    success "map_pullQ" $ GMAPQ r (MKQ r a (COMP b WRAP f))
map_pullQ _ _ = mzero

prod_pullQ :: Rule
prod_pullQ _ (MKQ r (Prod a b) (COMP _ f FST)) = 
    do Eq <- teq a b
       mzero
  `mplus` (success "prod_pullQ" $ GMAPQ r (MKQ r a f))
prod_pullQ _ (MKQ r (Prod a b) (COMP _ f SND)) = 
    do Eq <- teq a b
       mzero
  `mplus` (success "prod_pullQ" $ GMAPQ r (MKQ r b f))
prod_pullQ _ _ = mzero

wrap_pullQ :: Rule
wrap_pullQ _ (MKQ r (Data n _ _) (COMP _ WRAP TODYN)) = success "wrap_pullQ" (NAME n)
wrap_pullQ _ (MKQ r a (COMP _ WRAP TODYN)) = success "wrap_pullQ" SELF
wrap_pullQ _ _ = mzero

-- heristic rules for query generalization

everything_intro :: Rule
everything_intro _ (GMAPQ r (MKQ _ b f)) = success "everything_intro" $ GMAPQ r (EVERYTHING r (MKQ r b f))
everything_intro _ _ = mzero

gmapQ_everything :: Rule
gmapQ_everything _ (GMAPQ r (EVERYTHING _ q)) = success "gmapQ_everything" $ EVERYTHING r q
gmapQ_everything _ _ = mzero

desc_intro :: Rule
desc_intro _ CHILD = success "desc_intro" DESCorSELF
desc_intro _ (CHILD :/: DESCorSELF) = success "desc_intro" DESCorSELF
desc_intro _ (CHILD :/: (DESCorSELF :/: f)) = success "desc_intro" (DESCorSELF :/: f)
desc_intro _ _ = mzero

-- for non-recursive types
everything_mkq :: Rule
everything_mkq _ (APPLYQ a (MKQ r b f)) = 
    do Eq <- teq a b
       success "everthing_intro" $ (APPLYQ a (EVERYTHING r (MKQ r b f)))
everything_mkq _ _ = mzero

unpush_applyq :: Rule
unpush_applyq b t@(COMP _ (FOLD r) (LIST (APPLYQ a h))) =
    do let f = reduce sybq2pf b t
	   g = reduce sybq2pf b (APPLYQ (List a) h)
       guard (geq (PF b) f g)
       success "unpush_applyq" (APPLYQ (List a) h)
unpush_applyq x@(Func (Prod a b) r) q@(COMP (Prod rl rr) _ (PROD (APPLYQ _ t) h)) =
    do Eq <- teq rr r
       Eq <- teq rl r
       let f = reduce sybq2pf x q
	   g = reduce sybq2pf x (APPLYQ (Prod a b) t)
       guard (geq (PF x) f g)
       success "unpush_applyq" (APPLYQ (Prod a b) t)
unpush_applyq x@(Func (Prod a b) r) q@(COMP (Prod rl rr) _ (PROD h (APPLYQ _ t))) =
    do Eq <- teq rr r
       Eq <- teq rl r
       let f = reduce sybq2pf x q
	   g = reduce sybq2pf x (APPLYQ (Prod a b) t)
       guard (geq (PF x) f g)
       success "unpush_applyq" (APPLYQ (Prod a b) t)
unpush_applyq d@(Func a@(Data s ep c) _) q@(COMP b (APPLYQ _ t) OUT) =
    do Eq <- teq c b
       let f = reduce sybq2pf d q
	   g = reduce sybq2pf d (APPLYQ a t)
       guard (geq (PF d) f g)
       success "unpush_applyq" (APPLYQ a t)
unpush_applyq _ _ = mzero


-- Some rules to handle Datas

datamap_id :: Rule
datamap_id _ (DATAMAP _ ID) = success "datamap_id" ID
datamap_id _ _ = mzero

zero_fusion :: Rule
zero_fusion _ (COMP _ (ZERO m) _) = success "zero_fusion" $ ZERO m
zero_fusion _ _ = mzero

true_fusion :: Rule
true_fusion _ (COMP _ TRUE _) = success "true_fusion" TRUE
true_fusion _ _ = mzero


datas :: Rule
datas = once (top datamap_id ||| top zero_fusion)

-- Rules for foldl
{-
foldl_nil :: Rule
foldl_nil _ (COMP _ (FOLDL _) (SPLIT f (COMP _ NIL BANG))) =
    success "foldl_nil" $ f
foldl_nil _ _ = mzero

foldl_plus :: Rule
foldl_plus _ (COMP (Prod r _) (FOLDL (PLUS _)) (SPLIT (COMP _ (ZERO _) BANG) (COMP _ CONS (SPLIT h t)))) = 
    success "foldl_plus" $ COMP (Prod r r) (PLUS r) (SPLIT h (COMP (Prod r (List r)) (FOLDL (PLUS r)) (SPLIT (COMP One (ZERO r) BANG) t)))
foldl_plus _ (COMP _ (FOLDL (PLUS r)) (SPLIT f (COMP _ CONS (SPLIT (COMP _ (ZERO _) BANG) g)))) =
    success "foldl_plus" $ COMP (Prod r (List r)) (FOLDL (PLUS r)) (SPLIT f g)
foldl_plus _ _ = mzero

foldl_map :: Rule
foldl_map _ (COMP (Prod r (List t)) (FOLDL o) (SPLIT (COMP _ z BANG) (COMP (List a) (LIST f) g))) = 
    success "foldl_map" $ (COMP (Prod r (List a)) (FOLDL (COMP (Prod r t) o (PROD ID f))) (SPLIT (COMP One z BANG) g))
foldl_map _ _ = mzero

foldl_fst :: Rule
foldl_fst _ (FOLDL FST) = success "foldl_fst" FST
foldl_fst _ _ = mzero
-}

fold_mapzero :: Rule
fold_mapzero _ (COMP (List a) (FOLD m) (LIST (ZERO _))) = success "fold_mapzero" (ZERO m)
fold_mapzero _ (COMP (List a) (FOLD m) (COMP b (LIST (ZERO _)) f)) = success "fold_mapzero" (COMP b (ZERO m) f)
fold_mapzero _ _ = mzero

{-
fold_maptrue :: Rule
fold_maptrue _ (COMP (List Bool) (FOLD m) (LIST TRUE)) = success "fold_maptrue" TRUE
fold_maptrue _ (COMP (List Bool) (FOLD m) (COMP b (LIST TRUE) f)) = success "fold_maptrue" (COMP b TRUE f)
fold_maptrue _ _ = mzero
-}

fold_wrap :: Rule
fold_wrap _ (COMP (List (List a)) (FOLD m) (LIST WRAP)) = success "fold_wrap" ID
fold_wrap _ (COMP (List (List a)) (FOLD m) (LIST (COMP _ WRAP f))) = success "fold_wrap" (LIST f)
fold_wrap _ (COMP (List (List a)) (FOLD m) (COMP _ (LIST WRAP) f)) = success "fold_wrap" f
fold_wrap _ (COMP (List (List a)) (FOLD m) (COMP b (LIST (COMP _ WRAP g)) f)) = success "fold_wrap" (COMP b (LIST g) f)
fold_wrap _ (COMP (List a) (FOLD m) WRAP) = success "fold_wrap" ID
fold_wrap _ (COMP (List a) (FOLD m) (COMP _ WRAP f)) = success "fold_wrap" f
fold_wrap _ _ = mzero

fold_plus :: Rule
fold_plus (Func _ (List a)) (COMP _ (FOLD _) (PLUS _)) = success "fold_plus" (COMP (Prod (List a) (List a)) (PLUS mlist) (PROD (FOLD mlist) (FOLD mlist)))
fold_plus (Func _ (List a)) (COMP _ (FOLD _) (COMP b (PLUS _) g)) = success "fold_plus" (COMP (Prod (List a) (List a)) (PLUS mlist) (COMP b (PROD (FOLD mlist) (FOLD mlist)) g))
fold_plus _ _ = mzero

list_wrap :: Rule
list_wrap (Func _ (List b)) (COMP _ (LIST f) WRAP) = success "list_wrap" (COMP b WRAP f)
list_wrap (Func _ (List b)) (COMP _ (LIST f) (COMP c WRAP g)) = success "list_wrap" (COMP b WRAP (COMP c f g))
list_wrap _ _ = mzero

list_plus :: Rule
list_plus (Func _ b) (COMP _ (LIST f) (PLUS _)) = success "list_plus" (COMP (Prod b b) (PLUS mlist) (PROD (LIST f) (LIST f)))
list_plus (Func _ b) (COMP _ (LIST f) (COMP a (PLUS _) g)) = success "list_plus" (COMP (Prod b b) (PLUS mlist) (COMP a (PROD (LIST f) (LIST f)) g))
list_plus _ _ = mzero

list_zero :: Rule
list_zero _ (COMP _ (LIST _) (ZERO _)) = success "list_zero" (ZERO mlist)
list_zero _ (COMP _ (FOLD m) (ZERO _)) = success "list_zero" (ZERO m)
list_zero _ _ = mzero

list_list :: Rule
list_list _ (COMP (List a) (LIST f) (LIST g)) = success "list_list" (LIST (COMP a f g))
list_list _ (COMP (List a) (LIST f) (COMP b (LIST g) h)) = success "list_list" (COMP b (LIST (COMP a f g)) h)
list_list _ _ = mzero

plus_zero :: Rule
plus_zero _ (COMP _ (PLUS _) (SPLIT (ZERO _) f)) = success "plus_zero" f
plus_zero _ (COMP _ (PLUS _) (SPLIT f (ZERO _))) = success "plus_zero" f
plus_zero _ _ = mzero

{-
plus_wrap :: Rule
plus_wrap (Func _ (List a)) (COMP _ (PLUS _) (SPLIT (COMP _ WRAP g) h)) = success "plus_wrap" (COMP (Prod a (List a)) CONS (SPLIT g h))
plus_wrap _ _ = mzero

list_cons :: Rule
list_cons (Func _ (List a)) (COMP _ (LIST f) CONS) = success "list_cons" $ COMP (Prod a (List a)) CONS (PROD f (LIST f))
list_cons (Func _ (List a)) (COMP _ (LIST f) (COMP b CONS g)) = success "list_cons" $ COMP (Prod a (List a)) CONS (COMP b (PROD f (LIST f)) g)
list_cons _ _ = mzero

fold_cons :: Rule
fold_cons (Func _ (List a)) (COMP _ (FOLD m) CONS) = success "fold_cons" $ COMP (Prod (List a) (List a)) (PLUS m) (PROD ID (FOLD m))
fold_cons (Func _ (List a)) (COMP _ (FOLD m) (COMP b CONS f)) = success "fold_cons" $ COMP (Prod (List a) (List a)) (PLUS m) (COMP b (PROD ID (FOLD m)) f)
fold_cons _ _ = mzero
-}

list_fold :: Rule
list_fold (Func _ (List a)) (COMP _ (LIST f) (FOLD _)) = success "list_fold" $ COMP (List (List a)) (FOLD mlist) (LIST (LIST f))
list_fold (Func _ (List a)) (COMP _ (LIST f) (COMP b (FOLD _) g)) = success "list_fold" $ COMP (List (List a)) (FOLD mlist) (COMP b (LIST (LIST f)) g)
list_fold _ _ = mzero

{-
fold_list :: Rule
fold_list (Func (List (List b)) _) (COMP _ (FOLD m) (LIST (LIST f))) = success "fold_list" $ (COMP (List b) (LIST f) (FOLD mlist))
fold_list (Func _ _) (COMP _ (FOLD m) (COMP (List (List b)) (LIST (LIST f)) g)) = success "fold_list" $ (COMP (List b) (LIST f) (COMP (List (List b)) (FOLD mlist) g))
fold_list _ _ = mzero
-}

fold_fold_list :: Rule
fold_fold_list (Func _ (List a)) (COMP _ (FOLD m) (COMP _ (FOLD o) (LIST f))) = 
    success "fold_fold_list" $ COMP (List (List a)) (FOLD mlist) (LIST (COMP (List (List a)) (FOLD mlist) f))
fold_fold_list (Func _ (List a)) (COMP _ (FOLD m) (COMP _ (FOLD o) (COMP c (LIST f) g))) = 
    success "fold_fold_list" $ COMP (List (List a)) (FOLD mlist) (COMP c (LIST (COMP (List (List a)) (FOLD mlist) f)) g)
{-
fold_fold_list (Func _ (List a)) (COMP _ (FOLD m) (COMP _ (FOLD o) (LIST (COMP b (PLUS n) f)))) = 
    success "fold_fold_list" $ COMP (List (List a)) (FOLD mlist) (LIST (COMP (List (List a)) (FOLD mlist) (COMP b (PLUS n) f)))
fold_fold_list (Func _ (List a)) (COMP _ (FOLD m) (COMP _ (FOLD o) (COMP c (LIST (COMP b (PLUS n) f)) g))) = 
    success "fold_fold_list" $ COMP (List (List a)) (FOLD mlist) (COMP c (LIST (COMP (List (List a)) (FOLD mlist) (COMP b (PLUS n) f))) g)
-}
fold_fold_list _ _ = mzero


folds :: Rule
folds = innermost aux
    where aux :: Rule
	  aux = top comp_assocr ||| top nat_id ||| top zero_fusion ||| top plus_zero |||  top prod_cancel ||| top prod_fusion ||| top prod_def ||| top fold_mapzero -- ||| top foldl_fst ||| top foldl_nil ||| top foldl_plus ||| top foldl_map


filter_true :: Rule
filter_true _ (FILTER TRUE) = success "filter_true" ID
filter_true _ _ = mzero

filter_false :: Rule
filter_false _ (FILTER (ZERO _)) = success "filter_false" (ZERO mlist)
filter_false _ _ = mzero

filter_wrap :: Rule
filter_wrap _ (COMP _ (FILTER f) (COMP a WRAP g)) = 
    success "filter_wrap" $ COND (COMP a f g) (COMP a WRAP g) (ZERO mlist)
filter_wrap _ _ = mzero

filter_cat :: Rule
filter_cat _ (COMP (List a) (FILTER f) (COMP _ (PLUS m) g)) = 
    success "filter_cat" $ COMP (Prod (List a) (List a)) (PLUS m) (COMP (Prod (List a) (List a)) (PROD (FILTER f) (FILTER f)) g)
filter_cat _ _ = mzero

filter_map :: Rule
filter_map (Func (List b) _) (COMP (List a) (FILTER f) (LIST g)) = success "filter_map" $ COMP (List b) (LIST g) (FILTER (COMP a f g))
filter_map _ (COMP (List a) (FILTER f) (COMP (List b) (LIST g) h)) = success "filter_map" $ COMP (List b) (LIST g) (COMP (List b) (FILTER (COMP a f g)) h)
filter_map _ _ = mzero

filter_concat :: Rule
filter_concat _ (COMP (List a) (FILTER f) (FOLD m)) = success "filter_fold" $ COMP (List (List a)) (FOLD m) (LIST (FILTER f))
filter_concat _ (COMP (List a) (FILTER f) (COMP b (FOLD m) g)) = success "filter_fold" $ COMP (List (List a)) (FOLD m) (COMP b (LIST (FILTER f)) g)
filter_concat _ _ = mzero

filter_nil :: Rule
filter_nil _ (COMP _ (FILTER f) (ZERO m)) = success "filter_nil" $ ZERO m
filter_nil _ _ = mzero

cond_true :: Rule
cond_true _ (COND TRUE f g) = success "cond_true" f
cond_true _ _ = mzero

cond_false :: Rule
cond_false _ (COND (ZERO _) f g) = success "cond_false" g
cond_false _ _ = mzero


-----------------------------------------------------------------------------

optimize' :: Rule
optimize' = many (prods ||| maps ||| sums ||| datas)

t2pf :: Rule
t2pf = innermost (top nop_applyT ||| top seq_applyT ||| top gmapT_applyT ||| top mkT_applyT ||| top everywhere_applyT)

sybt2pf :: Rule
sybt2pf = t2pf >>> optimize' >>> prods_inv

pf2t :: Rule
pf2t = innermost (top id_pullT ||| top map_pullT ||| top prod_pullT ||| top data_pullT) >>>
       innermost (top id_seq ||| top id_all ||| top gmapT_fusion ||| top everywhere_def_inv ||| gtop sybt2pf everywhere_intro ||| gtop sybt2pf gmapT_everywhere)

{-       
pf2sybt :: Rule
pf2sybt = innermost (top id_pullT ||| top map_pullT ||| top prod_pullT ||| top data_pullT ||| top id_seq ||| top nop_applyT ||| top seq_applyT ||| top gmapT_applyT ||| top mkT_applyT' ||| top everywhere_applyT ||| top datamap_id ||| top prod_reflex ||| top listmap_id ||| top nat_id) >>> innermost (top everywhere_mkt ||| top unpush_apply)

t2pf' :: Rule
t2pf' = innermost (top nop_applyT ||| top seq_applyT ||| top gmapT_applyT ||| top mkT_applyT' ||| top everywhere_applyT)

generalize_t = t2pf' >>> optimize' >>> prods_inv >>> innermost (top everywhere_mkt ||| top unpush_apply)
-}

sybq2pf :: Rule
sybq2pf = innermost (top push_applyq) >>> folds >>> prods_inv

pf2q :: Rule
pf2q = innermost (top comp_assocl ||| top empty_pullQ ||| top map_pullQ ||| top prod_pullQ ||| top data_pullQ ||| top wrap_pullQ) >>>
       innermost (top comp_assocl ||| top child_inv ||| gtop sybq2pf everything_intro ||| gtop sybq2pf gmapQ_everything ||| gtop xpath desc_intro)

pf2xp :: Rule
pf2xp = innermost (top comp_assocl ||| top empty_pullQ ||| top map_pullQ ||| top prod_pullQ ||| top data_pullQ ||| top wrap_pullQ) >>>
       innermost (top comp_assocl ||| top child_inv)

{-
pf2sybq :: Rule
pf2sybq = innermost (top push_mkq ||| top push_applyq') >>> innermost (top everything_intro) >>> innermost (top unpush_applyq)
-}

-----------------------------------------------------------------------------

push_xpath :: Rule
push_xpath _ (APPLYQ (Data s _ _) (NAME l)) | s==l = success "push_xpath" (COMP Dynamic WRAP TODYN)
push_xpath _ (APPLYQ Dynamic (NAME l)) = mzero
push_xpath _ (APPLYQ a (NAME l)) = success "push_xpath" $ ZERO mlist
--push_xpath _ (APPLYQ Int CHILD) = success "push_xpath" $ COMP One (ZERO mlist) BANG
--push_xpath _ (APPLYQ String CHILD) = success "push_xpath" $ COMP One (ZERO mlist) BANG
--push_xpath _ (APPLYQ a@(Data s _ _) SELF) = success "push_path" (APPLYQ a (NAME s))
push_xpath _ (APPLYQ Dynamic SELF) = success "push_xpath" WRAP
push_xpath (Func a _) (COMP _ (APPLYQ Dynamic f) TODYN) = success "push_xpath" (APPLYQ a f)
push_xpath _ (COMP _ (APPLYQ Dynamic f) (COMP a TODYN g)) = success "push_xpath" (COMP a (APPLYQ a f) g)
push_xpath _ (APPLYQ a SELF) = success "push_xpath" (COMP Dynamic WRAP TODYN)
push_xpath (Func _ (List r)) (APPLYQ a (f :/: g)) = 
    success "push_xpath" $ COMP (List (List r)) (FOLD mlist) (COMP (List Dynamic) (LIST (APPLYQ Dynamic g)) (APPLYQ a f))
push_xpath (Func _ Bool) (APPLYQ a (f :/: g)) = 
    success "push_xpath" $ COMP (List Bool) (FOLD mbool) (COMP (List Dynamic) (LIST (APPLYQ Dynamic g)) (APPLYQ a f))
push_xpath _ (APPLYQ a (f :?: p)) = success "push_xpath" $ COMP (List Dynamic) (FILTER (APPLYQ Dynamic p)) (APPLYQ a f)
push_xpath _ (APPLYQ a NONEMPTY) = success "push_xpath" $ TRUE
push_xpath _ _ = mzero

xpath_alg :: Rule
--xpath_alg _ (SELF :/: SELF) = success "xpath_alg" CHILD
xpath_alg _ DESCorSELF = success "xpath_alg" (UNION mlist SELF DESCENDANT)
xpath_alg (TU (List a)) (UNION m f g :/: h) = success "xpath_alg" $ UNION mlist (f :/: h) (g :/: h)
xpath_alg _ DESCENDANT = success "xpath_alg" (EVERYTHING mlist CHILD)
xpath_alg _ CHILD = success "xpath_alg" (GMAPQ mlist SELF)
xpath_alg _ _ = mzero


child_inv :: Rule
child_inv _ (GMAPQ _ f) = success "child_def_inv" $ CHILD :/: f
child_inv _ _ = mzero

xpath :: Rule
xpath = innermost (top comp_assocr ||| top nat_id ||| top zero_fusion ||| top true_fusion ||| top plus_zero |||  top prod_cancel ||| top prod_fusion ||| top prod_def ||| top fold_wrap ||| top fold_plus ||| top list_zero ||| top list_wrap ||| top list_list ||| top list_plus ||| top fold_mapzero ||| top list_fold ||| top fold_fold_list ||| top filter_true ||| top filter_false ||| top filter_wrap ||| top cond_true ||| top cond_false ||| top filter_nil ||| top filter_cat ||| top filter_map ||| top filter_concat ||| top xpath_alg ||| top push_xpath ||| top push_applyq) 

optimizexp :: Typeable b => Type a -> PF (GQ b) -> PF (a -> b)
optimizexp a q = reduce xpath (Func a typeof) (APPLYQ a q)