{-# OPTIONS -fglasgow-exts #-}
-----------------------------------------------------------------------------
-- |
--
-- Combinators for Two-Level Transformation.
--
-- See
--
--     Type-safe Two-level Data Transformation
-- 
-- by
--
--     Alcino Cunha, Jose Nuno Oliveira, Joost Visser
--
-----------------------------------------------------------------------------

module Data.Transform.TwoLevelCombinators where

import Data.Transform.Type hiding (to,from)
import Pointless.Combinators

import Data.Map as Map
import Control.Monad
import Control.Monad.State hiding (when)

import Prelude hiding (all)
import Debug.Trace

-----------------------------------------------------------------------------

-- Determine whether the left alternative is present.
isLeft :: Either a b -> Bool
isLeft = either (const True) (const False)


-- Encapsulation of type-changing transformations
-- |Rep| is a pair of bi-directional value-level data transformation functions.
data Rep a b = Rep {to :: a -> b, from :: b -> a}

-- Pair of identity value-level tranformations. Used to define 'nop'.
idrep :: Rep a a
idrep = Rep {to = id, from = id}

-- Sequential composition of value-level transformations. Used to define '($>>>$)'.
comprep :: Rep a b -> Rep b c -> Rep a c
comprep f g = Rep {from = (from f) . (from g), to = (to g) . (to f)}

-- Lift value-level transformations to a functor.
-- Used to define traversal strategies, like 'once'.
maprep :: Functor f => Rep a b -> Rep (f a) (f b)
maprep r = Rep {to = fmap (to r), from = fmap (from r)}

prodrep :: Rep a b -> Rep c d -> Rep (a,c) (b,d)
prodrep (Rep {to = to1, from = from1}) (Rep {to = to2, from = from2})
    = Rep {to = to1 >< to2, from = from1 >< from2}


-- Lift value-level transformations to a bi-functor.
-- Used to define traversal strategies, like 'once'.
bimaprep :: Bifunctor f => Rep a b -> Rep c d -> Rep (f a c) (f b d)
bimaprep x y = Rep {to = bimap (to x) (to y), from = bimap (from x) (from y)}

-- Class of bi-functors. Needed for 'bimaprep'
class Bifunctor f where
    bimap :: (a -> b) -> (c -> d) -> f a c -> f b d

instance Bifunctor (,) where
    bimap = (><)

instance Bifunctor Either where
    bimap = (-|-)

instance Bifunctor Map where
    bimap f g = mapKeysMonotonic f . Map.map g



type NameSupply m a =  StateT ([String],Int) m a

-- This state monad transformer is used to transform the |Maybe| monad 
-- in two-level transformation rules:
type Rule = forall a . Type a -> NameSupply Maybe (View (Type a))

-- The View constructor expreses that a |type a| can be represented 
-- by a |type b|, as witnessed by value-level functions that convert 
-- between these types. Note that the target type does not escape, 
-- which means that it is implicitly existentially qualified.
data View a where
    View :: Rep a b -> Type b -> View (Type a)

instance Show (View a) where
    show (View _ b) = "(View (Rep <to> <from>) "++(show b)++")"

-- Print the string representation of the target type encapsulated in a view.
showType :: View a -> String
showType (View _ b) = show b

-- Sequential composition.
(>>>) :: Rule -> Rule -> Rule
(f >>> g) a = do (View r b) <- f a
                 (View s c) <- g b
                 return $ (View (comprep r s) c)

-- Left-biased choice.
(|||) :: Rule -> Rule -> Rule
(f ||| g) x = f x `mplus` g x

-- Do-nothing operation.
-- Unit for sequential composition.
-- Left-zero and right-unit for left-biased choice.
nop :: Rule
nop x = return $ (View idrep x)

-- Repeat until failure.
-- Zero or more repetitions.
many :: Rule -> Rule
many r = (r >>> many r) ||| nop

-- Apply argument rule exactly once, at arbitrary depth.
once :: Rule -> Rule
once r Int  = r Int
once r Bool = r Bool
once r Char = r Char
once r One  = r One
once r String = r String
once r (List a) = r (List a) `mplus` 
                  (do (View s b) <- once r a
                      return $ (View (maprep s) (List b)))
once r (Set a) = r (Set a) `mplus`
                 (do (View s b) <- once r a
                     return $ (View (maprep s) (Set b)))
once r (Map a b) = r (Map a b) `mplus`
                   (do (View s c) <- once r a
                       return $ (View (bimaprep s idrep) (Map c b))) `mplus`
                   (do (View s c) <- once r b
                       return $ (View (bimaprep idrep s) (Map a c)))
once r (Either a b) = r (Either a b) `mplus`
                      (do (View s c) <- once r a
                          return $ (View (bimaprep s idrep) (Either c b))) `mplus`
                      (do (View s c) <- once r b
                          return $ (View (bimaprep idrep s) (Either a c)))
once r (Prod a b) = r (Prod a b) `mplus`
                    (do (View s c) <- once r a
                        return $ (View (bimaprep s idrep) (Prod c b))) `mplus`
                    (do (View s c) <- once r b
                        return $ (View (bimaprep idrep s) (Prod a c)))
once r (Ann l a) = r (Ann l a) `mplus`
                   (do (View s b) <- once r a
                       return $ (View s (Ann l b)))
-- once r X = mzero
once r x = r x

outerMost :: Rule -> Rule
outerMost r = many (once r)

-- Apply argument rule everywhere
everywhere :: Rule -> Rule
everywhere r = all (everywhere r) >>> r

all :: Rule -> Rule
all r (List a) = do 
  (View s b) <- r a
  return $ (View (maprep s) (List b))
all r (Set a) = do 
  (View s b) <- r a
  return $ (View (maprep s) (Set b))
all r (Map a b) =  do 
  (View s c) <- r a
  (View t d) <- r b
  return $ (View (bimaprep s t) (Map c d))
all r (Either a b) =  do 
  (View s c) <- r a
  (View t d) <- r b
  return $ (View (bimaprep s t) (Either c d))
all r (Prod a b) =  do 
  (View s c) <- r a
  (View t d) <- r b
  return $ (View (bimaprep s t) (Prod c d))
all r (Ann l a) = do
  (View s b) <- r a
  return $ (View s (Ann l b))
all r t  = return $ View idrep t

-- Function |try| applies the argument rule if possible, but never fails.
try :: Rule -> Rule
try r = r ||| nop

(>>>>) :: Rule -> Rule -> Rule
f >>>> g = f >>> print >>> g
  where
    print :: Rule
    print t = return (View (Rep (tr "To") (tr "From")) t)
      where tr msg x = trace (msg++" "++gshow t x) x

-----------------------------------------------------------------------------

-- * Using dynamic types

-- Dynamic type
data DynType where
    DynT :: Type a -> DynType

-- Apply a generic function to a dynamically typed value.
applyDyn :: (forall a . Type a -> a -> b) -> Dynamic -> b
applyDyn f (Dyn ta a) = f ta a

-- Turn a function on a dynamic type into a generic function.
dynTApply :: (DynType -> b) -> Type a -> b
dynTApply f ta = f (DynT ta)

-- Apply a generic function to a dynamic type.
applyDynT :: (forall a . Type a -> b) -> DynType -> b
applyDynT f (DynT t) = f t

-- Turn a function on dynamics into a generic function.
dynApply :: (Dynamic -> b) -> Type a -> a -> b
dynApply f ta a = f (Dyn ta a)

instance Show Dynamic where
    show (Dyn a x) = "Dyn "++(gshow a x)
    
instance Show DynType where
    show (DynT t) = show t
    
-- Variation on 'forth' that does not need static knowledge of the
--   target type, and produces a dynamically typed value instead.
forthDyn :: View (Type a) -> a -> Maybe Dynamic
forthDyn (View rep b) x = return (Dyn b (to rep x))

-- Variation on 'forth' that does not need static knowledge of the
--   target type, and produces a dynamic type instead.
forthDynT :: View (Type a) -> Maybe DynType
forthDynT (View _ b) = return (DynT b)

-- Variation on 'back' that does not need static knowledge of the
--   original type, and produces a dynamically typed value instead.
backDyn :: View (Type a) -> Dynamic -> Maybe a
backDyn (View rep b) (Dyn c x) = do
  Eq <- teq b c
  return (from rep x)

-- Variation on 'back' that does not need static knowledge of the
--   original type, and produces a dynamically typed value instead.
-- Returns a Dynamic type.
backDyn2 :: Type a -> View (Type a) -> Dynamic -> Maybe Dynamic
backDyn2 t (View rep b) d@(Dyn c x) = do
        Eq <- teq b c
        return $ Dyn t $ from rep x


-----------------------------------------------------------------------------
