module Data.Transform.TwoLevelFunctions where

import Data.Transform.Type hiding (from,to)
import Data.Transform.TwoLevelCombinators
import Data.Transform.RulesWithRefs






import Data.List as List hiding (all)
import Data.Map as Map
import Data.Set as Set
import Pointless.Combinators
import Control.Monad.State hiding (when)
import Data.Maybe
import Pointless.Isomorphisms as ISO

import Debug.Trace

-- The |ReadP| module containts a library of parsing combinators for building Read instances.
import Text.ParserCombinators.ReadP (skipSpaces,ReadP,readP_to_S,readS_to_P,between,char,sepBy,string,(+++))

gread :: Type a -> ReadM a
gread = readS_to_M . readP_to_S . greadP 10

-- ** FrontEnd class

class FrontEnd t v | t -> v, v -> t where
	parsetype  :: t -> Maybe DynType
	printtype  :: Type a -> Maybe t
	parsevalue :: Type a -> v -> Maybe a
	printvalue :: Type a -> a -> Maybe v

transform :: (FrontEnd t v, FrontEnd t' v') => 
             t -> Rule -> Maybe (t', v -> Maybe v', v' -> Maybe v)
transform t r = do 
  DynT a <- parsetype t
  View (Rep to from) a' <- evalStateT (r a) (initNameSupply a)
  t' <- printtype a'
  let to' v = do {x <- parsevalue a v; printvalue a' (to x)}
      from' v' = do {x <- parsevalue a' v'; printvalue a (from x)}
  return (t', to', from')

transform2 :: (FrontEnd t v) =>
             t -> Rule -> Maybe (t, v -> Maybe v, v -> Maybe v)
transform2 t r = do 
  DynT a <- parsetype t
--  View (Rep to from) a' <- evalStateT (r a) (initNameSupply a)
--  t' <- printtype a'
  let to v = do {x <- parsevalue a v; printvalue a x}
      from v' = do {x <- parsevalue a v'; printvalue a (id x)}
  return (t, to, from)

instance FrontEnd String String where
  parsetype = readM :: ReadM DynType
  printtype = return . show
  parsevalue = gread
  printvalue t = return . gshow t








-- * ReadP Combinators

brackets :: ReadS a -> ReadS a
brackets = readP_to_S . bracketsP . readS_to_P

bracketsP :: ReadP a -> ReadP a
bracketsP = between (char '{') (char '}')

sepcommaP :: ReadP a -> ReadP [a]
sepcommaP = (flip sepBy) (char ',')

sepcomma :: ReadS a -> ReadS [a]
sepcomma = readP_to_S . sepcommaP . readS_to_P

readP :: Read a => ReadP a
readP = readS_to_P reads

infixr 9 ++++

f ++++ g = readP_to_S $ readS_to_P f +++ readS_to_P g

infixr 9 =>>

f =>> g = g >> f

apply :: (a -> b) -> ReadS a -> ReadS b
apply f = readP_to_S . applyP f . readS_to_P

applyP :: (a -> b) -> ReadP a -> ReadP b
applyP f = liftM f

readParenP :: Bool -> ReadP a -> ReadP a
readParenP True p = between (char '(') (char ')') p
readParenP False p =  between (char '(') (char ')') p +++ p








-- * Read Instance for Type

instance (Read k,Read a) => Read (Map k a) where
	readsPrec d = readParen (d > 9) (apply Map.fromDistinctAscList (brackets (sepcomma mapvalue)))
		where
		mapvalue = readP_to_S $ do
			x <- readP
			string ":="
			y <- readP
			return (x,y)

instance (Read a) => Read (Set a) where
	readsPrec d = readParen (d > 9) (apply Set.fromDistinctAscList (brackets (sepcomma reads)))

instance Read DynType where
	readsPrec _ = readP_to_S (readType 10)

readType :: Int -> ReadP DynType
readType d = skipSpaces >> (
	readParenP False (applyP (const $ DynT Int) (skipSpaces >> string "Int" >> skipSpaces))
    +++ readParenP False (applyP (const $ DynT Char) (skipSpaces >> string "Char" >> skipSpaces))
    +++ readParenP False (applyP (const $ DynT Bool) (skipSpaces >> string "Bool" >> skipSpaces))
    +++ readParenP False (applyP (const $ DynT String) (skipSpaces >> string "String" >> skipSpaces))
    +++ readParenP False (applyP (const $ DynT One) (skipSpaces >> string "One" >> skipSpaces))
--    +++ readParenP False (applyP (const $ DynT X) (skipSpaces >> string "<X>" >> skipSpaces))
    +++ readParenP (d < 10) (applyP (applyDynT (DynT . List)) (skipSpaces >> string "List" >> readType 9))
    +++ readParenP (d < 10) (applyP (applyDynT (DynT . Set)) (skipSpaces >> string "Set" >> readType 9))
    +++ readParenP (d < 10) (skipSpaces >> string "Map" >> readMapType 9)
    +++ readParenP (d < 10) (skipSpaces >> string "Prod" >> readProdType 9)
    +++ readParenP (d < 10) (skipSpaces >> string "Ann" >> skipSpaces >> readP >>= readTag 9)
    +++ readParenP (d < 10) (skipSpaces >> string "Mu" >> fail "Mu not implemented") =>> skipSpaces
    )
    where
	readMapType :: Int -> ReadP DynType
	readMapType d = do
		(DynT l) <- readType d
		(DynT r) <- readType d
		return $ DynT $ Map l r
	readProdType :: Int -> ReadP DynType
	readProdType d = do
		(DynT l) <- readType d
		(DynT r) <- readType d
		return $ DynT $ Prod l r
	readTag :: Int -> TypeAnn -> ReadP DynType
	readTag d tag = do
		(DynT t) <- readType d
		return $ DynT $ Ann tag t

type ReadM a = (MonadPlus m) => String -> m a

readM :: (Read a) => ReadM a
readM = readS_to_M reads

readS_to_M :: ReadS a -> ReadM a
readS_to_M p s = case (p s) of
	[(x,"")] -> return x
	otherwise -> mzero


greadP :: Int -> Type a -> ReadP a
greadP _ Int = readP
greadP _ Char = readP
greadP _ Bool = readP
greadP _ String = readP
greadP _ One = readP
-- greadP _ X = fail "<X> type accepts no values"
greadP d (List a) = readParenP (d < 10) (between (char '[') (char ']') (sepcommaP (greadP 9 a)))
greadP d (Set a) = readParenP (d < 10) $ applyP Set.fromDistinctAscList $ bracketsP $ sepcommaP $ greadP 9 a
greadP d (Map a b) = readParenP (d < 10) $ applyP Map.fromDistinctAscList $ bracketsP $ sepcommaP $ do
	x <- greadP 10 a
	string ":="
	y <- greadP 10 b
	return (x,y)
greadP d (Prod a b) = readParenP (d < 10) $ do
	char '('
	l <- greadP 9 a
	char ','
	r <- greadP 9 b
	return (l,r)
greadP _ (Ann _ a) = greadP 9 a
greadP _ (Mu _) = fail "Mu type not supported"











-----------------------------------------------------------------------------
-- * Rules for format evolution

-- If the root is an either, we need to create a Set for it to be converted with the mapping.
either2set :: Rule
either2set = (when isEither addSet) ||| nop
	where 
	isEither (Either _ _) = True
	isEither _ = False
	addSet :: Rule
	addSet t = return $ View addSetrep $ Set t
	addSetrep = Rep {to = Set.singleton, from = head . Set.elems}

-- Removes all possible ones from a type.
removeones :: Rule
removeones = many (once (removeones' ||| rmTagOne))
	where
	removeones' :: Rule
	removeones' (Prod a One) = return $ View rep a
		where rep = Rep {to = fst, from = \x -> (x,())}
	removeones' (Prod One b) = return $ View rep b
		where rep = Rep {to = snd, from = \x -> ((),x)}
	removeones' t = mzero
	
	rmTagOne :: Rule
	rmTagOne (Ann _ One) = return $ View idrep One
	rmTagOne _ = mzero

rmTagInsideEitherOne :: Rule
rmTagInsideEitherOne (Either (Ann l t) One)
	| isBasicType t = return $ View idrep (Either t One)
	| otherwise = mzero
rmTagInsideEitherOne _ = mzero







-- * Rules for Constraints creation evolution

type KeyMap = Map (String,String) KeyId

type KeyRule = forall a . Type a -> NameSupply Maybe (View (Type a),KeyMap)

-- ! Converts a KeyRule to a Rule.
key2rule :: KeyRule -> Rule
key2rule r = liftM fst . r

-- ! Applies a function over Rule to a KeyRule.
keyrule :: (Rule -> Rule) -> KeyRule -> KeyRule
keyrule f kr a = do
	(_,idxs) <- kr a
	vw <- (f (key2rule kr)) a
	return (vw,idxs)

-- ! Composes a KeyRule with a function that consumes a KeyMap and creates a Rule.
keycomprule :: forall a . KeyRule -> (KeyMap -> Rule) -> Rule
keycomprule kr rule a = do
	(View r b,idxs) <- kr a
        (View s c) <- (rule idxs) b
        return $ (View (comprep r s) c)

updatefk :: KeyMap -> String -> [String] -> Rule
updatefk keys s ks t@(List (Ann table prods))
	| isJust tname && (fromJust tname) == s = mzero
	| otherwise = do
		(View rep t') <- once createFk prods
		return $ View (maprep rep) $ List $ Ann table t'
	where
	tname = getFieldName table
	
	createFk :: Rule
	createFk (Ann l _) | isJust $ key l = mzero
	createFk (Ann l a) | isJust name && isJust nk && (fromJust name) `elem` ks = return $ View idrep $ Ann fktag a
		where
		name = getFieldName l
		nk = Map.lookup (s,fromJust name) keys
		fktag = addReference (Just $ "FK" ++ show (fromJust nk)) (fromJust nk) l
	createFk _ = mzero
updatefk _ _ _ _ = mzero

-- Converts a list of a sequence to a map, using the argument list as primary keys.
listsequence2map :: String	-- primary table name
		 -> [String]	-- primary keys
		 -> KeyRule	-- the generated rule
listsequence2map s ks t@(List (Ann table prods))
	| isJust tname && (fromJust tname) == s = mkPks
	| otherwise = mzero
	where
	tname = getFieldName table
	mkPks = do
		(View rep t)   <- many (once Data.Transform.RulesWithRefs.assocr) prods
		(View rep2 t2) <- emptyMap $ List t
		(View rep3 t3,idxs) <- mkMap s ks t2
		(View rep4 t4) <- removeones t3
		return (View (comprep (maprep rep) $ comprep rep2 $ comprep rep3 rep4) $ Ann (mapTag table) t4,idxs)
		
	-- Creates an empty map from a list.
	-- Used to start mkMap function calls.
	emptyMap :: Rule
	emptyMap (List t) = return $ View rep $ Map One t
		where rep = Rep {to = Map.fromDistinctAscList . zip (repeat ()), from = Map.elems}
	
	mapTag tag = tag {fieldName = appendName (getFieldName tag) "Map$"}

	keyTag key tag = tag {fieldName = appendName (getFieldName tag) "Key$", key = Just key, reference = getReference tag}

	appendName (Just x) s = Just $ s ++ x
	appendName Nothing _ = Nothing

	-- recursive function that moves keys to the left side of the map.
	mkMap :: String -> [String] -> KeyRule
	mkMap _ [] x = return (View idrep x,Map.empty)
	mkMap table ks m@(Map l a@(Ann x _))
			| isJust name && name' `elem` ks = do
				nk <- nextKey
				(View rep t2,idxs) <- mkMap table (List.filter (/=name') ks) $ Map (Prod l (alterTag (keyTag nk) a)) One
				return (View (comprep assocMapOneRep rep) t2,Map.insert (table,name') nk idxs)
			| otherwise = do
				(View rep2 (Map x y),idxs) <- mkMap table ks (Map l One)
				return (View (mapProdOneRep rep2) $ Map x (Prod a y),idxs)
		where
		name = getFieldName x
		name' = fromJust name
	mkMap table ks (Map l (Prod a@(Ann x _) b))
			| isJust name && name' `elem` ks = do
				nk <- nextKey
				(View rep t2,idxs) <- mkMap table (List.filter (/=name') ks) $ Map (Prod l (alterTag (keyTag nk) a)) b
				return (View (comprep assocMapRep rep) t2,Map.insert (table,name') nk idxs)
			| otherwise = do
				(View rep2 (Map x y),idxs) <- mkMap table ks (Map l b)
				return (View (mapProdRep rep2) $ Map x (Prod a y),idxs)
		where
		name = getFieldName x
		name' = fromJust name
	mkMap _ _ _ = mzero
	
	-- Value level functions
	assocMapRep :: Rep (Map a (b,c)) (Map (a,b) c)
	assocMapRep = Rep {to = Map.fromDistinctAscList . List.map ISO.assocl . Map.toAscList, from = Map.fromDistinctAscList . List.map ISO.assocr . Map.toAscList}
		
	assocMapOneRep :: Rep (Map a b) (Map (a,b) ())
	assocMapOneRep = Rep {to = Map.fromDistinctAscList . uncurry zip . ((uncurry zip . (Map.keys /\ Map.elems)) /\ (const $ repeat ())), from = Map.fromDistinctAscList . keys}
		
	mapProdRep :: Rep (Map a c) (Map d e) -> Rep (Map a (b,c)) (Map d (b,e))
	mapProdRep r = Rep {to = mapProdRep' (to r), from = mapProdRep' (from r)}
		where	
		mapProdRep' :: (Map a c -> Map d e) -> Map a (b,c) -> Map d (b,e)
		mapProdRep' r m = Map.fromDistinctAscList $ uncurry zip $ (keys /\ ((zip b) . Map.elems)) $ r $ Map.fromDistinctAscList $ (keys m) `zip` (List.map snd $ Map.elems m)
			where
			b = fst $ unzip $ Map.elems m

	mapProdOneRep :: Rep (Map a ()) (Map d e) -> Rep (Map a b) (Map d (b,e))
	mapProdOneRep r = Rep {to = to' (to r), from = from' (from r)}
		where
		to' :: (Map a () -> Map d e) -> Map a b -> Map d (b,e)
		to' r m = Map.fromDistinctAscList $ uncurry zip $ (keys /\ (zip b) . Map.elems) $ r $ Map.fromDistinctAscList $ (keys m) `zip` (repeat ())
			where
			b = Map.elems m
		
		from' :: (Map d e -> Map a ()) -> Map d (b,e) -> Map a b
		from' r m = Map.fromDistinctAscList $ ((flip zip) b) . keys $ r $ Map.fromDistinctAscList $ zip (keys m) (snd $ unzip $ Map.elems m)
			where
			b = fst $ unzip $ Map.elems m

listsequence2map _ _ _ = mzero

-- ! Converts a XML schema sequence to a map, using the argument list as primary keys

createMaps :: String -> [String] -> Rule
createMaps table ks = keyrule once (listsequence2map table ks) `keycomprule` \x -> many (once (updatefk x table ks))



putRefs :: String -> String -> KeyId -> Rule
putRefs x y id t@(Prod a b) = do
  guard (getName a == Just x)
  guard (getName b == Just y)  
  return $ View idrep (addRef (Just $ x++y) id t)
putRefs _ _ _ _ = mzero

list2map :: KeyId -> Rule
list2map id (List (Prod k v)) = do
  return (View (Rep (Map.fromDistinctAscList) (Map.toList)) (Map (putKey id k) v))
list2map id (List (Ann l (Prod k v))) = do
  let l' = emptyTag { fieldName = getFieldName l }
  return (View (Rep (Map.fromDistinctAscList) (Map.toList)) (Ann l' (Map ( putKey id k) v)))
list2map _ _ = mzero

assocl2 :: String -> String -> Rule
assocl2 x y t@(Prod a (Prod b c)) = do
  guard (getName a == Just x)
  guard (getName b == Just y)
  Data.Transform.RulesWithRefs.assocl t
assocl2 _ _ _ = mzero






-- 