{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Type-safe representation of types at the value level, including
-- representation of recursive types as fixpoints of functors.
--
-- See
--
--     Type-safe Two-level Data Transformation
-- 
-- by
--
--     Alcino Cunha, Jose Nuno Oliveira, Joost Visser
--
-----------------------------------------------------------------------------

module Data.Transform.Type where

import Data.Set as Set
import Data.Map as Map
import Data.List as List
import Control.Monad
import Control.Monad.Identity
import Text.Regex
import Data.Maybe

import GHC.Base (unsafeCoerce#)

-----------------------------------------------------------------------------
-- * Representation of Types

data EP a b = EP {to :: a -> b, from :: b -> a}

-- | A type-safe representation of types.
data Type a where

    -- Without fixpoints
    Int :: Type Int
    Bool :: Type Bool
    Char :: Type Char
    String :: Type String
    One :: Type One
    List :: Type a -> Type [a]
    Maybe :: Type a -> Type (Maybe a)           -- << --
    Set :: Type a -> Type (Set a)
    Tree :: Type a -> Type (Tree a) -- used in Data.Transform.Lenses
    Map :: Type a -> Type b -> Type (Map a b)
    Either :: Type a -> Type b -> Type (Either a b)
    Prod :: Type a -> Type b -> Type (a,b)
    Func :: Type a -> Type b -> Type (a -> b)
    Tag :: String -> Type a -> Type a
    Ann :: TypeAnn -> Type a -> Type a          -- << --

    Data :: String -> EP a b -> Type b -> Type a

    -- used to represent XML elements
    EL :: String -> Type a -> Type (El a)
    Tuple :: Tuple a -> Type a
    
    -- TNil :: Type (Tuple ())
    -- TCons :: Type a -> Type (Tuple b) -> Type (Tuple (a,b))
    
    -- Fixpoints
    Mu :: Dist f => Fctr f -> Type (Mu f)
    
    PF :: Type a -> Type (PF a)
    TP :: Type T
    TU :: Type a -> Type (GQ a)
    TUQ :: Type a -> Type (Q a) -- For PAT-Algebra
    Dynamic :: Type Dynamic
    
-- Type annotations
data TypeAnn = TypeAnn {
    fieldName    :: Maybe TagName,
    key        :: Maybe KeyId,
    reference    :: RefType
  } deriving (Show,Eq)
type TagName    = String
type KeyId    = (Int,[Int])
type RefName    = String
type RefType    = [(Maybe RefName, KeyId)]
getFieldName :: TypeAnn -> Maybe TagName
getFieldName t = fieldName t
getKey :: TypeAnn -> Maybe KeyId
getKey t = key t
getReference :: TypeAnn -> RefType
getReference t = reference t


data Tuple a where
    -- TNil :: Tuple ()
    TCons :: Type a -> Tuple b -> Tuple (a,b)
    -- TAppend :: Tuple a -> Tuple b -> Tuple (a,b)
    TSing :: Type a -> Tuple a

-- used in Type GADT above and in Data.Transform.Lenses
data Tree a = Tr (Map.Map a (Tree a))
    deriving (Eq, Show)

--instance Functor Tree where
--    fmap = fmapTree    -- Not possible because of Ord constraint.
-- fmapTree :: Ord b => (a->b) -> Tree a -> Tree b

fmapTree :: Ord b => (a->b) -> Tree a -> Tree b
fmapTree f (Tr m) = Tr (Map.map (fmapTree f) $ Map.mapKeys f m)

-- | Representation of XML elements.
data El a = El a 
  deriving (Eq,Show)



-- | A type-safe representation of lifted functors.
data Fctr f where
    Id    :: Fctr Id
    K     :: Type a -> Fctr (K a)
    (:+:) :: Fctr g -> Fctr h -> Fctr (g:+:h)
    (:*:) :: Fctr g -> Fctr h -> Fctr (g:*:h)
    (:@:) :: Functor g => (forall a . Type a -> Type (g a)) -> Fctr h -> Fctr (g:@:h)

instance Show (Type a) where
    show Int = "Int"
    show Bool = "Bool"
    show Char = "Char"
    show String = "String"
    show One = "One"
    show (List x) = "(List "++(show x)++")"
    show (Maybe x) = "(Maybe "++(show x)++")"
    show (Set x) = "(Set "++(show x)++")"
    show (Tree x) = "(Tree "++(show x)++")"
    show (Map x y) = "(Map "++(show x)++" "++(show y)++")"
    show (Either x y) = "(Either "++(show x)++" "++(show y)++")"
    show (Prod x y) = "(Prod "++(show x)++" "++(show y)++")"
    show (Func x y) = "("++show x++" -> "++show y++")"
    show (Tag n t) = "(Tag "++(show n)++" "++(show t)++")"
    show (Ann l t) = "(Ann "++(show l)++" "++(show t)++")"
    show (EL l a)       = "(EL"++show l++" "++show a++")"
    show (Data s _ _) = s

    show (PF a) = "(PF "++show a++")"
    show (Dynamic) = "Dynamic"
    show (Mu f) = "(Mu "++show f++")"
    
instance Show (Fctr f) where
  show Id = "Id"
  show (K t) = "(K "++show t++")"
  show (f:*:g) = "("++show f++":*:"++show g++")"
  show (f:+:g) = "("++show f++":+:"++show g++")"
  show (f:@:g) = "(<?>:@:"++show g++")"
 
----------------------------------------
-- ** Type One

-- data One
type One = ()

_L :: One
_L = ()

-- instance Show One where
--     show _ = "_L"

----------------------------------------
-- ** Type Dynamic

-- | Dynamic types
data Dynamic where
    Dyn :: Type a -> a -> Dynamic

-----------------------------------------------------------------------------
-- * Deriving type representations from types.

-- | Typeable class of types for which we can derive representations.
class Typeable a where
    typeof :: Type a

instance Typeable Int where
    typeof = Int

instance Typeable () where
    typeof = One

instance Typeable Bool where
    typeof = Bool

instance Typeable Char where
    typeof = Char

instance (Typeable a, Typeable b) => Typeable (a,b) where
{-
    typeof = Data "Prod" (EP to from) (Prod typeof typeof)
	where to (x,y) = (x,y)
	      from (x,y) = (x,y)
-}
    typeof = Prod typeof typeof

instance (Typeable a, Typeable b) => Typeable (Either a b) where
    typeof = Either typeof typeof

instance (Typeable a, Typeable b) => Typeable (Map a b) where
    typeof = Map typeof typeof

instance (Typeable a, Typeable b) => Typeable (a -> b) where
    typeof = Func typeof typeof

instance Typeable a => Typeable (PF a) where
    typeof = PF typeof

instance Typeable Dynamic where
    typeof = Dynamic

instance Typeable String where
    typeof = String

instance Typeable a => Typeable [a] where
    typeof = List typeof

instance Typeable a => Typeable (Set a) where
    typeof = Set typeof

instance Typeable a => Typeable (Tree a) where
    typeof = Tree typeof

instance Typeable T where
    typeof = TP

instance Typeable a => Typeable (GQ a) where
    typeof = TU typeof

instance Typeable a => Typeable (Q a) where
    typeof = TUQ typeof


-----------------------------------------------------------------------------
-- * Type-indexed functions

----------------------------------------
-- ** Some generic functions

-- | Explicitly contrain a value to a particular type.
constrain :: Type a -> a -> a
constrain _ a = a

-- | Explicitly coerce a value of a given type to another given type.
coerce :: MonadPlus m => Type a -> Type b -> a -> m b
coerce a b x = do Eq <- teq a b
                  return x

----------------------------------------
-- ** Equality of types

-- | Data type to express type-equality.    
data Equal a b where
    Eq :: Equal a a
    
-- | Test equality of types.
--   If the types are equal, a witness 'Eq' of the equality is returned.
teq :: MonadPlus m => Type a -> Type b -> m (Equal a b)
teq Dynamic Dynamic = return Eq
teq Int Int = return Eq
teq Bool Bool = return Eq
teq Char Char = return Eq
teq One One = return Eq
teq String String = return Eq
teq (List a) (List b) = 
    do Eq <- teq a b
       return Eq
teq (Set a) (Set b) = 
    do Eq <- teq a b
       return Eq
teq (Maybe a) (Maybe b) =
    do Eq <- teq a b
       return Eq
teq (Map a b) (Map c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Prod a b) (Prod c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Either a b) (Either c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Func a b) (Func c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (PF a) (PF b) = 
    do Eq <- teq a b
       return Eq
-- teq (Tag l a) (Tag l' b )=
--     do guard (l==l')
--        Eq <- teq a b
--        return Eq
teq (Tag _ a) b =
    do Eq <- teq a b
       return Eq
teq a (Tag _ b) =
    do Eq <- teq a b
       return Eq
teq (Ann _ a) b =
    do Eq <- teq a b
       return Eq
teq a (Ann _ b) =
    do Eq <- teq a b
       return Eq
teq (Data s _ a) (Data t _ b) =
    do guard (s==t)
       Eq <- teq a b
       return (unsafeCoerce# Eq)
teq (Tree a) (Tree b) = 
    do Eq <- teq a b
       return Eq
teq (EL l a) (EL l' a') = do
  guard (l==l')
  Eq <- teq a a'
  return Eq
teq TP TP = return Eq
teq (TU a) (TU b) = do
  Eq <- teq a b
  return Eq
teq (Mu f) (Mu g) = do
  error "Cannot determine equality for Mu's"
teq _ _ = mzero

-- | Determine whether a given type occurs in another type.
occursIn :: Type a -> Type b -> Bool
occursIn a b = or (crushType aux b)
  where
    aux :: Type x -> [Bool]
    aux x = maybe [False] (const [True]) (teq a x)

-- Determine whether a given (element) type has content (PCDATA).
hascontent :: Type a -> Bool
hascontent (EL s a) = aux a
  where
    aux :: Type a -> Bool
    aux String        = True
    aux (Prod a b)    = aux a || aux b
    aux (Either a b)  = aux a || aux b
    aux (List a)      = aux a
    aux Int           = False
    aux (EL _ _)      = False
hascontent ta = False

-----------------------------------------------------------------------------
-- * Representation of recursive datatypes.

----------------------------------------
-- ** Standard functors

infixr 5 :+:
infixr 6 :*:
infixr 9 :@:

newtype Id a        = Ident {unIdent :: a}       deriving Eq
newtype K b a       = Const {unConst :: b}       deriving Eq
data (g :+: h) a    = Inl (g a) | Inr (h a)      deriving Eq
data (g :*: h) a    = Pair (g a) (h a)           deriving Eq
newtype (g :@: h) a = Comp {unComp :: g (h a)}   deriving Eq

instance Functor Id where
    fmap f = Ident . f . unIdent

instance Functor (K a) where
    fmap f = Const . id . unConst

instance (Functor f, Functor g) => Functor (f :+: g) where
    fmap f (Inl x) = Inl (fmap f x)
    fmap f (Inr x) = Inr (fmap f x)

instance (Functor f, Functor g) => Functor (f :*: g) where
    fmap f (Pair x y) = Pair (fmap f x) (fmap f y)

instance (Functor g, Functor h) => Functor (g :@: h)
    where fmap f = Comp . fmap (fmap f) . unComp

----------------------------------------
-- ** Fixpoints and recursion patterns

-- | Explicit fixpoint operator
newtype Mu f = In {out :: f (Mu f)}

-- This instance requires -fallow-undecidable-instances.
instance Eq (f (Mu f)) => Eq (Mu f) where
  (In fmf) == (In gmg) = fmf==gmg

-- | Catamorphism or fold
cata :: Functor f => (f a -> a) -> Mu f -> a
cata g = g . fmap (cata g) . out

-- | Anamorphism or unfold
ana :: Functor f => (a -> f a) -> a -> Mu f
ana g = In . fmap (ana g) . g

-- | Natural transformation
nu :: Functor g => (forall a. f a -> g a) -> Mu f -> Mu g
nu t = In . fmap (nu t) . t . out

----------------------------------------
-- ** Functors and monads

-- | Class of functors that distribute over monads.
class Functor f => Dist f where
    dist :: Monad m => f (m a) -> m (f a)

instance Dist Id where
    dist (Ident x) = do y <- x
			return (Ident y)

instance Dist (K a) where
    dist (Const x) = return (Const x)

instance (Dist f, Dist g) => Dist (f :+: g) where
    dist (Inl x) = do y <- dist x
		      return (Inl y)
    dist (Inr x) = do y <- dist x
		      return (Inr y)

instance (Dist f, Dist g) => Dist (f :*: g) where
    dist (Pair l r) = do x <- dist l
			 y <- dist r
			 return (Pair x y)

instance (Dist g, Dist h) => Dist (g :@: h) where
    dist (Comp x) = do y <- return $ fmap dist x
		       z <- dist y
		       return (Comp z)

instance Dist [] where
    dist = sequence

instance Dist Set where
    dist s = sequence (Set.toList s) >>= return . Set.fromDistinctAscList

instance Functor Set where
    fmap = mapMonotonic

-- | Monadic catamorphism.
mcata :: (Monad m, Dist f) => (f a -> m a) -> Mu f -> m a
mcata g = cata (\x -> dist x >>= g)

-----------------------------------------------------------------------------
-- * Strategy combinators on type representations

-- | One-step type-unifying traversal combinator:
allType :: (forall a . Type a -> [b]) -> (forall a . Type a -> [b])
allType f Int                   = []
allType f String                = []
allType f (Prod a b)            = f a ++ f b
allType f (Either a b)          = f a ++ f b
allType f (List a)              = f a
allType f (EL l a)              = f a

-- | Type-unifying complete traversal.
crushType :: (forall a . Type a -> [b]) -> (forall a . Type a -> [b])
crushType f a = f a ++ allType (crushType f) a

-- | Type-unifying traversal that does not go below successful nodes.
crushTypeStop :: (forall a . Type a -> [b]) -> (forall a . Type a -> [b])
crushTypeStop f a = if List.null atRoot then belowRoot else atRoot
  where
    atRoot = f a
    belowRoot = allType (crushTypeStop f) a

-----------------------------------------------------------------------------
-- * Type-safe representation of point-free functions.

data Monoid r = Monoid {zero :: r, plus :: r -> r -> r}

class Result r where
    monoid :: Monoid r

mint = Monoid {zero = 0, plus = (+)}

instance Result Int where
    monoid = mint

mlist = Monoid {zero = [], plus = (++)}

instance Result [a] where
    monoid = mlist

mbool = Monoid {zero = False, plus = (||)}

instance Result Bool where
    monoid = mbool

-- | GADT to represent (point-free) functions.
data PF a where
    -- Point-free functions
    HOLE :: PF a
    FUN :: String -> (a -> b) -> PF (a -> b)
    BANG :: PF (a -> One)
    PNT :: a -> PF (One -> a)
    ID :: PF (a -> a)
    COMP :: Type b -> PF (b -> c) -> PF (a -> b) -> PF (a -> c)
    FST :: PF ((a,b) -> a)
    SND :: PF ((a,b) -> b)
    SPLIT :: PF (a -> b) -> PF (a -> c) -> PF (a -> (b,c))
    PROD :: PF (a -> b) -> PF (c -> d) -> PF ((a,c) -> (b,d))
    INL :: PF (a -> Either a b)
    INR :: PF (b -> Either a b)
    EITHER :: PF (a -> c) -> PF (b -> c) -> PF (Either a b -> c)
    SUM :: PF (a -> b) -> PF (c -> d) -> PF (Either a c -> Either b d)
    AP :: PF ((a -> b, a) -> b)
    CURRY :: PF ((a,b) -> c) -> PF (a -> (b -> c))
    EXP :: PF (a -> b) -> PF ((c -> a) -> (c -> b))
    NIL :: PF (() -> [a])
    CONS :: PF ((a,[a]) -> [a])
    LIST :: PF (a -> b) -> PF ([a] -> [b])
    WRAP :: PF (a -> [a])
    APPEND :: PF (([a],[a]) -> [a])
    LIST2MAP :: PF ([a] -> Map Int a)
--    FOLDL :: PF ((a,b) -> a) -> PF ((a,[b]) -> a)
    CATA :: PF (Either () (a,b) -> b) -> PF ([a] -> b)
    FOLD :: Monoid a -> PF ([a] -> a)
    FILTER :: PF (a -> Bool) -> PF ([a] -> [a])
    EMPTY :: PF (One -> Map a b)
    INSERT :: PF (((a,b),Map a b) -> Map a b)
    MAP :: PF (b -> c) -> PF (Map a b -> Map a c)
    MAPKEYS :: PF (a -> b) -> PF (Map a c -> Map b c)
    ELEMS :: PF (Map a b -> [b])
    LJOIN :: PF ((Map a b, a -> c) -> Map a (b,c))
    SET :: PF (a -> b) -> PF (Set a -> Set b)
    TODYN :: PF (a -> Dynamic)

    ELMAP :: PF (a->b) -> PF (El a -> El b)

    DATAMAP :: Type b -> PF (b->b) -> PF (a->a)
    OUT  :: PF (a -> b)

    ZERO :: Monoid a -> PF (b -> a)
    PLUS :: Monoid a -> PF ((a,a) -> a)
    TRUE :: PF (a -> Bool)
    COND :: PF (a -> Bool) -> PF (a -> b) -> PF (a -> b) -> PF (a -> b)

    REC    :: (PF f -> PF f) -> PF f

    -- Type-preserving strategy combinators
    APPLY  :: Type a -> PF T -> PF (a->a)
    ADHOC  :: Type a -> PF (a->a) -> PF T -> PF T
    MKT    :: Type a -> PF (a->a) -> PF T
    NOP    :: PF T
    SEQ    :: PF T -> PF T -> PF T
    ALL    :: PF T -> PF T
    EVERYWHERE :: PF T -> PF T

    -- Type-unifying strategy combinators
    GMAPQ      :: Monoid r -> PF (GQ r) -> PF (GQ r)
    MKQ        :: Monoid r -> Type a -> PF (a -> r) -> PF (GQ r)
    EVERYTHING :: Monoid r -> PF (GQ r) -> PF (GQ r)
    EMPTYQ     :: Monoid r -> PF (GQ r)
    UNION      :: Monoid r -> PF (GQ r) -> PF (GQ r) -> PF (GQ r)
    APPLYQ     :: Type a -> PF (GQ r) -> PF (a->r)

    -- XPath
    SELF :: PF (GQ [Dynamic])
    CHILD :: PF (GQ [Dynamic])
    DESCENDANT :: PF (GQ [Dynamic])
    DESCorSELF :: PF (GQ [Dynamic])
    NAME :: String -> PF (GQ [Dynamic])
    (:/:)  :: PF (GQ [Dynamic]) -> PF (GQ r) -> PF (GQ r)
    (:?:)  :: PF (GQ [Dynamic]) -> PF (GQ Bool) -> PF (GQ [Dynamic])
    QCOMP  :: Type a -> PF (GQ a) -> PF (a -> b) -> PF (GQ b)
    (:/\:) :: PF (GQ a) -> PF (GQ b) -> PF (GQ (a,b))
    NONEMPTY :: PF (GQ Bool)


    -- PAT-algebra expressions
    Empty           :: PF (Q x)
    Etn             :: String -> Type x -> PF (Q (El x))
    Union           :: PF (Q x) -> PF (Q x) -> PF (Q x)
    Intersect       :: PF (Q x) -> PF (Q x) -> PF (Q x)
    Difference      :: PF (Q x) -> PF (Q x) -> PF (Q x)
    ContainedIn     :: Type x -> Type y -> PF (Q x) -> PF (Q y) -> PF (Q x)
    Contains        :: Type x -> Type y -> PF (Q x) -> PF (Q y) -> PF (Q x)
    MatchContent    :: Type x -> (String->Bool) -> PF (Q x) -> PF (Q x)
    MatchAttribute  :: (String->Bool) -> PF (Q x) -> PF (Q x)
    IndexDown       :: Type y -> Map x [y] -> PF (Q x) -> PF (Q y)
    IndexUp         :: Type y -> Map x y   -> PF (Q x) -> PF (Q y)

    ALLQ    :: PF (Q b) -> PF (Q b)
    AtTag   :: String -> Type x -> PF (Q x)


    -- Generalization for T and Q (forget for now)
    -- SEQG   :: PF T' -> PF (G c) -> PF (G c) 
    
-- Smart constructors
pnt x = COMP One (PNT x) BANG

wrap :: Type a -> PF (a -> [a])
wrap a = COMP (Prod a (List a)) CONS (SPLIT ID (COMP One NIL BANG))

comp f g = COMP typeof f g

etn :: Type (El a) -> PF (Q (El a))
etn (EL l a)      = Etn l a

contains f g      = Contains (qrestype f) (qrestype g) f g
containedIn f g   = ContainedIn (qrestype f) (qrestype g) f g

match s f         = MatchContent (qrestype f) p f
  where
    p = maybe False (const True) . (matchRegex $ mkRegex s)
    
topdown f = REC (\self -> SEQ f (ALL self))

collect f = REC (\self -> Union f (ALLQ self))

crush :: Result r => PF (GQ r) -> PF (GQ r)
crush = EVERYTHING monoid

mkQ :: (Result r, Typeable a) => PF (a -> r) -> PF (GQ r)
mkQ f = MKQ monoid typeof f

infix 5 |>

(|>) :: Typeable a => PF (GQ a) -> PF (a -> b) -> PF (GQ b)
q |> f = QCOMP typeof q f

--

qrestype :: PF (Q b) -> Type b
qrestype Empty                  = error "7"   -- Perhaps we need a representation of \emph{any} type?
qrestype (Union f g)            = qrestype f
qrestype (Etn s t)              = EL s t
qrestype (Contains x y f g)     = x
qrestype (ContainedIn x y f g)  = x
qrestype (MatchContent x p f)   = x
qrestype f = error $ "qrestype missing pattern for: "-- ++show f

data T    where T :: (forall a . Type a -> a -> a) -> T
data Q b  where Q :: (forall a . Type a -> a -> [b]) -> Q b
unT (T t) = t
unQ (Q q) = q

newtype GQ b = GQ {unGQ :: GenericQ b}

data G c where G :: (forall a . Type a -> a -> c a) -> G c
type T' = G Identity
type Q' = G []

-- Evaluation of point-free terms VERY INCOMPLETE!

eval :: Type a -> PF a -> a
-- Point-free functions
eval _ (FUN n f) = f
eval _ BANG = const _L
eval _ (PNT x) = const x
eval _ ID = id
eval (Func a c) (COMP b f g) = eval (Func b c) f . eval (Func a b) g
eval _ FST = fst
eval _ SND = snd
eval (Func a (Prod b c)) (SPLIT f g) = \x -> (eval (Func a b) f x, eval (Func a c) g x)
eval (Func (Prod a b) (Prod c d)) (PROD f g) = \(x,y) -> (eval (Func a c) f x, eval (Func b d) g y)
eval _ INL = Left
eval _ INR = Right
eval (Func (Either a b) c) (EITHER f g) = either (eval (Func a c) f) (eval (Func b c) g)
eval (Func (Either a b) (Either c d)) (SUM f g) = either (Left . eval (Func a c) f) (Right . eval (Func b d) g)
eval _ AP = uncurry ($)
eval (Func a (Func b c)) (CURRY f) = curry (eval (Func (Prod a b) c) f)
eval (Func (Func c a) (Func _ b)) (EXP f) = (.) (eval (Func a b) f) 
eval _ NIL = const []
eval _ CONS = uncurry (:)
eval (Func (List a) (List b)) (LIST f) = Prelude.map (eval (Func a b) f)
eval _ WRAP = (:[])
--eval _ (FOLDL f) = uncurry $ foldl (curry (eval _ f))
eval (Func (List a) b) (CATA f) = 
    let g = eval (Func (Either One (Prod a b)) b) f
    in foldr (curry (g . Right)) (g (Left ()))
eval _ (FOLD t) = foldl (plus t) (zero t)
eval (Func (List a) _) (FILTER f) = List.filter (eval (Func a Bool) f)
eval _ LIST2MAP = Map.fromDistinctAscList . (zip [0..])
eval _ EMPTY = const Map.empty
eval _ INSERT = undefined
eval (Func (Map _ b) (Map _ c)) (MAP f) = Map.map (eval (Func b c) f)
eval (Func (Map a _) (Map b _)) (MAPKEYS f) = Map.mapKeysMonotonic (eval (Func a b) f) -- Wrong!!
eval _ ELEMS = Map.elems
eval _ LJOIN = undefined
eval (Func (Set a) (Set b)) (SET f) = Set.mapMonotonic (eval (Func a b) f) -- Wrong!!
eval (Func a _) (TODYN) = \x -> Dyn a x

eval (Func (EL _ a) (EL _ b)) (ELMAP f) = \(El x) -> El (eval (Func a b) f x)

eval (Func (Data _ ep a) _) (DATAMAP b f) = case teq' a b of Just Eq -> from ep . eval (Func b b) f . to ep
eval (Func (Data _ ep a) b) OUT = case teq' a b of Just Eq -> to ep

eval _ TRUE = \z -> True
eval _ (ZERO m) = \x -> zero m
eval _ (PLUS m) = uncurry (plus m)
eval (Func a b) (COND p f g) = \x -> if (eval (Func a Bool) p x) then eval (Func a b) f x else eval (Func a b) g x

-- Type-preserving strategic function combinators
eval _ (APPLY t s)  = case eval TP s of (T f) -> f t
eval _ (MKT t f)    = T (mkT t (eval (Func t t) f))
eval _ NOP          = T (const id)
eval _ (SEQ f g)    = T (seqT (unT $ eval TP f) (unT $ eval TP g))
eval _ (ALL f)      = T (allT (unT $ eval TP f))
eval _ (EVERYWHERE f) = eval TP (topdown f)

eval a (REC (ff::PF f -> PF f))     = eval a aux
  where
    aux :: PF f
    aux = ff aux

eval (TU r) (GMAPQ a f)        = GQ (gmapQ' (plus a) (zero a) (unGQ $ eval (TU r) f))
eval (TU r) (MKQ a t f)      = GQ (_mkQ t (zero a) (eval (Func t r) f))
eval (TU r) (EVERYTHING a f) = GQ (everything' (plus a) (unGQ $ eval (TU r) f))
eval (Func _ r) (APPLYQ t f)     = unGQ (eval (TU r) f) t
eval _ (EMPTYQ t) = GQ (emptyq (zero t))
eval (TU r) (UNION t f g) = GQ (unionq (plus t) (unGQ $ eval (TU r) f) (unGQ $ eval (TU r) g))

eval _ SELF = GQ aux
    where aux :: GenericQ [Dynamic]
	  aux Dynamic (Dyn a x) = aux a x
	  aux a x = [Dyn a x]
eval _ CHILD = eval (TU (List Dynamic)) (GMAPQ mlist SELF)
eval _ DESCENDANT = eval (TU (List Dynamic)) (EVERYTHING mlist (GMAPQ mlist SELF))
eval _ DESCorSELF = eval (TU (List Dynamic)) (UNION mlist SELF DESCENDANT)
eval _ (NAME n) = GQ (aux n)
    where aux :: String -> GenericQ [Dynamic]
	  aux s Dynamic (Dyn a x) = aux s a x
	  aux s a@(Data l _ _) x | s==l = [Dyn a x]
	  aux _ _ _ = []
eval (TU (List a)) (f :/: g) = GQ (\t x -> List.concatMap (unGQ (eval (TU (List a)) g) Dynamic) $ unGQ (eval (TU (List Dynamic)) f) t x)	-- 
eval (TU Bool) (f :/: g) = GQ (\t x -> List.or $ List.map (unGQ (eval (TU Bool) g) Dynamic) $ unGQ (eval (TU (List Dynamic)) f) t x)	-- 
eval _ NONEMPTY = GQ (\t x -> True)
eval _ (f :?: p) = GQ (\t x -> let y = unGQ (eval (TU (List Dynamic)) f) t x
		               in List.filter (unGQ (eval (TU Bool) p) Dynamic) y)
eval (TU b) (QCOMP a q f) = GQ (\t x -> eval (Func a b) f $ unGQ (eval (TU a) q) t x)
eval (TU (Prod a b)) (f :/\: g) = GQ (\t x -> (unGQ (eval (TU a) f) t x,unGQ (eval (TU b) g) t x))

--eval _ (APPLYQ t s)  = case eval _ s of (Q f) -> f t
{-
eval _ (Union f g)       = Q ((unQ $ eval _ f) `unionQ` (unQ $ eval _ g))
eval _ (Intersect f g)   = Q ((unQ $ eval _ f) `intersectQ` (unQ $ eval _ g))

-- PAT-algebra expressions
eval _ (Etn l a) = Q (_crush (q (EL l a)))
  where
    q :: Type b -> Type a -> a -> [b]
    q a b x = case teq' a b of
      Just Eq    -> [x]
      _          -> []
eval _ (ContainedIn _ ty (f::PF (Q b)) g) = Q q
  where
    q :: forall a . Type a -> a -> [b]
    q tx x = [ z | y <- (unQ $ eval _ g) tx x, z <- (unQ $ eval _ f) ty y ]
eval _ (Contains tz _ (f::PF (Q b)) g) = Q q
  where
    q :: forall a . Type a -> a -> [b]
    q tx x = [ z | z <- (unQ $ eval _ f) tx x, y <- (unQ $ eval _ g) tz z ]
eval _ (MatchContent (t::Type b) p f) = Q q
  where
    q :: forall a . Type a -> a -> [b]
    q ta a = [ b | b <- (unQ $ eval _ f) ta a, p (contentof t b) ]
-}
eval _ f = error $ "eval missing pattern for : " -- ++show f

-- Determine whether a given XML element type has text content (PCDATA, i.e. String)
-- as immediate child.
contentof :: Type a -> a -> String
contentof (EL s a) (El x) = aux a x
  where
    aux :: Type a -> a -> String
    aux String x                = x
    aux (Prod a b) (x,y)        = aux a x ++ aux b y
    aux (Either a b) (Left x)   = aux a x
    aux (Either a b) (Right y)  = aux b y
    -- aux (Maybe a) x             = maybe "" (aux a) x
    aux (List a) xs             = concatMap (aux a) xs
    aux (EL _ _) _              = ""
    aux Int _                   = ""
-- contentof ta a = trace ("NO CONTENT: "++show ta++" ") ""


-- Equality of functions (of the same type) -- VERY INCOMPLETE -- geq should be used instead!!
{-
feq :: PF f -> PF f -> Bool
feq ID ID  = True
feq (COMP t f g) (COMP t' f' g') 
           = maybe False (const True) $ do
                Eq <- teq t t'
                return $ feq f f' && feq g g'
feq (ContainedIn x y f g) (ContainedIn x' y' f' g') 
           = maybe False (const True) $ do
                Eq <- teq x x'; Eq <- teq y y'
                return $ feq f f' && feq g g'
feq (Etn s x) (Etn s' x')
           = maybe False (const True) $ do
                Eq <- teq x x'
                return $ s==s'
-}
{-
feq f g    = if (show f == show g)
               then error $ "PANICK -- feq incomplete -- "++show f
               else False
-}
{-
instance Eq (PF f) where
  (==) = feq
-}

-----------------------------------------------------------------------------
-- Strategic programming (type-indexed SYB)

mkT :: Type x -> (x->x) -> (forall a . Type a -> a -> a)
mkT t f t' = maybe id id (do Eq <- teq t t'; return f)

teq' :: Type a -> Type b -> Maybe (Equal a b)
teq' = teq

allT :: (forall a . Type a -> a -> a) -> (forall a . Type a -> a -> a)
allT f Int x                   = x
allT f String x                = x
allT f (Prod a b) (x,y)        = (f a x,f b y)
allT f (Either a b) (Left x)   = Left (f a x)
allT f (Either a b) (Right y)  = Right (f b y)
--allT f (Maybe a) x             = maybe Nothing (Just . (f a)) x
allT f (List a) xs             = List.map (f a) xs
allT f (Tag s a) x             = allT f a x
allT f (EL _ a) (El x)           = El (f a x)
allT f (Data s ep b) x = from ep (f b (to ep x)) 
allT f t x = error $ "allT missing pattern for "++show t

seqT :: (forall a . Type a -> a -> a) -> (forall a . Type a -> a -> a) -> (forall a . Type a -> a -> a) 
seqT f g a x  = g a (f a x)

seqQ :: (forall a . Type a -> a -> a) -> (forall a . Type a -> a -> x) -> (forall a . Type a -> a -> x) 
seqQ f g a x  = g a (f a x)

seqG :: (forall a . Type a -> a -> Identity a)
     -> (forall a . Type a -> a -> c a)
     -> (forall a . Type a -> a -> c a)
seqG f g a x = g a (runIdentity $ f a x)

letG :: (forall a . Type a -> a -> x)
     -> (x -> (forall a . Type a -> a -> c a))
     -> (forall a . Type a -> a -> c a)
letG f g a x = g (f a x) a x

letQ :: (forall a . Type a -> a -> [b])
     -> (b -> (forall a . Type a -> a -> [c]))
     -> (forall a . Type a -> a -> [c])
letQ f g a x = concatMap (\b -> g b a x) (f a x)

unionQ :: (forall a . Type a -> a -> [b]) 
       -> (forall a . Type a -> a -> [b]) 
       -> (forall a . Type a -> a -> [b])
unionQ f g a x = f a x ++ g a x

intersectQ :: (forall a . Type a -> a -> [b]) 
       -> (forall a . Type a -> a -> [b]) 
       -> (forall a . Type a -> a -> [b])
intersectQ f g a x = error "Need Eq" -- f a x `List.intersect` g a x

type GenericQ r = forall a . Type a -> a -> r

emptyq :: r -> GenericQ r
emptyq x a y = x

unionq :: (r -> r -> r) -> GenericQ r -> GenericQ r -> GenericQ r
unionq o f g a x = o (f a x) (g a x)

_mkQ :: Type a -> r -> (a -> r) -> GenericQ r
_mkQ a x f b y = case (teq' a b) of 
  Just Eq -> (f y)
  otherwise -> x

gmapQ' :: (r -> r -> r) -> r -> GenericQ r -> GenericQ r
gmapQ' o z f a = (foldl o z) . (gmapQ f a)
		   
gmapQ :: GenericQ u -> GenericQ [u]
gmapQ f Dynamic (Dyn a x)      = gmapQ f a x 
gmapQ f (Prod a b) (x,y)       = [f a x, f b y]
gmapQ f (Either a b) (Left x)  = [f a x]
gmapQ f (Either a b) (Right x) = [f b x]
--gmapQ f (EL _ a) (El x)        = [f a x]
gmapQ f (List a) l             = List.map (f a) l
gmapQ f (Data _ ep b) x        = gmapQN f b (to ep x)
gmapQ f Int _                  = []
gmapQ f String _               = []
gmapQ f One _                  = []

gmapQN :: GenericQ u -> GenericQ [u]
gmapQN f (Prod a b) (x,y)       = concat [gmapQN f a x, gmapQN f b y]
gmapQN f (Either a b) (Left x)  = concat [gmapQN f a x]
gmapQN f (Either a b) (Right x) = concat [gmapQN f b x]
gmapQN f (List a) l             = concat $ List.map (gmapQN f a) l
gmapQN f a x                    = [f a x]

everything' :: (r -> r -> r) -> GenericQ r -> GenericQ r
everything' o f a x = gmapQ' o (f a x) (everything' o f) a x

everything :: (r -> r -> r) -> GenericQ r -> GenericQ r
everything k f a x = foldl k (f a x) (gmapQ (everything k f) a x)


allQ :: (forall a . Type a -> a -> [b]) -> (forall a . Type a -> a -> [b])
allQ f Int x                   = []
allQ f String x                = []
allQ f (Prod a b) (x,y)        = f a x ++ f b y
allQ f (Either a b) (Left x)   = f a x
allQ f (Either a b) (Right y)  = f b y
-- allQ f (Maybe a) x             = maybe [] (f a) x
allQ f (List a) xs             = concatMap (f a) xs
allQ f (EL l a) (El x)         = allQ f a x -- Should it descend in this case?

_crush :: (forall a . Type a -> a -> [b]) -> (forall a . Type a -> a -> [b])
_crush r a x = r a x ++ allQ (_crush r) a x

recT :: ((forall a . Type a -> a -> a) -> (forall a . Type a -> a -> a))
     -> (forall a . Type a -> a -> a)
recT f a x = f' a x
  where
    f' :: forall a . Type a -> a -> a
    f' = f f'  
    
td :: (forall a . Type a -> a -> a) -> (forall a . Type a -> a -> a)
td s a x = recT (\f -> s `seqT` (allT f)) a x
-- td s = fix aux
--   where
--     aux :: ((forall a . Type a -> a -> a) -> (forall a . Type a -> a -> a))
--     aux f = s `seqT` (allT f)

-----------------------------------------------------------------------------
-- * Generic programming a la SYB revolutions


-- | The type of spines
data Typed a = Type a :| a

data Spine a where
    As :: a -> Con -> Spine a
    Ap :: Spine (a -> b) -> Typed a -> Spine b

data Fixity = Prefix | Infix deriving Eq

data Con = Con {name :: String, fixity :: Fixity}

-- | Converting from a spine to a value
fromSpine :: Spine a -> a
fromSpine (c `As` _) = c
fromSpine (Ap f (_ :| a)) = (fromSpine f) a

dyncon = Con {name = "Dyn", fixity = Prefix}
intcon n = Con {name = show n, fixity = Prefix}
onecon = Con {name = "_L", fixity = Prefix}
holecon = Con {name = "_", fixity = Prefix}
bangcon = Con {name = "bang", fixity = Prefix}
idcon = Con {name = "id", fixity = Prefix}
compcon = Con {name = ".", fixity = Infix}
fstcon = Con {name = "fst", fixity = Prefix}
sndcon = Con {name = "snd", fixity = Prefix}
splitcon = Con {name = "/\\", fixity = Infix}
prodcon = Con {name = "><", fixity = Infix}
inlcon = Con {name = "inl", fixity = Prefix}
inrcon = Con {name = "inr", fixity = Prefix}
eithercon = Con {name = "\\/", fixity = Infix}
sumcon = Con {name = "-|-", fixity = Infix}
apcon = Con {name = "ap", fixity = Prefix}
currycon = Con {name = "curry", fixity = Prefix}
expcon = Con {name = "exp", fixity = Prefix}
varcon n = Con {name = n, fixity = Prefix}
pntcon = Con{name = "pnt", fixity = Prefix}

nilcon = Con{name="nil", fixity=Prefix}
conscon = Con{name="cons", fixity=Prefix}
listelimcon = Con{name="listelim", fixity=Prefix}
listfoldcon = Con{name="listfold", fixity=Prefix}
listcon = Con{name="map", fixity=Prefix}
list2mapcon = Con{name="list2map", fixity=Prefix}

emptycon = Con{name="empty", fixity=Prefix}
insertcon = Con{name="insert", fixity=Prefix}
mapfoldcon = Con{name="mapfold", fixity=Prefix}
mapfoldccon = Con{name="mapfoldc", fixity=Prefix}
elemscon = Con{name="elems", fixity=Prefix}
ljoincon = Con{name="ljoin", fixity=Prefix}
mapcon = Con{name="map", fixity=Prefix}

somelist = Con{name="[...]",fixity=Prefix}

-- | Converting from a value to a spine
toSpine :: Type a -> a -> Spine a
toSpine Dynamic (Dyn t x) = Ap ((Dyn t) `As` dyncon) (t :| x)
toSpine Int n = n `As` (intcon n)
toSpine Bool n = n `As` (intcon n)
toSpine Char n = n `As` (intcon n)
toSpine String n = n `As` (intcon n)
toSpine One x = x `As` bangcon
toSpine (List a) [] = [] `As` Con{name="[]", fixity=Prefix}
toSpine (List a) (h:t) = Ap (Ap ((:) `As` Con{name=":", fixity=Infix}) (a :| h)) (List a :| t)
toSpine (Either a _) (Left x) = Ap (Left `As` Con{name="Left", fixity=Prefix}) (a :| x)
toSpine (Either _ b) (Right y) = Ap (Right `As` Con{name="Right", fixity=Prefix}) (b :| y)
toSpine (Prod a b) (x,y) = (,) `As` Con{name=",",fixity=Infix} `Ap` (a :| x) `Ap` (b :| y)
toSpine (EL _ a) (El x) = (El `As` Con{name="El",fixity=Prefix}) `Ap` (a :| x)
toSpine (Tag _ a) x = toSpine a x
toSpine (Data s ep b) x = (from ep `As` Con {name="in"++s,fixity=Prefix}) `Ap` (b :| to ep x) 
toSpine (PF _) HOLE = HOLE `As` holecon
toSpine (PF _) BANG = BANG `As` bangcon
toSpine (PF (Func a b)) (FUN n f) = (FUN n f) `As` (varcon n)
toSpine (PF (Func one a)) (PNT x) = PNT x `As` Con{name="PNT "++gshow a x, fixity=Prefix}
toSpine (PF _) ID = ID `As` idcon
toSpine (PF (Func a c)) (COMP b f g) = Ap (Ap ((COMP b) `As` compcon) (PF (Func b c) :| f)) (PF (Func a b) :| g)
toSpine (PF _) FST = FST `As` fstcon
toSpine (PF _) SND = SND `As` sndcon
toSpine (PF (Func a (Prod b c))) (SPLIT f g) = Ap (Ap (SPLIT `As` splitcon) (PF (Func a b) :| f)) (PF (Func a c) :| g)
toSpine (PF (Func (Prod a c) (Prod b d))) (PROD f g) = Ap (Ap (PROD `As` prodcon) (PF (Func a b) :| f)) (PF (Func c d) :| g)
toSpine (PF _) INL = INL `As` inlcon
toSpine (PF _) INR = INR `As` inrcon
toSpine (PF (Func (Either a b) c)) (EITHER f g) = Ap (Ap (EITHER `As` eithercon) (PF (Func a c) :| f)) (PF (Func b c) :| g)
toSpine (PF (Func (Either a c) (Either b d))) (SUM f g) = Ap (Ap (SUM `As` sumcon) (PF (Func a b) :| f)) (PF (Func c d) :| g)
toSpine (PF _) AP = AP `As` apcon
toSpine (PF (Func a (Func b c))) (CURRY f) = Ap (CURRY `As` currycon) (PF (Func (Prod a b) c) :| f) 
toSpine (PF (Func (Func _ a) (Func _ b))) (EXP f) = Ap (EXP `As` expcon) (PF (Func a b) :| f) 
toSpine (PF _) NIL = NIL `As` nilcon 
toSpine (PF _) CONS = CONS `As` conscon 
toSpine (PF (Func (List a) (List b))) (LIST f) = Ap (LIST `As` listcon) (PF (Func a b) :| f)
toSpine (PF _) WRAP = WRAP `As` Con{name="wrap",fixity=Prefix}
toSpine (PF (Func (List a) b)) (CATA f) = Ap (CATA `As` Con{name="cata",fixity=Prefix}) (PF (Func (Either One (Prod a b)) b) :| f)
toSpine (PF (Func _ Int)) (FOLD t) = FOLD t `As` Con{name="sum",fixity=Prefix}
toSpine (PF (Func _ Bool)) (FOLD t) = FOLD t `As` Con{name="listor",fixity=Prefix}
toSpine (PF (Func _ (List _))) (FOLD t) = FOLD t `As` Con{name="concat",fixity=Prefix}
toSpine (PF (Func (List a) _)) (FILTER f) = Ap (FILTER `As` Con{name="filter",fixity=Prefix}) (PF (Func a Bool) :| f)
toSpine (PF _) LIST2MAP = LIST2MAP `As` list2mapcon
toSpine (PF _) EMPTY = EMPTY `As` emptycon 
toSpine (PF _) INSERT = INSERT `As` insertcon 
toSpine (PF _) ELEMS = ELEMS `As` elemscon
toSpine (PF _) LJOIN = LJOIN `As` ljoincon
toSpine (PF _) (SET f) = error $ "toSpine missing case for SET"
toSpine (PF (Func a _)) (TODYN) = TODYN `As` Con{name="(mkDyn "++show a++")", fixity=Prefix}
toSpine (PF (Func (Map a b) (Map _ c))) (MAP f) = Ap (MAP `As` mapcon) (PF (Func b c) :| f)
toSpine (PF (Func (EL _ a) (EL _ b))) (ELMAP f) = Ap (ELMAP `As` Con{name="ELMAP",fixity=Prefix}) (PF (Func a b) :| f)
toSpine (PF (Func (Data s ep b) _)) (DATAMAP c f) = 
    case teq' b c of Just Eq -> Ap ((DATAMAP c) `As` Con{name=s,fixity=Prefix}) (PF (Func b b) :| f) 
toSpine (PF (Func (Data s ep a) b)) OUT = case teq a b of Just Eq -> OUT `As` Con{name="un"++s,fixity=Prefix}
toSpine (PF _) TRUE = TRUE `As` Con{name="true",fixity=Prefix}
toSpine (PF (Func _ Int)) (ZERO a) = ZERO a `As` Con{name="zero",fixity=Prefix}
toSpine (PF (Func _ Int)) (PLUS a) = PLUS a `As` Con{name="plus",fixity=Prefix}
toSpine (PF (Func _ Bool)) (ZERO a) = ZERO a `As` Con{name="false",fixity=Prefix}
toSpine (PF (Func _ Bool)) (PLUS a) = PLUS a `As` Con{name="or",fixity=Prefix}
toSpine (PF (Func _ (List _))) (ZERO a) = ZERO a `As` Con{name="nil",fixity=Prefix}
toSpine (PF (Func _ (List _))) (PLUS a) = PLUS a `As` Con{name="cat",fixity=Prefix}
toSpine (PF (Func a b)) (COND p f g) = Ap (Ap (Ap (COND `As` Con{name="cond",fixity=Prefix}) (PF (Func a Bool) :| p)) (PF (Func a b) :| f)) (PF (Func a b) :| g)
-- Type-preserving strategy combinators
toSpine (PF _) (APPLY a f) = (APPLY a `As` Con{name="applyT "++(show a),fixity=Prefix}) `Ap` (PF TP :| f)
toSpine (PF _) (MKT a f) = (MKT a `As` Con{name="mkT "++(show a),fixity=Prefix}) `Ap` (PF (Func a a) :| f)
toSpine (PF _) NOP = NOP `As` Con{name="nop",fixity=Prefix}
toSpine (PF _) (SEQ f g) = ((SEQ `As` Con{name="seq",fixity=Prefix}) `Ap` (PF TP :| f)) `Ap` (PF TP :| g)
toSpine (PF _) (ALL f) = (ALL `As` Con{name="gmapT",fixity=Prefix}) `Ap` (PF TP :| f) 
toSpine (PF _) (EVERYWHERE f) = (EVERYWHERE `As` Con{name="everywhere",fixity=Prefix}) `Ap` (PF TP :| f) 
toSpine (PF _) (REC f) = REC f `As` Con{name="REC ..",fixity=Prefix}
toSpine (PF (TU r)) (GMAPQ a f) = (GMAPQ a `As` Con{name="gmapQ",fixity=Prefix}) `Ap` (PF (TU r) :| f)
toSpine (PF (TU r)) (MKQ x a f) = (MKQ x a `As` Con{name="mkQ",fixity=Prefix}) `Ap` (PF (Func a r) :| f)
toSpine (PF (TU r)) (EVERYTHING x f) = (EVERYTHING x `As` Con{name="everything",fixity=Prefix}) `Ap` (PF (TU r) :| f)
toSpine (PF (Func _ r)) (APPLYQ a f) = (APPLYQ a `As` Con{name="applyQ "++show a,fixity=Prefix}) `Ap` (PF (TU r) :| f)
toSpine (PF _) (EMPTYQ a) = EMPTYQ a `As` Con{name="empty",fixity=Prefix}
toSpine (PF (TU r)) (UNION m f q) = ((UNION m `As` Con{name="union",fixity=Prefix}) `Ap` (PF (TU r) :| f)) `Ap` (PF (TU r) :| q)
--XPath
toSpine (PF _) SELF = SELF `As` Con {name = "self", fixity = Prefix}
toSpine (PF _) CHILD = CHILD `As` Con {name = "child", fixity = Prefix}
toSpine (PF _) DESCENDANT = DESCENDANT `As` Con {name = "descendant", fixity = Prefix}
toSpine (PF _) DESCorSELF = DESCorSELF `As` Con {name = "descorself", fixity = Prefix}
toSpine (PF _) (NAME s) = (NAME s) `As` Con {name = s, fixity = Prefix}
toSpine (PF (TU r)) (f :/: g) = (((:/:) `As` Con {name="/", fixity=Infix}) `Ap` (PF (TU (List Dynamic)) :| f)) `Ap` (PF (TU r) :| g)
toSpine (PF _) NONEMPTY = NONEMPTY `As` Con {name = "nonempty", fixity = Prefix}
toSpine (PF _) (f :?: p) =  (((:?:) `As` Con {name="?", fixity=Infix}) `Ap` (PF (TU (List Dynamic)) :| f)) `Ap` (PF (TU Bool) :| p)
toSpine (PF (TU b)) (QCOMP a q f) = (((QCOMP a) `As` Con {name =">", fixity=Infix}) `Ap` (PF (TU a) :| q)) `Ap` (PF (Func a b) :| f)
toSpine (PF (TU (Prod a b))) (f :/\: g) = (((:/\:) `As` Con {name ="/\\", fixity=Infix}) `Ap` (PF (TU a) :| f)) `Ap` (PF (TU b) :| g)
-- WARNING NEEDS TO BE DONE DIFFERENTLY!
-- toSpine (PF (Func (Tag _ a) b)) f = toSpine (PF (Func a b)) f
-- toSpine (PF (Func a (Tag _ b))) f = toSpine (PF (Func a b)) f
toSpine t a = error $ "toSpine missing case for "++show t

-- | Generic equality
geq :: Type a -> a -> a -> Bool
geq t x y = xyz t t x y
    where xyz :: Type a -> Type b -> a -> b -> Bool
	  xyz t1 t2 x y = aux (toSpine t1 x) (toSpine t2 y)
	  aux :: Spine a -> Spine b -> Bool
	  aux (_ `As` c1)        (_ `As` c2)         = name c1 == name c2
	  aux (Ap f1 (t1 :| a1)) (Ap f2 (t2 :| a2))  = aux f1 f2 && xyz t1 t2 a1 a2
	  aux _ _ = False

instance Typeable a => Eq (PF a) where
    (==) = geq (PF typeof)

-- | Generic show
gshow :: Type a -> a -> String
gshow Dynamic (Dyn t x) = gshow t x
gshow t x   =  aux (toSpine t x)
    where aux :: Spine a -> String
          aux (Ap f@(Ap (_ `As` c) (a :| x)) (b :| y)) | fixity c == Infix = "(" ++ gshow a x ++ " " ++ name c ++ " " ++ gshow b y ++ ")"
						       | otherwise = "(" ++ aux f ++ " " ++ gshow b y ++ ")"
          aux (_ `As` c)       = name c
          aux (Ap f (t :| a))  =  "(" ++ aux f ++ " " ++ gshow t a ++ ")"

instance Typeable a => Show (PF a) where
    show = gshow (PF typeof)

-- | Almost generic compare
gcompare :: Type a -> a -> a -> Ordering
gcompare Int x y = compare x y
gcompare Bool x y = compare x y
gcompare Char x y = compare x y
gcompare String x y = compare x y
gcompare One _ _ = EQ
gcompare (List a) x y = aux x y
  where
    aux [] [] = EQ
    aux [] _  = LT
    aux _  [] = GT
    aux (x:xs) (y:ys) 
     = case gcompare a x y of
         LT -> LT
         GT -> GT
         EQ -> aux xs ys
gcompare (Prod a b) (x1,y1) (x2,y2) 
 = let t = gcompare a x1 x2
   in case t of 
     EQ -> gcompare b y1 y2
     otherwise -> t
-- THE REST IS MISSING !!!


-----------------------------------------------------------------------------

