-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  stable
-- Portability :  portable
--
-- See
--
--     Haskell Misc Functions
-- 
-- by
--
--     Diogo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------



module Data.TwoLevelExchangeMisc where

import Data.Transform.Type
import Data.Transform.TwoLevelCombinators
import Data.Transform.RulesWithRefs
import Data.Map as Map
import System.IO.Unsafe

-- infix declarations
infix 5 <..>

-- * Functions

escapeQuotes :: String -> String
escapeQuotes [] = []
escapeQuotes ('\'':xs) = "\\'" ++ escapeQuotes xs
escapeQuotes (x:xs) = x : escapeQuotes xs

foldrM :: Monad m => (a -> b -> m b) -> b -> [a] -> m b
foldrM _ b [] = return b
foldrM f b l = f (last l) b >>= \fax -> foldrM f fax (init l)

-- | merge two DynTypes
mergeDynT :: Maybe DynType -> Maybe DynType -> Maybe DynType
mergeDynT (Just d) Nothing = Just d
mergeDynT (Just (DynT t1)) (Just (DynT t2)) = Just $ DynT (Prod t1 t2)

-- | creates a Maybe pair from 2 Maybes

mergeMb :: Maybe a -> Maybe b -> Maybe (a,b)
mergeMb a b = do
	x <- a
	y <- b
	return (x,y)

-- | Extract the first component of a triplet.
fst3 :: (a, b, c) -> a
fst3 (x,_,_) = x

-- | Extract the second component of a triplet.
snd3 :: (a, b, c) -> b
snd3 (_,y,_) = y

-- | Extract the third component of a triplet.
thr3 :: (a, b, c) -> c
thr3 (_,_,z) = z

-- | Extract the first component of a quartet.
fst4 :: (a, b, c, d) -> a
fst4 (x,_,_,_) = x

-- | Extract the second component of a quartet.
snd4 :: (a, b, c, d) -> b
snd4 (_,y,_,_) = y

-- | Extract the third component of a quartet.
thr4 :: (a, b, c ,d) -> c
thr4 (_,_,z,_) = z

-- | Extract the fourth component of a quartet
fth4 :: (a, b, c, d) -> d
fth4 (_,_,_,w) = w

-- | Extract the first component of a quintet.
fst5 :: (a, b, c, d, e) -> a
fst5 (x,_,_,_,_) = x

-- | Extract the second component of a quintet.
snd5 :: (a, b, c, d, e) -> b
snd5 (_,y,_,_,_) = y

-- | Extract the third component of a quintet.
thr5 :: (a, b, c ,d, e) -> c
thr5 (_,_,z,_,_) = z

-- | Extract the fourth component of a quintet
fth5 :: (a, b, c, d, e) -> d
fth5 (_,_,_,w,_) = w

-- | Extracts the fifth component of a quintet
fit5 :: (a, b, c, d, e) -> e
fit5 (_,_,_,_,e) = e

-- | Extract the first component of a tuple with 6 elements.
fst6 :: (a, b, c, d, e, f) -> a
fst6 (x,_,_,_,_,_) = x

-- | Extract the second component of a tuple with 6 elements.
snd6 :: (a, b, c, d, e, f) -> b
snd6 (_,y,_,_,_,_) = y

-- | Extract the third component of a tuple with 6 elements.
thr6 :: (a, b, c ,d, e, f) -> c
thr6 (_,_,z,_,_,_) = z

-- | Extract the fourth component of a tuple with 6 elements.
fth6 :: (a, b, c, d, e, f) -> d
fth6 (_,_,_,w,_,_) = w

-- | Extract the fifth component of a tuple with 6 elements.
fit6 :: (a, b, c, d, e, f) -> e
fit6 (_,_,_,_,e,_) = e

-- | Extract the sixth component of a tuple with 6 elements.
sth6 :: (a, b, c, d, e, f) -> f
sth6 (_,_,_,_,_,f) = f


-- | binary composition of functions
(<..>) :: (a -> b -> c) -> (d -> e -> (a,b)) -> (d -> e -> c)
(<..>) f g = curry (uncurry f . uncurry g)

-- | the (:) constructor for Maybe types
appendMaybe :: Maybe a -> [a] -> [a]
appendMaybe (Just x) l = x : l
appendMaybe Nothing l = l

-- | convert a Maybe to a Bool type
maybe2Bool :: Maybe a -> Bool
maybe2Bool (Just _) = True
maybe2Bool Nothing = False

-- | convert a Maybe of List to a List
maybe2List :: Maybe [a] -> [a]
maybe2List Nothing = []
maybe2List (Just l) = l

-- | concatenation for lists of assocs (concatenates also assocs for repeated keys)
assocUnion :: (Ord a) => Map a [b] -> Map a [b] -> Map a [b]
assocUnion x y = foldr assocGroup Map.empty $ (Map.toList x) ++ (Map.toList y)
	where
	assocGroup :: (Ord a) => (a,[b]) -> Map a [b] -> Map a [b]
	assocGroup p@(a,b) m =
		case (Map.lookup a m) of
			-- remember the left map is preferred for duplicate keys
			Nothing -> Map.singleton a b `Map.union` m
			(Just b1) -> Map.singleton a (b ++ b1) `Map.union` m

-- | checks if the key exists in the list
lookup6 :: (Eq a) => a -> [(a,b,c,d,e,f)] -> Bool
lookup6 _ [] = False
lookup6 a (x:xs) | fst6 x == a = True
		 | otherwise = lookup6 a xs

-- | debug function
trace :: (a -> b) -> String -> a -> b
trace f s a = unsafePerformIO (do {putStrLn s; return $ f a })

-- | Converts a lsit of strings to a list delimited by parenthesis.
parenthesis :: [String] -> String
parenthesis = ('(':) . (++")") . init . foldr (\h r -> h ++ ',' : r) []

-- | Returns the maximum value in a list on 'Int'.
-- If ths list is empty, returns 0 by default.
maxInt :: [Int] -> Int
maxInt [] = 0
maxInt x = maximum x
