
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  development
-- Portability :  portable
--
--
--     2LT Type to XML conversion
-- 
-- by
--
--     Diogo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Data.XML.TypetoXML
    where

import Data.Transform.Type
import Data.Transform.RulesWithRefs
import Text.XML.HaXml.Types
import Data.Set as Set
import Data.List as List
import Control.Monad.State
import Data.Maybe

-- | Tests wether a Type is a basic type without tags and not One.
testBasicType Int = True
testBasicType Bool = True
testBasicType Char = True
testBasicType String =True
testBasicType _ = False

-- | Creates a XSD attribute in HaXml 'Attribute' format.
genAttribute :: String -> String -> Attribute
genAttribute n v = (n, AttValue [Left $ v])


-- | Converts a 'Type' to it's equivalent XSD and XML representations.
type2HaXml :: MonadPlus m => Type a -> a -> m (Document, Document)
type2HaXml t v = do
		xsd <- type2Xsd t
		xml <- type2Xml t v
		return (xsd,xml)

-- | Converts a 'Type' to it's equivalent XSD Representation.
type2Xsd :: MonadPlus m => Type a -> m Document
type2Xsd t = do
	res <- type2Xsd' t False False
	return $ Document (Prolog  (Just (XMLDecl "1.0" (Just (EncodingDecl "utf-8")) Nothing)) Nothing) [] (Elem "xs:schema" [genAttribute "xmlns:xs" "http://www.w3.org/2001/XMLSchema",genAttribute "targetNamespace" "http://www.di.uminho.pt",genAttribute "xmlns" "http://www.di.uminho.pt",genAttribute "elementFormDefault" "qualified"] [CElem res])
	where
	type2Xsd' :: MonadPlus m => Type a 	-- the type to convert
		  -> Bool	-- boolean for minOccurs="0" or not
		  -> Bool	-- boolean for maxOccurs="unbounded" or not
		  -> m Element	-- resulting xml element

	type2Xsd' t@(Ann l a) _ _ | isAtt l = if (testBasicType a)
		then liftM (\x -> Elem "xs:attribute" [x,basicType a] []) $ liftMb (genAttribute "name") $ getAttFieldName l
		else fail $ "Error in Type to XSD conversion: attribute for type '" ++ show t ++ "' is not a basic type"
	type2Xsd' (Ann l t) b1 b | testBasicType t = liftM (\x -> Elem "xs:element" ([x,basicType t] ++ genListAtts b1 b) []) $ liftMb (genAttribute "name") $ getFieldName l
	type2Xsd' (Ann l (List a)) _ _
		| testBasicType a = liftMb (\x -> Elem "xs:element" [genAttribute "name" x] [CElem $ Elem "xs:simpleType" [] [CElem $ Elem "xs:list" [basicItemType a] []]]) $ getFieldName l
		| otherwise = type2Xsd'' a True >>= \y -> do {x <- mb2mplus $ getFieldName l; return $ Elem "xs:element" [genAttribute "name" x] y}
	type2Xsd' (List t@(Ann l a)) b1 _
		| testBasicType a = liftMb (\x -> Elem "xs:element" ([genAttribute "name" x, basicType a] ++ genListAtts b1 True) []) $ getFieldName l
		| otherwise = type2Xsd' t b1 True
	type2Xsd' (Ann l (Set a)) _ _ 
		| testBasicType a = liftMb (\x -> Elem "xs:element" [genAttribute "name" x] [CElem $ Elem "xs:simpleType" [] [CElem $ Elem "xs:list" [basicItemType a] []]]) $ getFieldName l
		| otherwise = type2Xsd'' a True >>= \y -> do {x <- mb2mplus $ getFieldName l; return $ Elem "xs:element" [genAttribute "name" x] y}
	type2Xsd' (Set (Ann l a)) b1 _
		| testBasicType a = liftMb (\x -> Elem "xs:element" ([genAttribute "name" x, basicType a] ++ genListAtts b1 True) []) $ getFieldName l
		| otherwise = type2Xsd' t b1 True
	type2Xsd' (Ann l (Maybe x)) b1 b = type2Xsd' (Ann l (Either x One)) b1 b
	type2Xsd' (Ann l (Either x One)) _ b = type2Xsd' (Ann l x) True b
	type2Xsd' (Ann l x@(Either _ _)) b1 b
		| isJust $ getFieldName l = type2Xsd'' x False >>= \y -> do {
			x <- mb2mplus $ getFieldName l;
			return $ Elem "xs:element" ([genAttribute "name" x] ++ genListAtts b1 b) y}
		| otherwise = type2Xsd' x b1 b
	type2Xsd' (Ann l x@(Prod _ _)) b1 b
		| isJust $ getFieldName l = type2Xsd'' x False >>= \y -> do {
			x <- mb2mplus $ getFieldName l; 
			return $ Elem "xs:element" ([genAttribute "name" x] ++ genListAtts b1 b) y}
		| otherwise = type2Xsd' x b1 b
	type2Xsd' x@(Either _ _) b1 b = nestedTypeEithers x >>= return . Elem "xs:choice" (genListAtts b1 b)
	type2Xsd' x@(Prod _ _) b1 b = nestedTypeProds x >>= return . Elem "xs:sequence" (genListAtts b1 b)
	type2Xsd' z _ _ = fail $ "Error in Type to XSD conversion: don't know how to convert '" ++ show z ++ "'"
	
	type2Xsd'' :: MonadPlus m => Type a -> Bool -> m [Content]
	type2Xsd'' x@(Either _ _) b = do
		res <- nestedTypeEithers x
		return [CElem $ Elem "xs:complexType" [] [CElem $ Elem "xs:choice" (genListAtts False b) res]]
	type2Xsd'' x@(Prod _ _) b = do
		res <- nestedTypeProds x
		return [CElem $ Elem "xs:complexType" [] [CElem $ Elem "xs:sequence" (genListAtts False b) res]]
	type2Xsd'' x b = type2Xsd' x False b >>= return . (:[]) . CElem
	
	genListAtts :: Bool -> Bool -> [Attribute]
	genListAtts True False = [genAttribute "minOccurs" "0"]
	genListAtts _ True = [genAttribute "minOccurs" "0" , genAttribute "maxOccurs" "unbounded"]
	genListAtts False False = []
	
	basicType :: Type a -> Attribute
	basicType Int = genAttribute "type" "xs:integer"
	basicType Bool = genAttribute "type" "xs:boolean"
	basicType Char = genAttribute "type" "xs:string"
	basicType String = genAttribute "type" "xs:string"
	
	basicItemType :: Type a -> Attribute
	basicItemType Int = genAttribute "itemType" "xs:integer"
	basicItemType Bool = genAttribute "itemType" "xs:boolean"
	basicItemType Char = genAttribute "itemType" "xs:string"
	basicItemType String = genAttribute "itemType" "xs:string"
	
	nestedTypeProds :: MonadPlus m => Type a -> m [Content]
	nestedTypeProds (Prod a b) = do
		l <- nestedTypeProds a
		r <- nestedTypeProds b
		return $ l ++ r
	nestedTypeProds x = type2Xsd' x False False >>= return . (:[]) . CElem

	nestedTypeEithers :: MonadPlus m => Type a -> m [Content]
	nestedTypeEithers (Either One b) = do
		res <- type2Xsd' b False False
		return [CElem $ Elem "xs:sequence" [genAttribute "minOccurs" "0"] [CElem res]]
	nestedTypeEithers (Either a One) = do
		res <- type2Xsd' a False False
		return [CElem $ Elem "xs:sequence" [genAttribute "minOccurs" "0"] [CElem res]]
	nestedTypeEithers (Either a b) = do
		l <- nestedTypeEithers a
		r <- nestedTypeEithers b
		return $ l ++ r
	nestedTypeEithers x = type2Xsd' x False False >>= return . (:[]) . CElem

-- | Converts a 'Type' to it's equivalent XML representation.
type2Xml :: MonadPlus m => Type a -> a -> m Document
type2Xml t v = type2Xml' t v >>= return . Document (Prolog  (Just (XMLDecl "1.0" (Just (EncodingDecl "utf-8")) Nothing)) Nothing) [] 
	where
	type2Xml' :: MonadPlus m => Type a -> a -> m Element
	type2Xml' (Ann l t) v | testBasicType t = liftMb (\x -> Elem x [] [CString False (showBasicType t v)]) $ getFieldName l
	type2Xml' (Ann l (List a)) la 
		| testBasicType a = liftMb (\x -> Elem x [] [CString False (showBasicList a la)]) $ getFieldName l
		| otherwise = mapM (type2Xml'' a) la >>= return . concat >>= \y -> do {x <- mb2mplus $ getFieldName l; return $ Elem x [] y}
	type2Xml' (Ann l (Set a)) sa
		| testBasicType a =  liftMb (\x -> Elem x [] [CString False (showBasicSet a sa)]) $ getFieldName l
		| otherwise = mapM (type2Xml'' a) (Set.toAscList sa) >>= return . concat >>= \y -> do {x <- mb2mplus $ getFieldName l; return $ Elem x [] y}
	type2Xml' (Ann l (Maybe t)) v = type2Xml' (Ann l (Either t One)) $ maybe (Right ()) Left v
	type2Xml' (Ann l t@(Either _ _)) v
		| isJust $ getFieldName l = type2Xml'' t v >>= \y -> do {x <- mb2mplus $ getFieldName l; return $ Elem x [] y}
		| otherwise = type2Xml' t v
	type2Xml' (Ann l t@(Prod _ _)) v@(_,_)
		| isJust $ getFieldName l = do
			atts <- genProdAtts t v
			type2Xml'' t v >>= \y -> do {x <- mb2mplus $ getFieldName l; return $ Elem x atts y}
		| otherwise = type2Xml' t v
	type2Xml' t v = fail $ "Error in Type to XML conversion:  don't know how to convert '" ++ show t ++ "'"
	
	type2Xml'' :: MonadPlus m => Type a -> a -> m [Content]
	type2Xml'' t@(Either _ _) = nestedValueEithers t
	type2Xml'' t@(Prod _ _) = nestedValueProds t
	type2Xml'' x = \arg -> type2Xml' x arg >>= return . (:[]) . CElem

	genProdAtts :: MonadPlus m => Type a -> a -> m [Attribute]
	genProdAtts (Ann l a) v | isAtt l = if (testBasicType a)
		then liftMb (return . (flip genAttribute) (showBasicType a v)) $ getAttFieldName l	
		else fail $ "Error in Type to XML conversion: attribute for type '" ++ show t ++ "' and value '" ++ gshow a v ++ "' is not a basic type"
	genProdAtts (Prod a b) (x,y) = do
		l <- genProdAtts a x
		r <- genProdAtts b y
		return $ l ++ r
	genProdAtts t v = return []

	showBasicList :: Type a -> [a] -> String
	showBasicList t = init . foldr (\h r -> showBasicType t h ++ ' ' : r) []
	
	showBasicSet :: Type a -> Set a -> String
	showBasicSet t = init . fold (\h r -> showBasicType t h ++ ' ' : r) []

	showBasicType :: Type a -> a -> String
	showBasicType Int = show
	showBasicType Bool = show
	showBasicType Char = (:[])
	showBasicType String = id

	aux :: MonadPlus m => Type a -> a -> m Content
	aux t x = do {x' <- type2Xml' t x; return $ CElem x'}

	nestedValueProds :: MonadPlus m => Type a -> a -> m [Content]
	nestedValueProds (Prod a b) (x,y) = do
		l <- nestedValueProds a x
		r <- nestedValueProds b y
		return $ l ++ r
	nestedValueProds (List t@(Ann l a)) la = mapM (aux t) la
	nestedValueProds (Set t@(Ann l a)) sa =  mapM (aux t) $ Set.toAscList sa
	nestedValueProds t@(Either _ _) v = type2Xml'' t v
	nestedValueProds (Ann l _) _ | isAtt l = return [] -- ignore attributes
	nestedValueProds (Ann l (Either t One)) (Left v) = type2Xml' (Ann l t) v >>= return . (:[]) . CElem
	nestedValueProds (Ann l (Either t One)) (Right v) = return []
	nestedValueProds t v = type2Xml' t v >>= return . (:[]) . CElem

	nestedValueEithers :: MonadPlus m => Type a -> a -> m [Content]
	nestedValueEithers (Either One _) (Left ()) = return []
	nestedValueEithers (Either One b) (Right bv) = nestedValueEithers b bv
	nestedValueEithers (Either a One) (Left av) = nestedValueEithers a av
	nestedValueEithers (Either _ One) (Right ()) = return []
	nestedValueEithers (Either a _) (Left av) = nestedValueEithers a av
	nestedValueEithers (Either _ b) (Right bv) = nestedValueEithers b bv
	nestedValueEithers (Ann l (Either t One)) (Left v) = type2Xml' (Ann l t) v >>= return . (:[]) . CElem
	nestedValueEithers (Ann _ (Either t One)) (Right v) = return []
	nestedValueEithers t@(Prod _ _) v = type2Xml'' t v
	nestedValueEithers t v = type2Xml' t v >>= return . (:[]) . CElem


