----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  development
-- Portability :  portable
--
--
--     XML to 2LT Type conversion
-- 
-- by
--
--     Diogo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------


module Data.XML.XMLtoType where

import Text.XML.HaXml.Types
import Data.Transform.Type
import Data.Transform.RulesWithRefs
import Data.TwoLevelExchangeMisc
import Data.Maybe
import Data.Set as Set
import GHC.Base as Base
import Data.List as List
import Control.Monad

-- * Main functions

-- | Converts a list of xml documents for a single schema to a Dynamic type.
-- Creates a set for the list of files.
xmls2Type :: MonadPlus m => Type a -> [Document] -> m Dynamic
xmls2Type t ld = case ld of
			[d] -> do
				v <- xml2Values d t
				return $ Dyn t v
			otherwise -> aux t
	where
	aux :: MonadPlus m => Type a -> m Dynamic
	aux t = insertValues t `mplus` fail "Error in XML to Type conversion: unknown"
	
	insertValues :: MonadPlus m => Type a -> m Dynamic
	insertValues t = return $ Dyn (Set t) $ fromDistinctAscList $ Base.map (\(Document x y e) -> fromJust $ findTagE True e t) ld

-- | Read the values from a document.
xml2Values :: MonadPlus m => Document -> Type a -> m a
xml2Values (Document _ _ e) t = findTagE True e t `mplus` fail "Error in XML to Type conversion: unknown."
	
-- * Reading functions

-- | Need something with a tag and try to find it.
findTagE :: MonadPlus m => Bool -> Element -> Type a -> m a
findTagE _ e (Mu f) = error "TODO: Mu!"
findTagE b e (Ann l (Either t One)) = findTagE b e (Either (Ann l t) One)
findTagE b (Elem n1 a c) (Ann l t)
	| isJust n && n1 == fromJust n = if b then readE [] c t else readE a c t
	| otherwise = mzero
	where
	n = getFieldName l
findTagE b e (Either x y)
	| isJust auxL = return $ Left $ fromJust auxL
  	| isJust oneY = return $ Right $ fromJust $ oneY
  	| isJust auxR = return $ Right $ fromJust auxR
 	| otherwise = mzero
	where
	auxL = findTagE b e x
	auxR = findTagE b e y
	oneY = findOne y
	
findTagE b e (Prod x y)
	| isJust oneX && isJust aux2 = return (fromJust oneX, fromJust aux2)
	| isJust oneY && isJust aux1 = return (fromJust aux1, fromJust oneY)
	| isJust oneX && isJust oneY = return (fromJust oneX, fromJust oneY)
	| isJust aux1 = do
		oneY <- mkOne y
		return (fromJust aux1,oneY)
	| isJust aux2 = do
		oneX <- mkOne x
		return (oneX,fromJust aux2)
	where
	oneX = findOne x
	oneY = findOne y
	aux1 = findTagE b e x
	aux2 = findTagE b e y

findTagE b e (List t) | isJust $ findTagE b e t = return [fromJust $ findTagE b e t]
findTagE _ _ _ = mzero

-- | Create an empty value for a One data type.
mkOne :: MonadPlus m => Type a -> m a
mkOne One = return ()
mkOne (List _) = return []
mkOne (Prod x y) = do
	oneX <- mkOne x
	oneY <- mkOne y
	return (oneX,oneY)
mkOne (Either x y)
	| isJust oneX = return $ Left $ fromJust oneX
	| isJust oneY = return $ Right $ fromJust oneY
	where
	oneX = mkOne x
	oneY = mkOne y
mkOne (Ann _ x) = mkOne x
mkOne _ = mzero

-- | Find One types that may be hidden! not empty lists too!
findOne :: Type a -> Maybe a
findOne One = Just ()
findOne (Ann _ x) = findOne x
findOne (Either a b)
		| isJust $ findOne a = Just $ Left $ fromJust $ findOne a
		| isJust $ findOne b = Just $ Right $ fromJust $ findOne b
findOne (Prod a b)
		| (isJust $ findOne a) && (isJust $ findOne b) = Just (fromJust $ findOne a, fromJust $ findOne b)
findOne (List l)
		| isJust $ findOne l = Just $ [fromJust $ findOne l]
		| otherwise = Nothing
findOne _ = Nothing

-- | Try to read Element values.
readE :: MonadPlus m => [Attribute] -> [Content] -> Type a -> m a
readE [] [] t = mkOne t
readE a [] t = readA a t
readE [] c t = readC c t
readE a c (Prod x y) = do
	auxR <- readA a y
	auxL <- readC c x
	return (auxL,auxR)
readE _ _ _ = mzero

-- | Try to read a list of Attributes
readA :: MonadPlus m => [Attribute] -> Type a -> m a
readA [(n1, AttValue [al])] (Ann l t)
	| isJust n && n1 == fromJust n && isBasicType t = do
		att <- getAttr al
		readV att t
	where
	n = getAttFieldName l
readA [(n1, AttValue l)] (Ann l2 (List t))
	| isJust n && n1 == fromJust n && isBasicType t = foldrM aux [] l
	where
	n = getAttFieldName l2
	aux a b = do
		x <- getAttr a
		y <- readV x t
		return $ y : b
readA (h:t) (Prod x y) = do
	auxL <- readA [h] x
	auxR <- readA t y
	return (auxL,auxR)
readA _ _ = fail "Error in XML to Type conversion: warning -  read attribute ignored."

-- | Try to read a list of Contents
readC :: MonadPlus m => [Content] -> Type a -> m a
readC c l@(List _) = do
	clst <- readCList c l
	case clst of
		(vs,[])   -> return vs
		otherwise -> mzero
readC [x] t = readUC x t 
readC e p@(Prod x y)
	| isJust oneX && isJust aux1 = return (fromJust oneX, fromJust aux1)
	| isJust oneY && isJust aux2 = return (fromJust aux2, fromJust oneY)
	| isJust oneX && isJust oneY = return (fromJust oneX, fromJust oneY)
	| otherwise = readCP e p
	where
	oneX = findOne x
	oneY = findOne y
	aux1 = readC e y
	aux2 = readC e x
readC c (Either x y)
	| isJust l = return $ Left $ fromJust l
	| isJust r = return $ Right $ fromJust r
	where
	l = readC c x
	r = readC c y
readC a b = mzero

-- | Try to read the Value of a Type. It's expected to receive a basic type.
readV :: MonadPlus m => String -> Type a -> m a
readV s String = return s
readV s Int = return $ fromEnum (read s :: Float)
readV s _ = fail "Error in XML to Type conversion: read value."

-- * Auxiliary functions

-- | Get Attribute value. Throws an error when attribute is a reference!
getAttr :: MonadPlus m => Either String Reference -> m String
getAttr (Left s) = return s
getAttr _ = fail "Error in XML to Type conversion: attribute fetching error."

-- | readC aux function. Try to read Content to Products.
readCP :: MonadPlus m => [Content] -> Type a -> m a
readCP [x] t = readUC x t
readCP c (Prod x y)
	| hasList x = do
		(l,rest) <- readCP' c x
		readc <- readC rest y
		return (l,readc)
readCP (h:t) (Prod x y) = do
	auxL <- readUC h x
	auxR <- readC t y
	return (auxL,auxR)
readCP _ _ = mzero

-- | readCP aux function.
readCP' :: MonadPlus m => [Content] -> Type a -> m (a,[Content])
readCP' c (Prod x y) = do
	(l,restL) <- readCP' c x
	(r,restR) <- readCP' restL y
	return ((l,r),restR)
readCP' c l@(List _) = readCList c l
readCP' (x:xs) t
	| isJust l && not (hasList t) = return (fromJust l,xs)
	where
	l = readUC x t
readCP' c t = fail "Error in XML to Type conversion: error reading contents to products."

-- | Reads a list of Content.
-- Stops in the first content that doesn't match the list, returning the list of values read and the not consumed contents
readCList :: MonadPlus m => [Content] -> Type [a] -> m ([a],[Content])
readCList [] _ = return ([],[])
readCList c@(x:xs) l@(List _) = do
	(vs,rest) <- readCList xs l
	case (readUC x l) of
		Nothing  -> return ([],x:xs)
		(Just res) -> return (res ++ vs,rest)
readCList x y = fail $ "Error in XML to Type conversion: reading content list."

-- readC aux function. Try to read singular types.
readUC :: MonadPlus m => Content -> Type a -> m a
readUC (CString _ s) t
	| isBasicType t = readV s t
--readUC _ (List (Ann [FieldName "review"] t)) = error $ "readUC:" ++ show t
readUC (CString _ s) (List t)
	| isBasicType t && (not (hasTag t)) = mapM (\x -> readV x t) $ words s
readUC (CElem e) t = findTagE False e t
readUC c t = fail "Error in XML to Type conversion: reading single types."

hasTag (Ann _ _) = True
hasTag _ = False

-- | Tests if a 'Type' has lists inside.
-- This is used when assigning values to products.
hasList :: Type a -> Bool
hasList (Prod a b) = hasList a || hasList b
hasList (List t) = True
hasList _ = False

