-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     Creates a Dynamic Type from a XSD file
-- 
-- by
--
--     Diogo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Data.XML.XSDtoType where

import Language.XMLSchema.Syntax
import Language.XMLSchema.NameSpaces
import Data.Transform.Type
import Data.Transform.TwoLevelCombinators
import Data.Transform.RulesWithRefs hiding (assocl)
import Pointless.Combinators
import Pointless.Isomorphisms (assocl)
import Text.XML.HaXml.OneOfN
import Control.Monad (MonadPlus,mapM)
import Data.Set as Set
import Data.Map as Map
import Data.List as List
import Data.Maybe
import Text.XML.HaXml.Xml2Haskell
import Data.TwoLevelExchangeMisc

infixr 5 !>

-- * NameSpaces

-- | Representation of NameSpaces.
type MyNS = (Prefix -- ^ identification
	   , XSTypes) -- ^ types definition 

-- | Creates a map from type name to XSType with correct prefixes.
createTypeMap :: MonadPlus m => MyNS -> m (Map String XSType)
createTypeMap (p,l) = foldrM f Map.empty l
	where
	f x r = do
		a <- aux p x
		return $ Map.union a r
	
	aux :: MonadPlus m => Prefix -> XSType -> m (Map String XSType)
	aux p (XSTTag n t) = return $ Map.singleton (resolveTagName p n) (aux2 p t)
	aux p _ = fail "Error in XSD to Type conversion: trying to create a map without a name"

	aux2 :: Prefix -> XSType -> XSType
	aux2 p (XSTTag n t) = XSTTag n (aux2 p t)
	aux2 p XSTNull = XSTNull
	aux2 p (XSTList t) = XSTList (aux2 p t)
	aux2 p (XSTSet t) = XSTSet (aux2 p t)
	aux2 p (XSTEither t1 t2) = XSTEither (aux2 p t1) (aux2 p t2)
	aux2 p (XSTProd t1 t2) = XSTProd (aux2 p t1) (aux2 p t2)
	aux2 p (XSTMaybe t) = XSTMaybe (aux2 p t)
	aux2 p (XSTRef r) = XSTRef (resolveTagName p r)
	
-- | 
resolveTagName :: Prefix -> String -> String
resolveTagName p n | p == "" = n
		   | foldr (\x r -> (x == ':') || r) False n = n
		   | otherwise = p ++ ":" ++ n

-- | Get information about current NameSpace.
getInternalMyNS :: MonadPlus m => Schema -> m MyNS
getInternalMyNS x = do
	let l = aux $ targetPrefix $ nseSchema x
	r <- getSchemaTypes x
	return (l,r)
	where
	aux Nothing = ""
	aux (Just p) = p

-- | Get types definition from a schema.
getSchemaTypes :: MonadPlus m => Schema -> m XSTypes
getSchemaTypes x = (getXSTypes $ xs'schemaTypes $ parsedSchema x) >>= return . removeNulls

-- | Get types information from Xs'schema
xs'schemaTypes :: Xs'schema -> [((OneOf7 Xs'simpleType Xs'complexType Xs'element Xs'attribute Xs'attributeGroup Xs'group Xs'notation),[(Xs'annotation)])]
xs'schemaTypes (Xs'schema _ _ types) = types

-- | Resolve the Type from a given schema and a list of external schemas.
xsd2Type :: MonadPlus m => [Schema] -> Schema -> m DynType
xsd2Type x l = aux (x,l)
	where
	aux (a,b) = do
		l <- mapM (\x -> getInternalMyNS x >>= createTypeMap) a
		let r = (b,collectRootElementNames b)
		uncurry createDynType $ (fst /\ removeNullsFromXSType . uncurry (uncurry mk_root) . assocl) $ (foldr Map.union Map.empty l,r) 
	
	mk_root m sc = foldr (\x r -> XSTEither (XSTTag x $ XSTRef $ resolveTagName (prefix sc) x) r) XSTNull
	
	prefix s | isJust aux = fromJust aux
		 | otherwise = ""
		where
		aux = targetPrefix $ nseSchema s

-- | Create a dynamic type from the types definitions and the root elements.
createDynType :: MonadPlus m => Map String XSType -> XSType -> m DynType
createDynType m t = expands "_root_" (Map.insert "_root_" t m) Set.empty >>= uncurry xsType2Dyn

-- | Takes the XSType and expands all it's references.
expands :: MonadPlus m => String -> Map String XSType -> Set String -> m (XSType, Map String XSType)
expands s m rec = do
		xt <- s !> m
		new_rec <- recSet xt 
		normalize new_rec xt
	where
	recSet :: MonadPlus m => XSType -> m (Set String)
	recSet XSTNull = return Set.empty
	recSet (XSTTag tg t) = recSet t
	recSet (XSTList x) = recSet x
	recSet (XSTSet x) = recSet x
	recSet (XSTMaybe x) = recSet x
	recSet (XSTEither x y) = do
		a <- recSet x
		b <- recSet y
		return $ a `Set.union` b
	recSet (XSTProd x y) = do
		a <- recSet x
		b <- recSet y
		return $ a `Set.union` b
	recSet (XSTRef r) = do
		test <- isRecursive r rec m
		if ((not (r `Set.member` rec)) && test)
			then return $ Set.singleton r
			else return Set.empty

	normalize :: MonadPlus m => Set String -> XSType -> m (XSType, Map String XSType)
	normalize new_rec XSTNull = return (XSTNull, Map.empty)
	normalize new_rec (XSTTag tg t) = normalize new_rec t >>= return . (XSTTag tg >< id)
	normalize new_rec (XSTList x) = normalize new_rec x >>= return . (XSTList >< id)
	normalize new_rec (XSTSet x) = normalize new_rec x >>= return . (XSTSet >< id)
	normalize new_rec (XSTMaybe x) = normalize new_rec x >>= return . (XSTMaybe >< id)
	normalize new_rec (XSTEither x y) = do
		a <- normalize new_rec x
		b <- normalize new_rec y
		return $ ((uncurry XSTEither) . (fst >< fst) /\ (uncurry Map.union) . (snd >< snd)) (a,b)
	normalize new_rec (XSTProd x y) = do
		a <- normalize new_rec x
		b <- normalize new_rec y
		return $ ((uncurry XSTProd) . (fst >< fst) /\ (uncurry Map.union) . (snd >< snd)) (a,b)
	normalize new_rec t@(XSTRef r)
		| r `Set.member` new_rec = do
			exp <- expands r m (Set.union rec new_rec)
			return (t, (uncurry (Map.insert r)) exp)
		| isBasicXsType r || r `Set.member` rec = return (t, Map.empty)
		| otherwise = expands r m (Set.union rec new_rec)

-- | Takes a expanded XSType and transforms if to a Dynamic Type.
xsType2Dyn :: MonadPlus m => XSType -> Map String XSType -> m DynType
xsType2Dyn XSTNull _ = fail "Fatal error in XSD to Type conversion: no root element defined?"
xsType2Dyn (XSTTag tg t) m = xsType2Dyn t m >>= return . createDynTTag tg
xsType2Dyn (XSTList x) m = xsType2Dyn x m >>= return . createDynTList
xsType2Dyn (XSTSet x) m = xsType2Dyn x m >>= return . createDynTSet
xsType2Dyn (XSTMaybe x) m = xsType2Dyn x m >>= return . createDynTMaybe
xsType2Dyn (XSTEither x y) m = do
	l <- xsType2Dyn x m
	r <- xsType2Dyn y m
	return $ createDynTEither l r
xsType2Dyn (XSTProd x y) m = do
	l <- xsType2Dyn x m
	r <- xsType2Dyn y m
	return $ createDynTProd l r
xsType2Dyn (XSTRef r) m
	| isBasicXsType r = convert2basicType r
	| otherwise = do
		x <- r !> m
		fctr <- toFctr x r Set.empty m
		return $ fctr2Dyn fctr

fctr2Dyn :: EncapFctr -> DynType
fctr2Dyn (EFctr f) = DynT $ Mu f

toFctr :: MonadPlus m => XSType -> String -> Set String -> Map String XSType -> m EncapFctr
toFctr  XSTNull _ _ _ = fail "Fatal error in XSD to Type conversion: recursivity error - XSTNull not expected"
toFctr (XSTTag tg t) s st m = do
	x <- toFctr t s st m
	return $ fctrAddTag x tg
toFctr (XSTList x) s st m = do
	x <- toFctr x s st m
	return $ fctrList x
toFctr (XSTSet x) s st m = do
	x <- toFctr x s st m
	return $ fctrSet x
toFctr (XSTMaybe x) s st m = do
	x <- toFctr x s st m
	return $ fctrEither x (EFctr $ K One)
toFctr (XSTEither x y) s st m = do
	l <- toFctr x s st m
	r <- toFctr y s st m
	return $ fctrEither l r
toFctr (XSTProd x y) s st m = do
	l <- (toFctr x s st m)
	r <- (toFctr y s st m)
	return $ fctrProd l r
toFctr (XSTRef r) s st m
	| isBasicXsType r = convert2basicType r >>= return . fctrK
	| r == s = return $ EFctr Id
	| r `Set.member` st = fail "Error in XSD to Type conversion: multiple recursion not supported"
	| otherwise = do
		x <- (r !> m)
		toFctr x r (s `Set.insert` st) m >>= return . fctrMu

fctrMu :: EncapFctr -> EncapFctr
fctrMu (EFctr f) = EFctr $ K $ Mu f

fctrK :: DynType -> EncapFctr
fctrK (DynT t) = EFctr $ K t

fctrEither :: EncapFctr -> EncapFctr -> EncapFctr
fctrEither (EFctr f) (EFctr g) = EFctr $ f :+: g

fctrProd :: EncapFctr -> EncapFctr -> EncapFctr
fctrProd (EFctr f) (EFctr g) = EFctr $ f :*: g

fctrList :: EncapFctr -> EncapFctr
fctrList (EFctr f) = EFctr $ List :@: f

fctrSet :: EncapFctr -> EncapFctr
fctrSet (EFctr f) = EFctr $ Set :@: f

fctrAddTag :: EncapFctr -> String -> EncapFctr
fctrAddTag (EFctr e) s = EFctr $ fctrAddTag' e s
	where
	fctrAddTag' :: (Dist f) => Fctr f -> String -> Fctr f
	fctrAddTag' (K t) tg = K $ Ann (emptyTag {fieldName = Just tg}) t 
	fctrAddTag' (f:@:g) tg = (f . Ann (emptyTag {fieldName = Just tg})) :@: g
	fctrAddTag' a _ = a --TODO - ID = ID and +/* are not possible

createDynTList :: DynType -> DynType
createDynTList (DynT x) = DynT (List x)

createDynTSet :: DynType -> DynType
createDynTSet (DynT x) = DynT (Set x)

createDynTMaybe :: DynType -> DynType
createDynTMaybe (DynT (Ann l x)) = DynT $ Ann l (Either x One)

createDynTEither :: DynType -> DynType -> DynType
createDynTEither (DynT x) (DynT y) = DynT (Either x y)

createDynTProd :: DynType -> DynType -> DynType
createDynTProd (DynT x) (DynT y) = DynT (Prod x y)

createDynTTag :: String -> DynType -> DynType
createDynTTag n (DynT t) = DynT $ Ann (emptyTag {fieldName = Just n}) t

-- | Check if a given XSType is recursive.
isRecursive :: MonadPlus m => String -> Set String -> Map String XSType -> m Bool
isRecursive s st  m | isBasicXsType s = return False
		    | otherwise = do
		    	x <- s !> m
		    	aux s x m (s `Set.insert` st)
	where
	aux :: MonadPlus m => String -> XSType -> Map String XSType -> Set String -> m Bool
	aux s XSTNull m n = return False
	aux s (XSTTag tg t) m n = aux s t m n
	aux s (XSTList x) m n = aux s x m n
	aux s (XSTSet x) m n = aux s x m n	
	aux s (XSTEither x y) m n = do
		a <- aux s x m n
		b <- aux s y m n
		return (a || b)
	aux s (XSTProd x y) m n = do
		a <- aux s x m n
		b <- aux s y m n
		return (a || b)
	aux s (XSTMaybe x) m n = aux s x m n
	aux s (XSTRef r) m n
		| isBasicXsType r = return False
		| r == s = return True
		| otherwise =
			if (r `Set.member` n)
			     then return False
			     else do
				x <- r !> m
				aux s x m (r `Set.insert` n)

-- | Intermediate data type, with references.
data XSType where
	XSTTag :: String -> XSType -> XSType
	XSTRef :: String -> XSType
	XSTList :: XSType -> XSType
	XSTSet :: XSType -> XSType
	XSTEither :: XSType -> XSType -> XSType
	XSTProd :: XSType -> XSType -> XSType
	XSTMaybe :: XSType -> XSType
	XSTNull :: XSType

instance Show XSType where
	show (XSTTag n t) = "(" ++ n ++ ": " ++ (show t) ++ ")"
	show (XSTRef s) = s
	show (XSTSet s) = "(Set " ++ (show s) ++ ")"
	show (XSTList l) = "(List " ++ (show l) ++ ")"
	show (XSTEither l r) = "(Either " ++ (show l) ++ " " ++ (show r) ++ ")"
	show (XSTProd l r) = "(" ++ (show l) ++ ", " ++ (show r) ++ ")"
	show (XSTMaybe t) = "(Maybe " ++ (show t) ++ ")"
	show XSTNull = "Null"

type XSTypes = [XSType]

data EncapFctr where
	EFctr :: (Dist f) => Fctr f -> EncapFctr

-- | Check if it's a basic type.
--   NOTE: Incomplete.
isBasicXsType :: String -> Bool
isBasicXsType s | isSchemaType s "string" = True
	      | isSchemaType s "integer" = True
	      | isSchemaType s "positiveInteger" = True
	      | isSchemaType s "int" = True
	      | isSchemaType s "decimal" = True
	      | isSchemaType s "float" = True
	      | isSchemaType s "double" = True
	      | isSchemaType s "boolean" = True
	      | isSchemaType s "date" = True
	      | isSchemaType s "NMTOKEN" = True
	      | isSchemaType s "NMTOKENS" = True
	      | isSchemaType s "ID" = True
	      | isSchemaType s "anyURI" = True
	      | isSchemaType s "language" = True
	      | isSchemaType s "NCName" = True
	      | isSchemaType s "dateTime" = True
	      | isSchemaType s "nonNegativeInteger" = True
	      | isSchemaType s "short" = True
	      | isSchemaType s "long" = True
	      | isSchemaType s "token" = True
	      | isSchemaType s "duration" = True
	      | isSchemaType s "normalizedString" = True
	      | isSchemaType s "Name" = True
	      | otherwise = False

-- | Convert basic type to DynType.
--   NOTE: Floats are converted to Ints and Dates are converted to Strings.
convert2basicType :: MonadPlus m => String -> (m DynType)
convert2basicType s | isSchemaType s "string" = return $ DynT String
		    | isSchemaType s "integer" = return $ DynT Int
		    | isSchemaType s "positiveInteger" = return $ DynT Int
		    | isSchemaType s "int" = return $ DynT Int
		    | isSchemaType s "decimal" = return $ DynT Int
		    | isSchemaType s "float" = return $ DynT Int
		    | isSchemaType s "double" = return $ DynT Int
		    | isSchemaType s "boolean" = return $ DynT Bool
		    | isSchemaType s "date" = return $ DynT String
		    | isSchemaType s "NMTOKEN" = return $ DynT String
		    | isSchemaType s "NMTOKENS" = return $ DynT (List String)
		    | isSchemaType s "ID" = return $ DynT String
		    | isSchemaType s "anyURI" = return $ DynT String
		    | isSchemaType s "language" = return $ DynT String
		    | isSchemaType s "NCName" = return $ DynT String
		    | isSchemaType s "dateTime" = return $ DynT String
		    | isSchemaType s "nonNegativeInteger" = return $ DynT Int
		    | isSchemaType s "short" = return $ DynT Int
		    | isSchemaType s "long" = return $ DynT Int
		    | isSchemaType s "token" = return $ DynT String
		    | isSchemaType s "duration" = return $ DynT String
		    | isSchemaType s "normalizedString" = return $ DynT String
		    | isSchemaType s "Name" = return $ DynT String
		    | otherwise = fail $ "Error in XSD to Type conversion: basic type '" ++ s ++ "' is not yet supported"
		    
-- | Check basic type prefix.
isSchemaType :: String -> String -> Bool
isSchemaType s str = (s == "xs:" ++ str) || (s == "xsd:" ++ str) || (s == str)

-- | Get a list of XSTypes  from schema content.
getXSTypes :: MonadPlus m => [((OneOf7 Xs'simpleType Xs'complexType Xs'element Xs'attribute Xs'attributeGroup Xs'group Xs'notation),[(Xs'annotation)])] -> m XSTypes
getXSTypes = mapM (getXSType . fst)

getXSType :: MonadPlus m => (OneOf7 Xs'simpleType Xs'complexType Xs'element Xs'attribute Xs'attributeGroup Xs'group Xs'notation) -> m XSType
getXSType (OneOf7 st) = createSimpleType st
getXSType (TwoOf7 ct) = createComplexType ct
getXSType (ThreeOf7 e) = createElementType e
getXSType (FourOf7 at) = createAttributeType at
getXSType (FiveOf7 ag) = createAttributeGroupType ag
getXSType (SixOf7 g) = createGroupType g
getXSType (SevenOf7 x) = fail "Error in XSD to Type conversion: Xs'notation not supported"

-- | Create a simple type.
--   Nested simpleTypes are ignored.
createSimpleType :: MonadPlus m => Xs'simpleType -> m XSType
createSimpleType (Xs'simpleType a _ c) = createSimpleType' c >>=  checkName False a xs'simpleTypeName
	where
	createSimpleType' :: MonadPlus m => (OneOf3 Xs'restriction Xs'list Xs'union) -> m XSType
	createSimpleType' (OneOf3 (Xs'restriction a _ _ _)) = getAttribute a xs'restrictionBase >>= return . XSTRef
	createSimpleType' (TwoOf3 (Xs'list a _ _))	    = getAttribute a xs'listItemType >>= return . XSTList . XSTRef
	createSimpleType' (ThreeOf3 (Xs'union a _ _))       = do
		att <- getAttribute a xs'unionMemberTypes
		return $ foldr (\s r -> XSTEither (XSTRef s) r) XSTNull (words att)

-- | Create a complexType.
createComplexType :: MonadPlus m => Xs'complexType -> m XSType
createComplexType (Xs'complexType a _ c) = createComplexType' c >>= checkName False a xs'complexTypeName
	where
	createComplexType' :: MonadPlus m => (OneOf3 Xs'simpleContent Xs'complexContent ((Maybe (OneOf4 Xs'all Xs'choice Xs'sequence Xs'group)),([(OneOf2 Xs'attribute Xs'attributeGroup)],(Maybe (Xs'anyAttribute))))) -> m XSType
	createComplexType' (OneOf3 sc)  = createComplexType'simpleContent sc
	createComplexType' (TwoOf3 cc)   = createComplexType'complexContent cc
	createComplexType' (ThreeOf3 wt) = createWeirdType wt

	createComplexType'simpleContent :: MonadPlus m => Xs'simpleContent -> m XSType
	createComplexType'simpleContent (Xs'simpleContent _ _ (OneOf2 r)) = createRestrictionType r
	createComplexType'simpleContent (Xs'simpleContent _ _ (TwoOf2 e)) = createExtensionType e

	-- TODO: mixed not implemented.
	createComplexType'complexContent :: MonadPlus m => Xs'complexContent -> m XSType
	createComplexType'complexContent (Xs'complexContent _ _ (OneOf2 r)) = createRestrictionType r
	createComplexType'complexContent (Xs'complexContent _ _ (TwoOf2 e)) = createExtensionType e

-- | Get away with restrictions.
createRestrictionType :: MonadPlus m => Xs'restriction -> m XSType
createRestrictionType (Xs'restriction a _ _ _) = getAttribute a xs'restrictionBase >>= return . XSTRef

createExtensionType :: MonadPlus m => Xs'extension -> m XSType
createExtensionType (Xs'extension a _ w) = (createWeirdType w) >>= return . XSTProd (XSTRef (getMustAttribute a xs'extensionBase))

createWeirdType :: MonadPlus m => (((Maybe (OneOf4 Xs'all Xs'choice Xs'sequence Xs'group)),([(OneOf2 Xs'attribute Xs'attributeGroup)],(Maybe (Xs'anyAttribute))))) -> m XSType
createWeirdType (Just t, (a, _)) = do
	l <- createWeirdType2 t
	r <- createAttributesType a
	return $ XSTProd l r
createWeirdType (Nothing, (a, _)) = createAttributesType a

createWeirdType2 :: MonadPlus m => (OneOf4 Xs'all Xs'choice Xs'sequence Xs'group) -> m XSType
createWeirdType2 (OneOf4 a) = createAllType a
createWeirdType2 (TwoOf4 c) = createChoiceType c
createWeirdType2 (ThreeOf4 s) = createSequenceType s
createWeirdType2 (FourOf4 g) = createGroupType g

-- | Create a Set of nested Eithers to represent a All type.
createAllType :: MonadPlus m => Xs'all -> m XSType
createAllType (Xs'all a _ l) = (foldrM aux XSTNull l) >>= return . XSTSet
	
	where
	aux :: MonadPlus m => Xs'element -> XSType -> (m XSType)
	aux t r = do
		  elem <- createElementType t
		  return $ XSTEither elem r

-- | Create nested Either to represent a Choice type.
createChoiceType :: MonadPlus m => Xs'choice -> m XSType
createChoiceType (Xs'choice a _ l) = foldrM aux XSTNull l >>= checkMinMax a xs'choiceMinOccurs xs'choiceMaxOccurs
	where
	aux t r = do
		x <- createWeirdType3 t
		return $ XSTEither x r

-- | Create nestes Products to represent a Sequence type.
createSequenceType :: MonadPlus m => Xs'sequence -> m XSType
createSequenceType (Xs'sequence a _ l) = foldrM aux XSTNull l >>= checkMinMax a xs'sequenceMinOccurs xs'sequenceMaxOccurs
	where
	aux t r = do
		x <- createWeirdType3 t
		return $ XSTProd x r
	
createWeirdType3 :: MonadPlus m => (OneOf5 Xs'element Xs'group Xs'choice Xs'sequence Xs'any) -> m XSType
createWeirdType3 (OneOf5 e) = createElementType e
createWeirdType3 (TwoOf5 g) = createGroupType g
createWeirdType3 (ThreeOf5 c) = createChoiceType c
createWeirdType3 (FourOf5 s) = createSequenceType s
createWeirdType3 (FiveOf5 a) = createAnyType a 

-- | Create a reference to root type to represent a Any type. Other name spaces attribute is ignored.
createAnyType :: MonadPlus m => Xs'any -> m XSType
createAnyType (Xs'any a _) = checkMinMax a (defToMaybe . xs'anyMinOccurs) (defToMaybe . xs'anyMaxOccurs) $ XSTRef "_root_"

createAttributesType :: MonadPlus m => [(OneOf2 Xs'attribute Xs'attributeGroup)] -> m XSType
createAttributesType [] = return XSTNull
createAttributesType ((OneOf2 a):t) = do
	l <- createAttributeType a
	r <- createAttributesType t
	return $ XSTProd l r
createAttributesType ((TwoOf2 ag):t) = do
	l <- (createAttributeGroupType ag)
	r <- (createAttributesType t)
	return $ XSTProd l r

createAttributeType :: MonadPlus m => Xs'attribute -> m XSType
createAttributeType x@(Xs'attribute a _ _) = createAttributeType' x >>= checkName True a xs'attributeName
	where
	createAttributeType' :: MonadPlus m => Xs'attribute -> m XSType
	createAttributeType' (Xs'attribute _ _ (Just st))				 = createSimpleType st
	createAttributeType' x@(Xs'attribute a _ Nothing)
		| hasAttribute a xs'attributeRef  = getAttribute a xs'attributeRef >>= return . XSTRef
		| hasAttribute a xs'attributeType = getAttribute a xs'attributeType >>= return . XSTRef
		| otherwise = return $ XSTEither (XSTEither (XSTRef "xs:boolean") (XSTRef "xs:string")) (XSTRef "xs:integer")

-- | anyAttribute is ignored.
createAttributeGroupType :: MonadPlus m => Xs'attributeGroup -> m XSType
createAttributeGroupType ag@(Xs'attributeGroup a _ _ _)  = createAttributeGroupType' ag >>= checkName False a xs'attributeGroupName
	where
	createAttributeGroupType' :: MonadPlus m => Xs'attributeGroup -> m XSType
	createAttributeGroupType' (Xs'attributeGroup a _ as _)
		| hasAttribute a xs'attributeGroupRef = getAttribute a xs'attributeGroupRef >>= return . XSTRef
		| otherwise = createAttributesType as

-- | Create an Element type.
createElementType :: MonadPlus m => Xs'element -> m XSType
createElementType e@(Xs'element a _ _ _) = createElementType' e >>= checkName False a xs'elementName >>= checkMinMax a xs'elementMinOccurs xs'elementMaxOccurs
	where
	createElementType' :: MonadPlus m => Xs'element -> m XSType
	createElementType' (Xs'element _ _ (Just (OneOf2 ct)) _) = createComplexType ct
	createElementType' (Xs'element _ _ (Just (TwoOf2 st)) _) = createSimpleType st
	createElementType' e@(Xs'element a _ Nothing _)
		| (hasAttribute a xs'elementRef) && (hasAttribute a xs'elementName) = getAttribute a xs'elementRef >>= return . XSTRef
		| (hasAttribute a xs'elementRef) = do
			att1 <- getAttribute a xs'elementRef
			att <- getAttribute a xs'elementRef
			return $ XSTTag att1 $ XSTRef att
		| hasAttribute a xs'elementType = getAttribute a xs'elementType >>= return . XSTRef
		| hasAttribute a xs'elementName && hasAttribute a xs'elementType = do
			att <- getAttribute a xs'elementName
			fail $ "Error in XSD to Type conversion: : element '" ++ att ++ "' is empty. Please remove it from the schema and try again."
		| hasAttribute a xs'elementName = do
			att <- getAttribute a xs'elementName
			fail $ "Error in XSD to Type conversion: : elemen '" ++ att ++ "' without type or reference to type."

createGroupType :: MonadPlus m => Xs'group -> m XSType
createGroupType g@(Xs'group a _ _) =  createGroupType' g >>= checkName False a xs'groupName
	where
	createGroupType' :: MonadPlus m => Xs'group -> m XSType
	createGroupType' (Xs'group _ _ (Just (OneOf3 a))) = createAllType a
	createGroupType' (Xs'group _ _ (Just (TwoOf3 c))) = createChoiceType c
	createGroupType' (Xs'group _ _ (Just (ThreeOf3 s))) = createSequenceType s
	createGroupType' (Xs'group a _ Nothing) | hasAttribute a xs'groupRef =  getAttribute a xs'groupRef >>= return . XSTRef
						| otherwise = fail "Error in XSD to Type conversion': xs:group without xs:all, xs:choice or xs:sequence"


-- * Auxiliary functions

-- | Tranform a Defaultable into a Maybe.
defToMaybe (Default a) = Just a
defToMaybe (NonDefault a) = Just a

checkName :: MonadPlus m => Bool -> t -> (t -> Maybe String) -> XSType -> m XSType
checkName isatt a f t
	| hasAttribute a f = do
		att <- getAttribute a f
		return $ XSTTag (c ++ att) t
	| otherwise = return t
	where
	c = if isatt then "att$" else ""

-- | Remove Nulls from XSTypes.
removeNulls :: XSTypes -> XSTypes
removeNulls = List.map removeNullsFromXSType

removeNullsFromXSType :: XSType -> XSType
removeNullsFromXSType (XSTMaybe t) = XSTMaybe (removeNullsFromXSType t)
removeNullsFromXSType (XSTTag n t) = XSTTag n (removeNullsFromXSType t)
removeNullsFromXSType (XSTList t) = XSTList (removeNullsFromXSType t)
removeNullsFromXSType (XSTSet t) = XSTSet (removeNullsFromXSType t)
removeNullsFromXSType (XSTEither XSTNull t) = removeNullsFromXSType t
removeNullsFromXSType (XSTEither t XSTNull) = removeNullsFromXSType t
removeNullsFromXSType (XSTEither t1 t2) = XSTEither (removeNullsFromXSType t1) (removeNullsFromXSType t2)
removeNullsFromXSType (XSTProd XSTNull t) = removeNullsFromXSType t
removeNullsFromXSType (XSTProd t XSTNull) = removeNullsFromXSType t
removeNullsFromXSType (XSTProd t1 t2) = XSTProd (removeNullsFromXSType t1) (removeNullsFromXSType t2)
removeNullsFromXSType t = t

-- | Check if min/max occurrences are declared and create appropriate XSType.
checkMinMax :: MonadPlus m => t -> (t -> Maybe String) -> (t -> Maybe String) -> XSType -> m XSType
checkMinMax a min max t
	| hasAttribute a min && hasAttribute a max = do
		min' <- getAttribute a min
		max' <- getAttribute a max
		auxMinMax (myread min') (myread max') t
	| hasAttribute a max = do
		max' <- getAttribute a max
		auxMinMax (Num 1) (myread max') t
	| hasAttribute a min = do
		min' <- getAttribute a min
		auxMinMax (myread min') (Num 1) t
	| otherwise = return t
	where
	auxMinMax :: MonadPlus m => MyRR -> MyRR -> XSType -> m XSType
	auxMinMax _ Unbounded t = return $ XSTList t
	auxMinMax (Num x) (Num y) t | x == 0 && y == 0 = return XSTNull
				    | x == 0 && y == 1 = return $ XSTMaybe t
				    | x == 1 && y == 1 = return t
				    | x > y = fail "Error in XSD to Type conversion: minOccurs greater than maxOccurs"
				    | otherwise = return $ XSTList t

-- | Auxiliary type to check min/max occurrences.
data MyRR = Num Int | Unbounded deriving Show

-- | Tranform a string into min/max definition
myread :: String -> MyRR
myread s | s == "unbounded" = Unbounded
	 | otherwise = Num (read s :: Int)

getMustAttribute :: t -> (t -> String) -> String
getMustAttribute a f = f a

getAttribute :: MonadPlus m => t -> (t -> Maybe String) -> m String
getAttribute a f = aux (f a)
	where
	aux (Just s) = return s
	aux Nothing  = fail "Error in XSD to Type conversion: : fetching an unexistant attribute"

hasAttribute :: t -> (t -> Maybe String) -> Bool
hasAttribute a f = aux (f a)
	where
	aux (Just _) = True
	aux Nothing  = False

-- | Look for root elements in schema
collectRootElementNames :: Schema -> [String]
collectRootElementNames  = catMaybes . foldr aux [] . xs'schemaTypes . parsedSchema
	where
	aux ((ThreeOf7 (Xs'element a _ _ _)), _) l = (xs'elementName a):(xs'elementRef a):l
	aux _ l = l

-- | Binding for Map.lookup
(!>) :: (MonadPlus m,Show a) => String -> Map String a -> m a
(!>) k m = case Map.lookup k m of
	Just s -> return s
	Nothing -> case Map.lookup (aux k) m of
			Just s -> return s
			Nothing -> fail $ "Error in XSD to Type conversion: : string '" ++ show k ++ "' not in map '" ++ show m ++ "'"
    where
    aux = tail' . dropWhile (/=':')
    tail' [] = []
    tail' l = tail l
