
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  stable
-- Portability :  portable
--
-- See
--
--     HSQL PostgreSQL binding
-- 
-- by
--
--     Digo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Text.SQL.HSQLAliases where

import Data.List as List
import Data.Char
import Data.TwoLevelExchangeMisc (fst3)
import Data.Map as Map (fromDistinctAscList)
import Data.SQL.DBaseTS
import Text.SQL.PrettyPrintDBaseTS
import Database.HSQL
import Control.Monad (liftM)
import Control.Exception (throwDyn,finally)

-- | Splits a valid SQL string into a list of SQL string statements.
statements :: String -> [String]
statements s
	| null (drop 1 r) = [newl]
	| otherwise = newl : statements (tail r)
	where
	(l,r) = break (==';') s
	newl = (dropWhile isSpace l) ++ ";"

-- | Filters CREATE statements from a SQL string.
creates :: String -> [String]
creates  = filter (isPrefixOf "CREATE") . statements

-- | Filters INSERT statements from a SQL string.
inserts :: String -> [String]
inserts = filter (isPrefixOf "INSERT") . statements

-- | Creates tables in a database from a 'DDL' description.
createDB :: Connection -> DDL -> IO ()
createDB c ddl = do
	let stmts = creates $ sshow ddl
	mapM (execute c) stmts >> return ()

-- | Inserts values in a database from a 'DML' description.
loadDB :: Connection -> DML -> IO ()
loadDB c dml = do
	let stmts = inserts $ sshow dml
	mapM (execute c) stmts >> return ()

-- | Reads values from a database into a 'DML' description.
-- Needs a 'DDL' table definition as argument because database keys are not created in database dumps.
dumpDB :: Connection -> DDL -> IO DML
dumpDB c (DB dbts) = do
	ts <- tables c
	liftM DB $ mapM (applyDBTable (liftM DBTable . dumpTable c ts)) dbts

-- | Fills the values from a database into a 'DBtable'.
dumpTable :: Connection -> [String] -> DBtable a b -> IO (DBtable a b)
dumpTable c tables (DBtable tn pks atts _)
	| (map toLower tn) `elem` tables = do
		l <- dumpField pks c tn
		r <- dumpField atts c tn
		return $ DBtable tn pks atts $ Map.fromDistinctAscList $ zip l r
	| otherwise = fail $ "table '" ++ tn ++ "' not found in database"
	where

-- | Converts values for SQL fields to 'DBtype' values.
dumpField :: DBtype a -> Connection -> String -> IO [a]
dumpField (Prod a b) c tn = do
	l <- dumpField a c tn
	r <- dumpField b c tn
	return $ zip l r
dumpField (Null t) c tn = do
	dumpt <- dumpField' t c tn
	return $ map mkNull $ snd dumpt
	where
	mkNull (Just x) = Left x
	mkNull Nothing = Right ()
dumpField (FK _ _ _ t) c tn = dumpField t c tn
dumpField t c tn = do
		(n,x) <- dumpField' t c tn
		mapM (testVal n) x
		where
		testVal _ (Just x) = return x
		testVal name _ = throwDyn (SqlFetchNull name)

dumpField' :: DBtype a -> Connection -> String -> IO (String,[Maybe a])
dumpField' (Integer (n,def)) c tn = dumpField'' n def c tn
dumpField' (Numeric (n,def) _ _) c tn = dumpField'' n def c tn
dumpField' (Boolean (n,def)) c tn = dumpField'' n def c tn
dumpField' (Float (n,def)) c tn = dumpField'' n def c tn
dumpField' (VarChar (n,def) max) c tn = do
	(name,x) <- dumpField'' n def c tn
	mapM (testMax max) x >> return (name,x) 
	where
	testMax :: Int -> Maybe String -> IO ()
	testMax max (Just s)
		| length s > max = fail $ "length of string '" ++ s ++ "' greater than its allowed maximum " ++ show max
		| otherwise = return ()
	testMax _ Nothing = return ()
dumpField' One c tn = do
	i <- countRows c tn
	return ("",replicate i $ Just ())

dumpField'' :: SqlBind a => String -> Maybe a -> Connection -> String -> IO (String,[Maybe a])
dumpField'' n (Just def) c tn = do
	let tname = map toLower tn
	d <- describe c tname
	let name = map toLower n
	if (name `elem` map fst3 d)
		then
		query c ("select " ++ n ++ " from " ++ tn) >>=
		collectRows (liftM Just . \s -> getFieldValue' s n def) >>=
		return . \l -> (n,l)
		else
		fail $ "field '" ++ n ++ "' not found in table '" ++ tn ++ "'"
dumpField'' n Nothing c tn = do
	let tname = map toLower tn
	d <- describe c tname
	let name = map toLower n
	if (name `elem` map fst3 d)
		then
		query c ("select " ++ n ++ " from " ++ tn) >>=
		collectRows (\s -> getFieldValueMB s name) >>=
		return . \x -> (n,x)
		else
		fail $ "field '" ++ n ++ "' not found in table '" ++ tn ++ "'"

-- | Returns the number of rows of a table.
countRows :: Connection -> String -> IO Int
countRows c tn = query c ("select * from " ++ tn) >>= countRows' 0
	where
	countRows' i stmt = do
		success <- fetch stmt
		if success
			then countRows' (i+1) stmt
			else return i `finally` closeStatement stmt








