
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     Parsec SQL Lexer
-- 
-- by
--
--     Digo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Text.SQL.Lexer where

import Char
import List
import Text.SQL.Tokens
import Text.SQL.ParsecCombinators
import Text.ParserCombinators.Parsec
import Pointless.Combinators

-- * Parser Combinators

-- ** Character validation functions

isPlusOrMinus x = (x == '+') || (x == '-')

isHorizSpace x = (x == ' ') || (x == '\t') || (x == '\f')

isNewline x = (x == '\n') || (x == '\r')

isSlashOrQuote x = (x == '\\') || (x == '\'')

isSlashOrStar x = (x == '/') || (x == '*')

isHexadecimal x = (isDigit x) || ((x >= 'A') && (x <= 'F')) || ((x >= 'a') && (x <= 'f'))

isDollar x = (isAlpha x) || (x == '_') || ((x >= '\200') && (x <= '\377'))

isDollarCont x = isDollar x || isDigit x

isSelf x = (x == '\37') || ((x >= '\40') && (x <= '\47')) || ((x >= '\58') && (x <= '\62')) || (x == '\91') || (x == '\93') || (x == '\94')

isOpChar x = or $ map (x ==) ['\33','\35','\37','\38','\42','\43','\45','\47','\60','\61','\62','\63','\64','\94','\96','\124','\126']

isAllowed x = or $ map (x ==) ['~','!','@','#','^','&','|','`','?','%']

between0And7 x = (x >= '0') && (x <= '7')

-- ** Regular Expressions

-- | space (regular expression = [ \t\n\r\f])
space :: CharParser st ()
space = do	inp <- getInput
		let (a,b) = span isSpace inp
		if (null a)
			then pzero
			else setInput b
-- ^ removes spaces from input

-- | horizontal space (regular expression = [ \t\f])
horiz_space :: CharParser st ()
horiz_space = do	inp <- getInput
			let (a,b) = span isHorizSpace inp
			if (null a)
				then pzero
				else setInput b
-- ^ removes horizontal spaces from input

-- | newline (regular expression = [\n\r])
newline :: CharParser st ()
newline = do	inp <- getInput
		let (a,b) = span isNewline inp
		if (null a)
			then pzero
			else setInput b
-- ^ removes newlines from input

-- | is not newline (regular expression = [^\n\r])
non_newline :: CharParser st ()
non_newline = do	inp <- getInput
			let (a,b) = span (not . isNewline) inp
			if (null a)
				then pzero
				else setInput b
-- ^ removes non-newlines from input

-- | line comment (regular expression = "--"{non_newline})
comment :: CharParser st ()
comment = curry snd <$> (string "--" >> return ()) <*> non_newline
-- ^ processes a line comment

-- | whitespace (regular expression = ({space}+|{comment}))
whitespace :: CharParser st ()
whitespace = ((oneOrMore $ Text.SQL.Lexer.space >> return []) >> return ()) <!> comment
-- ^ ignores spaces and comments

-- | special whitespace (regular expression = ({space}+|{comment}{newline}))
special_whitespace :: CharParser st ()
special_whitespace = ((oneOrMore $ Text.SQL.Lexer.space >> return []) >> return ()) <!> curry snd <$> comment <*> Text.SQL.Lexer.newline
-- ^ ignores multiple-lines whitespaces

-- | horizontal whitespace (regular expression = ({horiz_space}|{comment}))
horiz_whitespace :: CharParser st ()
horiz_whitespace = horiz_space <!> comment
-- ^ ignores horizontal spaces and comments

-- | whitespace with newline (regular expression = ({horiz_whitespace}*{newline}{special_whitespace}*))
whitespace_with_newline :: CharParser st ()
whitespace_with_newline = curry snd <$> (zeroOrMore $ horiz_whitespace >> return []) <*> curry fst <$> Text.SQL.Lexer.newline <*> (zeroOrMore $ special_whitespace >> return [])
-- ^ ignores horizontal spaces and newlines

-- | quote (regular expression = ')
quote :: CharParser st ()
quote = char '\'' >> return ()
-- ^ scans a quote character

-- | quoted string stop (regular expression = {quote}{whitespace}*)
quotestop :: CharParser st ()
quotestop = curry fst <$> quote <*> (zeroOrMore $ whitespace >> return [])
-- ^scans the last quote of a quoted string

-- | quoted string inside (regular expression = {quote}{whitespace_with_newline}){quote})
quotecontinue :: CharParser st ()
quotecontinue = curry snd <$> quote <*> curry snd <$> whitespace_with_newline <*> quote
-- ^ scans the string between quotes

-- | quoted string end (regular expression = {quote}{whitespace}*"-")
quotefail :: CharParser st ()
quotefail = curry fst <$> quote <*> (curry fst <$> (zeroOrMore $ whitespace >> return []) <*> char '-')
-- ^ a variant of quotestop

-- | bit string start (regular expression = [bB]{quote})
xbstart :: CharParser st ()
xbstart = curry snd <$> (char 'b' <!> char 'B') <*> quote
-- ^ scans the start of a bit string

-- | bit string value (regular expression = [^']*)
xbinside :: CharParser st String
xbinside = do	inp <- getInput
		let (a,b) = span (/='\'') inp
		if (null a)
			then pzero >> return []
			else setInput b >> return a
-- ^ scans the value of a bit string

-- | hexadecimal string start (regular expression = [xX]{quote})
xhstart :: CharParser st ()
xhstart = curry snd <$> (char 'x' <!> char 'X') <*> quote
-- ^ scans the start of an hexadecimal string

-- | hexadecimal string inside (regular expression = [^']*)
xhinside :: CharParser st String
xhinside = xbinside
-- ^ scans the value of an hexadecimal string

-- | national character string start (regular expression = [nN]{quote})
xnstart :: CharParser st ()
xnstart = curry snd <$> (char 'n' <!> char 'N') <*> quote
-- ^scans the start of a national character string

-- | quoted string with escape start (regular expression = [eE]{quote})
xestart :: CharParser st ()
xestart = curry snd <$> (char 'e' <!> char 'E') <*> quote
-- ^ scans the start of a quoted string that allows backslash escapes 

-- | quoted string start (regular expression = {quote})
xqstart :: CharParser st ()
xqstart = quote
-- ^ same as 'quote'

-- | extended double quote (regular expression = {quote}{quote})
xqdouble :: CharParser st ()
xqdouble = curry fst <$> quote <*> quote
-- ^ scans a double 'quote'

-- | quoted string inside (regular expression = [^\\']+)
xqinside :: CharParser st String
xqinside = do	inp <- getInput
		let (a,b) = span (not . isSlashOrQuote) inp
		if (null a)
			then pzero >> return []
			else setInput b	>> return a

-- | quoted string with escape (regular expression = [\\][^0-7])
xqescape :: CharParser st Char
xqescape = curry snd <$> char '\\' <*> satisfy (not . between0And7)

-- | octal string escape (regular expression = [\\][0-7]{1,3})
xqoctesc :: CharParser st String
xqoctesc = curry snd <$> char '\\' <*> (:) <$> satisfy between0And7 <*> ((:) <$> char '1' <*> ((:) <$> char '1' <*> (:[]) <$> char '1'))

-- | hexadecimal string escape (regular expression = [\\]x[0-9A-Fa-f]{1,2})
xqhexesc :: CharParser st String
xqhexesc = curry snd <$> char '\\' <*> curry snd <$> char 'x' <*> (:) <$> satisfy isHexadecimal <*> ((:) <$> char '1' <*> (:[]) <$> char '1')

-- | dollar quoted string start (regular expression = [A-Za-z\200-\377_])
dolq_start :: CharParser st Char
dolq_start = satisfy isDollar

-- _ dollar quoted string continue (regular expression = [A-Za-z\200-\377_0-9])
dolq_cont :: CharParser st Char
dolq_cont = satisfy isDollarCont

-- | dollar quoted string delimiters (regular expression = \$({dolq_start}{dolq_cont}*)?\$)
dolqdelim :: CharParser st String
dolqdelim = (:) <$> char '$' <*> ((++) <$> dolq <*> (:[]) <$> char '$')
	where
	dolq = (:) <$> dolq_start <*> (zeroOrMore $ (:[]) <$> dolq_cont) <!> succeed []

-- | dollar quoted string end (regular expression = \${dolq_start}{dolq_cont}*)
dolqfailed :: CharParser st String
dolqfailed = (:) <$> char '$' <*> ((:) <$> dolq_start <*> (zeroOrMore $ (:[]) <$> dolq_cont))

-- | dollar quoted string inside (regular expression = [^$]+)
dolqinside :: CharParser st String
dolqinside = oneOrMore $ (:[]) <$> satisfy (/='$')

-- | double quote (regular expression = \")
dquote :: CharParser st Char
dquote = char '"'

-- | double quoted string start (regular expression = {dquote})
xdstart :: CharParser st Char
xdstart = dquote

-- | double quoted string stop (regular expression = {dquote})
xdstop :: CharParser st Char
xdstop = dquote

-- | two quotes (regular expression = {dquote}{dquote})
xddouble :: CharParser st String
xddouble = (:) <$> dquote <*> ((:[]) <$> dquote)

-- | double quoted string inside (regular expression = [^\"]+)
xdinside :: CharParser st String
xdinside = (oneOrMore $ (:[]) <$> satisfy (/='"'))

-- | C comment start (regular expression = \/\*)
xcstart :: CharParser st ()
xcstart = curry (const () . (id >< id)) <$> char '/' <*> char '*'

-- | C comment stop (regular expression = \*+\/)
xcstop :: CharParser st ()
xcstop = curry (const () . (id >< id)) <$> (oneOrMore $ char '*' >> return []) <*> char '/'

-- C comment inside (regular expression = [^*/]+)
xcinside :: CharParser st ()
xcinside = (oneOrMore $ satisfy (not . isSlashOrStar) >> return []) >> return ()

-- | digit (regular expression = [0-9])
digit :: CharParser st Char
digit = satisfy isDigit
-- ^ returns a digit

-- | string identifier start (regular expression = [A-Za-z\200-\377_])
ident_start :: CharParser st Char
ident_start = dolq_start
-- ^ returns a dollar quote

-- | string identifier continue (regular expression = [A-Za-z\200-\377_0-9\$])
ident_cont :: CharParser st Char
ident_cont = satisfy isDollarCont <!> char '$'
-- ^ returns a dollar or dollar continue character

-- | string identifier (regular expression = {ident_start}{ident_cont}*)
identifier :: CharParser st String
identifier = (:) <$> ident_start <*> (zeroOrMore $ (:[]) <$> ident_cont)
-- ^ returns a string identifier

-- | typecast (regular expression = "::")
typecast :: CharParser st ()
typecast = string "::" >> return ()

-- | self character (regular expression = [,()\[\].;\:\+\-\*\/\%\^\<\>\=])
self :: CharParser st Char
self = satisfy isSelf
-- ^ returns a self character

-- | operator characters (regular expression = [\~\!\@\#\^\&\|\`\?\+\-\*\/\%\<\>\=])
op_chars :: CharParser st Char
op_chars =  satisfy isOpChar
-- ^ returns an operator character

-- | operator (regular expression = {op_chars}+)
operator :: CharParser st String
operator = oneOrMore $ (:[]) <$> op_chars
-- ^ returns an operator

-- | integer (regular expression = {digit}+)
integer :: CharParser st String
integer = oneOrMore $ (:[]) <$> Text.SQL.Lexer.digit
-- ^ returns an integer

-- | decimal (regular expression = (({digit}*\.{digit}+)|({digit}+\.{digit}*)))
decimal :: CharParser st String
decimal = (a <!> b)
	where
	a = (++) <$> (zeroOrMore $ (:[]) <$> Text.SQL.Lexer.digit) <*> ((:) <$> char '.' <*> (oneOrMore $ (:[]) <$> Text.SQL.Lexer.digit))
	b = (++) <$> (oneOrMore $ (:[]) <$> Text.SQL.Lexer.digit) <*> ((:) <$> char '.' <*> (zeroOrMore $ (:[]) <$> Text.SQL.Lexer.digit))
-- ^ returns a decimal

-- | real (regular expression = ({integer}|{decimal})[Ee][-+]?{digit}+)
real :: CharParser st String
real = (++) <$> (integer <!> decimal) <*> ((:) <$> (char 'e' <!> char 'E') <*> ((++) <$> (((:[]) <$> char '-') <!> ((:[]) <$> char '+') <!> succeed []) <*> (oneOrMore $ (:[]) <$> Text.SQL.Lexer.digit)))
-- ^ returns a real

-- | real failure (regular expression = ({integer}|{decimal})[Ee])
realfail1 :: CharParser st String
realfail1 = (++) <$> (integer <!> decimal) <*> ((:[]) <$> (char 'e' <!> char 'E'))
-- ^ returns the current real value

-- | real failure (regular expression = ({integer}|{decimal})[Ee][-+])
realfail2 :: CharParser st String
realfail2 = (++) <$> (integer <!> decimal) <*> ((:) <$> (char 'e' <!> char 'E') <*> ((:[]) <$> (char '-' <!> char '+')))
-- ^ returns the current real value

-- | parameter (regular expression = \${integer})
param :: CharParser st String
param = curry snd <$> char '$' <*> integer
-- ^ returns a parameter

-- | any character (regualr expression = .)
other :: CharParser st Char
other = do	inp <- getInput
		if (null inp)
			then pzero
			else (setInput $ tail inp) >> (return $ head inp)
-- ^ returns the scanned character

-- ** Environments

findComment :: String -> Int
findComment = uncurry max . ((tk_indexmb . findCComment) /\ (tk_indexmb . findDashComment))
	where
	tk_indexmb (Just x) = x
	tk_indexmb (Nothing) = -1
	
	findCComment :: String -> Maybe Int
	findCComment str
		| slash == Nothing || star == Nothing = Nothing
        	| tk_indexmb star == (tk_indexmb slash) + 1 = slash
		| otherwise = Nothing
	     where
       	     slash = findIndex (=='/') str
	     star = findIndex (=='*') str

	findDashComment :: String -> Maybe Int
	findDashComment str
		| length indices < 2 = Nothing
                | (!!) indices 1 == (head indices) + 1 = Just $ head indices
                | otherwise = Nothing
             where
	     indices = findIndices (=='-') str
-- ^ finds a comment

stringToOctalChar :: String -> Char
stringToOctalChar str = chr $ fromEnum (read ("0o" ++ str) :: Integer)
-- ^ converts a string to an octal
 
stringToHexChar :: String -> Char
stringToHexChar str = chr $ fromEnum (read ("0x" ++ str) :: Integer)
-- ^converts a string to an hexadecimal character

-- | escape characters
unescape_single_char :: Char -> Char
unescape_single_char x =
	case x of
	 'b' -> '\b'
	 'f' -> '\f'
	 'n' -> '\n'
	 'r' -> '\r'
	 't' -> '\t'
	 otherwise -> x
-- ^ create escape characters

yyless :: Int -> String -> CharParser st ()
yyless n inp = setInput $ drop n inp
-- ^ clone of Flex's yyless(n) that  returns all but the first n characters of the current token back to the input stream

namedatalen = 64
-- ^ the maximum string size

-- | the environment sensible scanner selector
getEnv :: Int -> CharParser (Int,Int,String,String) [Token]
getEnv env = case env of
	 0 ->
		    (		whitespaceINITIAL	>> return [])
		<!> (		xcstartINITIAL		>> return [])
		<!> (		xbstartINITIAL		>> return [])
		<!> (		xhstartINITIAL		>> return [])
		<!> ((:[]) <$>	xnstartINITIAL			    )
		<!> (		xqstartINITIAL		>> return [])
		<!> (		xestartINITIAL		>> return [])
		<!> (		dolqdelimINITIAL	>> return [])
		<!> ((:[]) <$>	dolqfailedINITIAL		    )
		<!> (		xdstartINITIAL		>> return [])
		<!> ((:[]) <$>	typecastINITIAL			    )
		<!> ((:[]) <$>	operatorINITIAL			    )
		<!> ((:[]) <$>	selfINITIAL			    )
		<!> ((:[]) <$>	paramINITIAL			    )
		<!> ((:[]) <$>	decimalINITIAL			    )
		<!> ((:[]) <$>	realINITIAL			    )
		<!> ((:[]) <$>	realfail1INITIAL		    )
		<!> ((:[]) <$>	realfail2INITIAL		    )
		<!> ((:[]) <$>	integerINITIAL			    )
		<!> ((:[]) <$>	identifierINITIAL		    )
		<!> ((:[]) <$>	otherINITIAL			    )
		<!> (		setState (-1,0,"","")	>> return [])

	 1 ->
		    (		xcstartXC		>> return [])
		<!> (		xcstopXC		>> return [])
		<!> (		xcinsideXC		>> return [])
		<!> (		op_charsXC		>> return [])
		<!> (		asteriskXC		>> return [])
		<!> (		eofXC			>> return [])

	 2 ->
		    ((:[]) <$>	quotestopXB			    )
		<!> (		xbinsideXB		>> return [])
		<!> (		quotecontinueXB		>> return [])
		<!> (		eofXB			>> return [])

	 3 ->
		    (		xhinsideXH		>> return [])
		<!> (		quotecontinueXH		>> return [])
		<!> ((:[]) <$>	quotestopXH			    )
		<!> (		eofXH			>> return [])
	
	 4 ->
		    ((:[]) <$>	quotestopXQ			    )
		<!> (		xqdoubleXQ		>> return [])
		<!> (		xqinsideXQ		>> return [])
		<!> (		xqescapeXQ		>> return [])
		<!> (		xqoctescXQ		>> return [])
		<!> (		xqhexescXQ		>> return [])
		<!> (		quotecontinueXQ		>> return [])
		<!> (		otherXQ			>> return [])
		<!> (		eofXQ			>> return [])
	
	 5 ->
		    (		dolqdelimXDOLQ			    )
		<!> (		dolqinsideXDOLQ		>> return [])
		<!> (		dolqfailedXDOLQ		>> return [])
		<!> (		otherXDOLQ		>> return [])
		<!> (		eofXDOLQ		>> return [])
	
	 6 ->
		    ((:[]) <$>	xdstopXD			    )
		<!> (		xddoubleXD		>> return [])
		<!> (		xdinsideXD		>> return [])
		<!> (		eofXD			>> return [])
	
	 otherwise -> unexpected "Unknown environment" >> return []
-- ^ applies a different lexer according to the current environment state
-- 
-- State	Attribute description
-- (Int		current environment
-- ,Int		depth of C-like comments
-- ,String	literal buffer
-- ,String	token accumulator or dolqstart accumulator
-- )

-- | the lexer function
lexPostgreSQL = runParser lexer (0,0,"","") "root"
-- ^ applies the lexer to a SQL contents string and returns a '[Token]'

-- | the main monadic lexer function
lexer :: CharParser (Int,Int,String,String) [Token]
lexer = do
	(env,_,_,_) <- getState
	case env of
	 -1 -> eof >> return []
	 otherwise -> (++) <$> getEnv env <*> lexer

-- *** Initial Environment (Code 1)

whitespaceINITIAL :: CharParser st ()
whitespaceINITIAL = whitespace

xcstartINITIAL :: CharParser (Int,Int,String,String) ()
xcstartINITIAL = do	xcstart
			setState (1,0,"","")
			return ()

xbstartINITIAL :: CharParser (Int,Int,String,String) ()
xbstartINITIAL = do	xbstart
			(_,y,_,z) <- getState
			setState (2,y,"b",z)
		
xhstartINITIAL :: CharParser (Int,Int,String,String) ()
xhstartINITIAL = do	xhstart
			(_,y,_,z) <- getState
			setState (3,y,"x",z)

xnstartINITIAL :: CharParser st Token
xnstartINITIAL = do	inp <- getInput
			xnstart
			yyless 1 inp
			st <- getParserState
			return $ tk_mb $ scanKeyword "nchar" $ statePos st
	where
	tk_mb (Just x) = x
	tk_mb Nothing = error "keyword nchar not in keywords list as it should"

xqstartINITIAL :: CharParser (Int,Int,String,String) ()
xqstartINITIAL = do	xqstart
			(_,y,_,w) <- getState
			setState (4,y,"",w)
			
xestartINITIAL :: CharParser (Int,Int,String,String) ()
xestartINITIAL = do	xestart
			(_,y,_,w) <- getState
			setState (4,y,"",w)

dolqdelimINITIAL :: CharParser (Int,Int,String,String) ()
dolqdelimINITIAL = do	str <- dolqdelim
			(_,y,_,_) <- getState
			setState (5,y,"",str)

dolqfailedINITIAL :: CharParser st Token
dolqfailedINITIAL = do	inp <- getInput
			string <- dolqfailed
			yyless 1 inp
			st <- getParserState
			return $ Token (TokSymbol $ head string) $ statePos st

xdstartINITIAL :: CharParser (Int,Int,String,String) ()
xdstartINITIAL = do	xdstart
			(_,y,z,w) <- getState
			setState (6,y,"",w)

typecastINITIAL :: CharParser st Token
typecastINITIAL = do	typecast
			st <- getParserState
			return $ Token TokTypecast $ statePos st

selfINITIAL :: CharParser st Token
selfINITIAL = do	char <- self
			st <- getParserState
			return $ Token (TokSymbol char) $ statePos st

operatorINITIAL :: CharParser (Int,Int,String,String) Token
operatorINITIAL = do
-- we will use the state for local use, and then restore it, because we can't have let inside an if/then/else clause
-- (_,nchars/nchars1,_,_)
			inp  <- getInput
			text <- operator
			(x,y,z,w) <- getState
			
			let n = findComment text
			if (n == -1)
			 then setState (0,length text,"","")
			 else setState (0,n,"","")

			(_,nchars,_,_) <- getState
			let str = take nchars text
			let i = findIndex isPlusOrMinus str
		 	if (i == Nothing)
			 then setState (0,nchars,"","")
			 else
			  if ((findIndex isAllowed (take (succ $ tk_mb i) str)) == Nothing)
			   then setState (0,x,"","")
			   else setState (0,nchars,"","")
			(_,nchars1,_,_) <- getState
			yyless nchars1 inp

			-- restore the original state
			setState (x,y,z,w)
			
			st <- getParserState
			let pos = statePos st
			
			if ((length text == 1) && (isSelf $ head text))
			 then return $ Token (TokSymbol $ head text) pos
			 else
			  if (nchars1 > namedatalen)
			   then unexpected "operator too long"
			   else
			    if (text == "!=")
			     then return $ Token (TokOp "<>") pos
			     else return $ Token (TokOp $ take nchars1 str) pos
	where
	tk_mb (Just j) = j

paramINITIAL :: CharParser st Token
paramINITIAL = do	text <- param
			st <- getParserState
			return $ Token (TokParam (read text :: Int)) $ statePos st

integerINITIAL :: CharParser st Token
integerINITIAL = do	text <- integer
			let ival = (read text :: Int)
			st <- getParserState
			let pos = statePos st
			if (ival < 0)
				then return $ Token (TokFconst text) pos
				else return $ Token (TokIconst ival) pos

decimalINITIAL :: CharParser st Token
decimalINITIAL = do	text <- decimal
			st <- getParserState
			return $ Token (TokFconst text) $ statePos st

realINITIAL :: CharParser st Token
realINITIAL = do	text <- real
			st <- getParserState
			return $ Token (TokFconst text) $ statePos st

realfail1INITIAL :: CharParser st Token
realfail1INITIAL = do	inp <- getInput
			text <- realfail1
			yyless (length text - 1) inp
			st <- getParserState
			return $ Token (TokFconst text) $ statePos st
		
realfail2INITIAL :: CharParser st Token
realfail2INITIAL = do	inp <- getInput
			text <- realfail2
			yyless (length text - 2) inp
			st <- getParserState
			return $ Token (TokFconst text) $ statePos st
		
identifierINITIAL :: CharParser st Token
identifierINITIAL = do	text <- identifier
			st <- getParserState
			let pos = statePos st
			case (scanKeyword text pos) of
				Nothing -> return $ Token (TokIdent $ take namedatalen text) pos
				Just key -> return key

otherINITIAL :: CharParser st Token
otherINITIAL = do	char <- other
			st <- getParserState
			return $ Token (TokSymbol char) $ statePos st

-- ** Double quoted string Environment (Code 6)

xdstopXD :: CharParser (Int,Int,String,String) Token
xdstopXD = do	str <- xdstop
		(_,y,z,w) <- getState
		setState (0,y,z,w)
		if (null z)
			then unexpected "zero-length delimited identifier"
			else
				do
				st <- getParserState
				return $ Token (TokIdent $ take namedatalen z) $ statePos st

xddoubleXD :: CharParser (Int,Int,String,String) ()
xddoubleXD = do 	xddouble
			(x,y,z,w) <- getState
			setState (x,y,z ++ ((:[]) '"'),w)

xdinsideXD :: CharParser (Int,Int,String,String) ()
xdinsideXD = do	text <- xdinside
		(x,y,z,w) <- getState
		setState (x,y,z ++ text,w)

eofXD :: CharParser st ()
eofXD = curry fst <$> eof <*> unexpected "unterminated quoted identifier"

-- ** Dollar quoted string Environment (Code 5)

dolqdelimXDOLQ :: CharParser (Int,Int,String,String) [Token]
dolqdelimXDOLQ = do	(x,y,z,w) <- getState
			inp <- getInput
			str <- dolqdelim
			if (w == str)
				then
					do
					setState (0,y,z,"")
					st <- getParserState
					return [Token (TokSconst z) $ statePos st]
				else yyless (length str - 1) inp >> setState (x,y,z ++ (init str),w) >> return []

dolqinsideXDOLQ :: CharParser (Int,Int,String,String) ()
dolqinsideXDOLQ = do	(x,y,z,w) <- getState
			str <- dolqinside
			setState (x,y,z ++ str,w)

dolqfailedXDOLQ :: CharParser (Int,Int,String,String) ()
dolqfailedXDOLQ = do	(x,y,z,w) <- getState
			str <- dolqfailed
			setState (x,y,z ++ str,w)

otherXDOLQ :: CharParser (Int,Int,String,String) ()
otherXDOLQ = do	char <- other
		(x,y,z,w) <- getState
		setState (x,y,z ++ ((:[]) char),w)

eofXDOLQ :: CharParser st ()
eofXDOLQ = curry fst <$> eof <*> unexpected "unterminated dollar-quoted string"

-- ** Quoted string Environment (Code 4)

quotestopXQ :: CharParser (Int,Int,String,String) Token
quotestopXQ = do	inp <- getInput
			quotestop <!> quotefail
			yyless 1 inp
			(_,y,z,w) <- getState
			setState (0,y,"",w)
			st <- getParserState
			return $ Token (TokSconst z) $ statePos st

xqdoubleXQ :: CharParser (Int,Int,String,String) ()
xqdoubleXQ = do	xqdouble
		(x,y,z,w) <- getState
		setState (x,y,z ++ "\'",w)
	
xqinsideXQ :: CharParser (Int,Int,String,String) ()
xqinsideXQ = do	(x,y,z,w) <- getState
		text <- xqinside
		setState (x,y,z ++ text,w)

xqescapeXQ :: CharParser (Int,Int,String,String) ()
xqescapeXQ = do	char <- xqescape
		case char of
			'\'' -> label (succeed ()) "nonstandard use of \\' in a string literal\nUse '' to write quotes in strings, or use the escape string syntax (E'...')."
			'\\' -> label (succeed ()) "nonstandard use of \\\\ in a string literal\nUse the escape string syntax for backslashes, e.g., E'\\\\'."
			otherwise -> label (succeed ()) "nonstandard use of escape in a string literal\nUse the escape string syntax for escapes, e.g., E'\\r\\n'."
		(x,y,z,w) <- getState
		setState (x,y,z ++ ((:[]) $ unescape_single_char char),w)

xqoctescXQ :: CharParser (Int,Int,String,String) ()
xqoctescXQ = do	str <- xqoctesc
		label (succeed ()) "nonstandard use of escape in a string literal\nUse the escape string syntax for escapes, e.g., E'\\r\\n'."
		(x,y,z,w) <- getState
		setState (x,y,z ++ ((:[]) $ stringToOctalChar str),w)

xqhexescXQ :: CharParser (Int,Int,String,String) ()
xqhexescXQ = do	str <- xqhexesc
		label (succeed ()) "nonstandard use of escape in a string literal\nUse the escape string syntax for escapes, e.g., E'\\r\\n'."
		(x,y,z,w) <- getState
		setState (x,y,z ++ ((:[]) $ stringToHexChar str),w)

quotecontinueXQ :: CharParser st ()
quotecontinueXQ = quotecontinue

otherXQ :: CharParser (Int,Int,String,String) ()
otherXQ = do	char <- other
		(x,y,z,w) <- getState
		setState (x,y,z ++ ((:[]) char),w)

eofXQ :: CharParser st ()
eofXQ = curry fst <$> eof <*> unexpected "unterminated quoted string"

-- ** Bit string environment (Code 2)

quotestopXB :: CharParser (Int,Int,String,String) Token
quotestopXB = do	inp <- getInput
			quotestop <!> quotefail
			yyless 1 inp
			(_,y,str,w) <- getState
			setState (0,y,"",w)
			st <- getParserState
			return $ Token (TokBconst str) $ statePos st

xbinsideXB :: CharParser (Int,Int,String,String) ()
xbinsideXB = do	(x,y,z,w) <- getState
		text <- xbinside
		setState(x,y,z ++ text,w)

quotecontinueXB :: CharParser st ()
quotecontinueXB = quotecontinue

eofXB :: CharParser st ()
eofXB = curry fst <$> eof <*> unexpected "unterminated bit string literal"

-- ** Hexadecimal string Environment (Code 3)
xhinsideXH :: CharParser (Int,Int,String,String) ()
xhinsideXH = do	(x,y,z,w) <- getState
		text <- xhinside
		setState (x,y,z ++ text,w)
		
quotecontinueXH :: CharParser st ()
quotecontinueXH = quotecontinue

quotestopXH :: CharParser (Int,Int,String,String) Token
quotestopXH = do	inp <- getInput
			quotestop <!> quotefail
			yyless 1 inp
			(_,y,str,z) <- getState
			setState (0,y,"",z)
			st <- getParserState
			return $ Token (TokXconst str) $ statePos st

eofXH :: CharParser st ()
eofXH = curry fst <$> eof <*> unexpected "unterminated hexadecimal string literal"

-- ** C-like comment Environment (Code 1)

xcstartXC :: CharParser (Int,Int,String,String) ()
xcstartXC = do	xcstart
		(_,xcdepth,z,w) <- getState
		setState $ (1,xcdepth + 1,z,w)
		
xcstopXC :: CharParser (Int,Int,String,String) ()
xcstopXC = do	xcstop
		(_,xcdepth,z,w) <- getState
		if (xcdepth == 0)
			then setState (0,xcdepth,z,w)
			else setState $ (1,xcdepth - 1,z,w)

xcinsideXC :: CharParser st ()
xcinsideXC = xcinside

op_charsXC :: CharParser st ()
op_charsXC = op_chars >> return ()

asteriskXC :: CharParser st ()
asteriskXC = (oneOrMore $ char '*' >> return []) >> return ()

eofXC :: CharParser st ()
eofXC = curry fst <$> eof <*> unexpected "unterminated /* comment"
