
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  dtlapa@yahoo.com, flavioxavier@gmail.com, hpacheco@gmail.com
-- Stability   :  stable
-- Portability :  portable
--
-- See
--
--     Pretty Printing for the DBaseTS type
-- 
-- by
--
--     Digo Lapa, Flavio Ferreira and Hugo Pacheco
--
-----------------------------------------------------------------------------

module Text.SQL.PrettyPrintDBaseTS (
    SQLShow(..)
    ) where

import Data.SQL.DBaseTS as TS
import Pointless.Combinators
import GHC.Base as Base
import Data.Map as Map
import Data.List as List
import Data.TwoLevelExchangeMisc

-- * Class SQLShow

-- | Class responsible for creating valid SQL from the type
class SQLShow a where
    sshow :: a -> String

-- * Instances of SQLShow

instance SQLShow DBase where
	sshow (DB l) = concat $ Base.map sshow l

instance SQLShow DBTable where
	sshow (DBTable x) = sshow x

instance SQLShow (DBtable a b) where
	sshow (DBtable n pks TS.One m) = "CREATE TABLE " ++ n ++ " (\n" ++
				       concat (mapDBType sshow (DBType pks)) ++
				       "\tPRIMARY KEY" ++ drawPks ++ "\n"
				       ++ drawFks ++ "\t\n);\n\n" ++
				       drawInserts m ++ "\n"
		where		
		drawPks = parenthesis $ mapDBType (applyDBType getName) $ DBType pks
		drawFks = joinFKeys $ concat $ mapDBType (applyDBType getFKeys) $ DBType pks
		
		drawInserts = concat . Base.map drawInsert . uncurry zip .(Map.keys /\ Map.elems)

		drawInsert (k,a) = "INSERT INTO " ++ n ++ " " ++ parenthesis variables ++ "\n\t VALUES " ++ parenthesis (drawValues $ DBValue pks k) ++ ";\n"
		
		variables = mapDBType (applyDBType getName) $ DBType pks

	sshow (DBtable n pks atts m) = "CREATE TABLE " ++ n ++ " (\n" ++
				       concat (mapDBType sshow (DBType pks)) ++
				       concat (mapDBType sshow (DBType atts)) ++
				       "\tPRIMARY KEY" ++ drawPks ++ "\n"
				       ++ drawFks ++ "\t\n);\n\n" ++
				       drawInserts m ++ "\n"
		where		
		drawPks = parenthesis $ mapDBType (applyDBType getName) $ DBType pks
		drawFks = joinFKeys $ concat $ mapDBType (applyDBType getFKeys) $ DBType $ TS.Prod pks atts
		
		drawInserts = concat . Base.map drawInsert . uncurry zip .(Map.keys /\ Map.elems)

		drawInsert (k,a) = "INSERT INTO " ++ n ++ " " ++ parenthesis variables ++ "\n\t VALUES " ++ parenthesis ((drawValues $ DBValue pks k) ++ (drawValues $ DBValue atts a)) ++ ";\n"
		
		variables = mapDBType (applyDBType getName) $ DBType $ TS.Prod pks atts

joinFKeys :: [(Maybe String,VariableName,TableName,VariableName)] -> String
joinFKeys [] = []
joinFKeys lst@(x:_) = showFK l ++ joinFKeys r
	where
	name a = case (fst4 $ head a) of
		Nothing -> ",\n"
		(Just n) -> ",\nCONSTRAINT " ++ n ++ " "
	(l,r) = List.partition (sameFKey x) lst
	sameFKey (n1,_,t1,_) (n2,_,t2,_) = n1 == n2 && t1 == t2
	showFK x = name x ++ "FOREIGN KEY " ++ (parenthesis $ List.map snd4 x) ++ " REFERENCES " ++ (thr4 $ head x) ++ (parenthesis $ List.map fth4 x)

instance SQLShow DBType where
	sshow (DBType x) = sshow x

instance SQLShow DBValue where
	sshow = show

instance SQLShow (DBtype a) where
	sshow (FK _ _ _ t) = sshow t
	sshow (Null x) = sshow' True x
	sshow x = sshow' False x
		
drawNull True = " NULL"
drawNull False = " NOT NULL"

drawDefault Nothing = ""
drawDefault (Just x) = " DEFAULT (" ++ show x ++ ")"

sshow' :: Bool -> DBtype a -> String
sshow' _ (TS.Prod a b)      = sshow a ++ sshow b
sshow' b (VarChar (n,def) i)   = '\t' : n ++ "\tVARCHAR(" ++ show i ++ ")" ++ drawNull b ++ drawDefault def ++ ",\n"
sshow' b (Numeric (n,def) x y) = '\t' : n ++ "\tNUMERIC(" ++ show x ++ ',' : show y ++ ")" ++ drawNull b ++ drawDefault def ++ ",\n"
sshow' b (Integer (n,def))     = '\t' : n ++ "\tINTEGER " ++ "" ++ drawNull b ++ drawDefault def ++ ",\n"
sshow' b (Boolean (n,def))     = '\t' : n ++ "\tBOOLEAN " ++ "" ++ drawNull b ++ drawDefault def ++ ",\n"
sshow' b (Float (n,def))	      = '\t' : n ++ "\tFLOAT " ++ "" ++ drawNull b ++ drawDefault def ++ ",\n"
sshow' _ TS.One = "\tTLT_One\tBOOLEAN NOT NULL DEFAULT (True),\n"


